require 'common.glutil'
require 'common.util'
require 'common.hsv'
require 'common.pingpong'
require 'common.shader'
require 'fractal.vec'

hsvTex = HsvTex.new(1024)

log2size = 10	--highest log2 texture dimension before the gpu craps itself
size = glush.bitshiftleft(1,log2size)
texs = PingPong.new{
	width=size;
	height=size;
	internalFormat=gl.GL_RGBA_FLOAT32;
	minFilter=gl.GL_NEAREST;
	magFilter=gl.GL_NEAREST;
}

-- create our fractal shader

iterateShader = ShaderProgram.new{
	vertexCode=[[
varying vec2 c, tc;	// transformed coordinate
void main() {
// huh ... you can't multiply the texture matrix, but you can multiply the modelview matrix ... huh ...
	c = gl_Vertex.xy;
	tc = (gl_ModelViewMatrix * gl_Vertex).xy;
	gl_Position = gl_ProjectionMatrix * gl_Vertex;
}
]];
	fragmentCode=[[
varying vec2 c, tc;
uniform sampler2D tex;
uniform float reset;

void main() {
	vec4 n = texture2D(tex, c);	// here's how it'll go: the first two componnets are z[n], the second two are the position in C
	if (n.z == 0.) {
		//there's a mac ati compiler bug:
		// this n2/n was combined into one line
		//and that worked fine in xp
		//but on mac it messed up.  i'm guessing it calc'd one then the other then used that for both
		// (like i was getting in my gpu flame fractal code)
		vec2 n2 = vec2(n.x*n.x - n.y*n.y + tc.x, 2.*n.x*n.y + tc.y);
		n = vec4(n2.x, n2.y, 0., n.w + 1.);
		if (n.x*n.x+n.y*n.y > 4.) n.z = 1.;		// use n.z as our 'has bailed out' flag
	}
	gl_FragColor = n * reset;	// add tc for mandelbrot, add some constant for julia
}
]];
	uniforms={
		tex=0;
	};
}

renderShader = ShaderProgram.new{
	vertexCode=[[
varying vec2 tc;
void main() {
	tc = gl_MultiTexCoord0.xy;	// should be [0,1]^2
	gl_Position = gl_ProjectionMatrix * gl_Vertex;
}]];
	fragmentCode=[[
varying vec2 tc;
uniform sampler2D tex;
uniform sampler1D grad;
void main() {
	vec4 n = texture2D(tex,tc);
	if (n.z == 0.) {
		gl_FragColor = vec4(0., 0., 0., 1.);
	} else {

		//float f = length(n.xy) * 0.25;		//magnitude based
		float f = n.w / 100.;	//iteration based
		gl_FragColor = vec4( texture1D(grad, f).rgb, 1.);
	}
}]];
	uniforms={
		tex=0;
		grad=1;
	};
}

iter = 1

-- state
STATE = {
	RESET = {};
	UPDATE = {};
	PAUSE = {};
}
state = STATE.PAUSE
zoom = 3
pos = vec2:new()
dragDelta = vec2:new()
dragZoom = 1
mousepos = vec2:new()
mouselastpos = vec2:new()
mouseLeftDown = false
mouseLeftWasDown = false
mouseRightDown = false
mouseRightWasDown = false

function glush.display()
	-- mouse system getting kinks worked out
	do
		local x,y,b
		x,y,b = glush.mouse()
		mouselastpos[1] = mousepos[1]
		mouselastpos[2] = mousepos[2]
		mousepos[1] = x
		mousepos[2] = y
		if state == STATE.PAUSE then
			state = STATE.RESET
		else
			state = STATE.UPDATE
			dragDelta:set(0,0)
			dragZoom = 1
		end
		
		mouseLeftWasDown = mouseLeftDown
		mouseLeftDown = glush.bitand(b,1) == 1 

		mouseRightWasDown = mouseRightDown
		mouseRightDown = glush.bitand(b,2) == 2
	end
	
	do
		if mouseLeftDown then
			local screenDelta = mousepos - mouselastpos
			local delta = screenDelta * zoom
			dragDelta = dragDelta + screenDelta
			pos = pos - delta
			-- if we're in the process of panning then set the state to 'dont update, dont reset'
			state = STATE.PAUSE
		elseif mouseRightDown then
			local screenZoom = math.pow(2, 10*(mousepos[2] - mouselastpos[2]))
			dragZoom = dragZoom * screenZoom
			zoom = zoom / screenZoom
			state = STATE.PAUSE
		end
	end
	--done n done

	glreport('fbo start')
	
	--print(table.concat(getTexData(texs:cur().id), ','))

	if state ~= STATE.PAUSE then
		texs:swap()
		texs:draw{
			viewport={0,0,texs.width,texs.height};
			resetProjection=true;
			shader=iterateShader;
			texs={texs:prev()};
			uniforms={
				iter=iter;
				reset=({[STATE.RESET]=0})[state] or 1;
			};
			callback=function()
				gl.glMatrixMode(gl.GL_MODELVIEW)
				gl.glPushMatrix()
				gl.glTranslate(pos[1], pos[2],0)
				gl.glScale(zoom,zoom,0)
				gl.glTranslate(-.5, -.5, 0)
				drawScreenQuad()
				gl.glPopMatrix()
				
			end
		}

		iter = iter + 1
	end
	glreport('fbo done')

	-- next run them through a remapping program (or save the iteration in the texture) to render to screen

	gl.glMatrixMode(gl.GL_PROJECTION)
	gl.glLoadIdentity()
	gl.glOrtho(0,1,0,1,-1,1)
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glLoadIdentity()

	renderShader:use()

	gl.glActiveTexture(gl.GL_TEXTURE0)
	texs:cur():bind()
	gl.glActiveTexture(gl.GL_TEXTURE1)
	hsvTex:bind()

	-- todo - fill the viewport area - and make sure the viewport is aspect ratio correct
	local vtxs = {
		vec2:new(0,0);
		vec2:new(1,0);
		vec2:new(1,1);
		vec2:new(0,1);
	}

	gl.glBegin(gl.GL_QUADS)
	for _,v in ipairs(vtxs) do
		gl.glTexCoord(v)
		gl.glVertex(dragDelta + (v - vec2:new(.5,.5)) * dragZoom + vec2:new(.5,.5))
	end
	gl.glEnd()
	
	gl.glActiveTexture(gl.GL_TEXTURE1)
	hsvTex:unbind()
	gl.glActiveTexture(gl.GL_TEXTURE0)
	texs:cur():unbind()
	
	ShaderProgram.useNone()
end

