/*

    File: bindings.cpp

    Copyright (C) 2000-2010 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

/*
I need to separate stuff out ...
which most likely means doing so per-extension to keep up with what I've already done
but what about the OpenGL versions?
and what about loading them?
I should look into making use of GLEW
*/

#include <iostream>
#include <string>
#include <sstream>

#include "math/vec.h"
#include "system/exception.h"
#include "system/gl.h"
#include "base/main.h"
#include "base/filebuffer.h"
#include "base/glext.h"
#include "base/opengl.h"

#include "program.h"
#include "bindings.h"
#include "bindings_ARB_vertex_buffer_object.h"
#include "bindings_ARB_shader_objects.h"
#include "bindings_ARB_vertex_program.h"
#include "bindings_ARB_vertex_shader.h"
#include "bindings_EXT_framebuffer_object.h"
#include "bindings_util.h"

//because windows sucks
inline const char *safestr(const char *str) {
#ifdef WIN32
	return str ? str : "(null)";
#else
	return str;
#endif
}

#ifdef WIN32
#define strcasecmp _stricmp
#endif

using namespace std;

//I stole this from texture.cpp.  then expanded on it.
//TODO - make it a static class method, or a TextureSystem method
//lua error on fail
static int GL_Format_to_Components(lua_State *L, int format) {
	switch (format) {
	case GL_COLOR_INDEX:
	case GL_RED:
	case GL_GREEN:
	case GL_BLUE:
	case GL_ALPHA:
	case GL_LUMINANCE:
	case GL_DEPTH:		//glCopyPixels
	case GL_DEPTH_COMPONENT:	//glReadPixels
	case GL_STENCIL:	//glCopyPixels
	case GL_STENCIL_INDEX:	//glReadPixels
	case GL_ALPHA_FLOAT32_APPLE:
	case GL_INTENSITY_FLOAT32_APPLE:
	case GL_LUMINANCE_FLOAT32_APPLE:
		return 1;
	case GL_LUMINANCE_ALPHA:
		return 2;
	case GL_RGB:
	case GL_RGB_FLOAT32_APPLE:
		return 3;
	case GL_COLOR:	//glCopyPixels
	case GL_RGBA:
	case GL_RGBA_FLOAT32_APPLE:
		return 4;
		
	default:
		luaL_error(L, "got an unknown GL format: %d", format);
		break;
	}
	return 0;
}

static int GL_Type_to_BPP(int type) {
	switch (type) {
	case GL_BITMAP:			return 1;
	case GL_UNSIGNED_BYTE:
	case GL_BYTE:			return 8;
	case GL_UNSIGNED_SHORT:
	case GL_SHORT:			return 16;
	case GL_UNSIGNED_INT:
	case GL_INT:
	case GL_FLOAT:			return 32;
	default:	assert(!"got an unknown GL type!");	return 0;
	}
	return 0;
}


//OpenGL 1.1


LUAF1i1d		(glAccum)
LUAFi			(glActiveTextureARB)
LUAF1i1d		(glAlphaFunc)
LUAFi			(glArrayElement)
LUAFi			(glBegin)
LUAF2i			(glBindTexture)
LUAF2i			(glBlendFunc)
LUAFimany		(glCallList)		//can lua unravel tables to parameter lists? meh.  forget about calllists anyways
LUAFi			(glClear)
LUAF4d			(glClearAccum)
LUAF4d			(glClearColor)
LUAFd			(glClearDepth)
LUAFd			(glClearIndex)
LUAFi			(glClearStencil)

LUAFstart(glClipPlane)
	LUAFvarcase(2, glClipPlane(LUAPi(1), luaReadVec4d(L)))
	LUAFvarcase(5, glClipPlane(LUAPi(1), vec4d(LUAPd(2), LUAPd(3), LUAPd(4), LUAPd(5))))
	LUAFvarendcase("1 and table or 5")
LUAFend()

LUAFstart(glColor)
	if (argc == 1) {
		if (!lua_istable(L,1)) {
			luaL_error(L, "expected a table");
			return 0;
		}
		int len = lua_objlen(L, 1);
		if (len == 3) glColor3dv(luaReadVec3d(L));
		else if (len == 4) glColor4dv( luaReadVec4d(L));
		else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
	} else 
	LUAFvarcase(3, glColor3d(LUAPd(1), LUAPd(2), LUAPd(3)))
	LUAFvarcase(4, glColor4d(LUAPd(1), LUAPd(2), LUAPd(3), LUAPd(4)))
	LUAFvarendcase("table, 3, or 4")
LUAFend()

LUAF4b			(glColorMask)
LUAF2i			(glColorMaterial)

//not mt safe.
LUAFstart(glColorPointer)
	if (argc == 1) {
		static vector<float> buffer;
		buffer.resize(0);

		int t = lua_gettop(L);
		if (!lua_istable(L,t)) luaL_error(L, "expected a table as the single param");

		int plen = -1;	//persistant length of the member tables
		
		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			//fprintf(stderr, "%s - %s\n", lua_typename(L, lua_type(L, -2)), lua_typename(L, lua_type(L, -1)));
				
			int t2 = lua_gettop(L);
			if (!lua_istable(L,t2)) luaL_error(L, "found vector elem %s that wasnt a table", lua_tostring(L,-2));

			int newlen = lua_objlen(L,t2);	//make sure the length of the subtables are all the same
			if (plen == -1) plen = newlen;	//save the first length as persistent
			if (plen != newlen) luaL_error(L, "i was reading vecs of length %d when i found one of differing length %d", plen, newlen);

			lua_pushnil(L);
			while (lua_next(L,t2) != 0) {
				buffer.push_back((float)lua_tonumber(L,-1));
				lua_pop(L,1);
			}

			lua_pop(L,1);
		}
		
		//done!
		glColorPointer(plen, GL_FLOAT, 0, &(buffer[0]));
		
	} else if (argc == 4) {
		//TODO - won't (char*) casting mess up with windows when it defaults char to wchar_t ?
		//	... and what's a good cross-platform workaround for that
		glColorPointer(LUAPi(1), LUAPi(2), LUAPi(3), (char*)NULL+LUAPi(4));
	} else {
		luaL_error(L, "expected either 1 or 4 args");
	}
LUAFend()

LUAF5i	(glCopyPixels)
LUAF7i	(glCopyTexImage1D)
LUAF8i	(glCopyTexImage2D)
LUAF6i	(glCopyTexSubImage1D)
LUAF8i	(glCopyTexSubImage2D)
LUAFi	(glCullFace)
LUAF2i	(glDeleteLists)			//unlike textures, lists will be allowed their range functionality, because they allocate congruently

LUAFstart(glDeleteTexture)
	for (int i = 0; i < argc; i++) {
		GLuint id = LUAPi(i);
		glDeleteTextures(1, &id);
	}
LUAFend()

LUAFi			(glDepthFunc)
LUAFb			(glDepthMask)
LUAF2d			(glDepthRange)
LUAFimany		(glDisable)
LUAFimany		(glDisableClientState)
LUAF3i			(glDrawArrays)
LUAFi			(glDrawBuffer)

//not MT safe
LUAFstart(glDrawElements)
	static vector<int> indexes;

	int geom = LUAPi(1);
	int elems = LUAPi(2);
	int type = GL_UNSIGNED_INT;
		
	//if the args aren't re-provided then they are re-membered!
	if (argc > 3) {
		int t = 3;
		if (!lua_istable(L,t)) luaL_error(L, "expected a table for the 3rd param");
			
		indexes.resize(0);
			
		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			indexes.push_back((int)lua_tonumber(L,-1));
			lua_pop(L,1);
		}
	}
		
	if (elems > indexes.size()) luaL_error(L, "you asked to render %d indexes but we only found %d in the table", elems, indexes.size());

	glDrawElements(geom, elems, type, &(indexes[0]));
LUAFend()


LUAFb(glEdgeFlag)

//not mt safe.
LUAFstart(glEdgeFlagPointer)

	if (argc == 1) {
		static vector<GLboolean> buffer;
		buffer.resize(0);

		int t = lua_gettop(L);
		if (!lua_istable(L,t)) luaL_error(L, "expected a table as the single param");

		int plen = -1;	//persistant length of the member tables
			
		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			buffer.push_back((GLboolean)lua_toboolean(L,-1));
			lua_pop(L,1);
		}
			
		//done!
		glEdgeFlagPointer(0, &(buffer[0]));
			
	} else if (argc == 2) {
		glEdgeFlagPointer(LUAPi(1), (char*)NULL+LUAPi(2));
	} else {
		luaL_error(L, "expected either 2 or 4 args");
	}
LUAFend()

LUAFimany		(glEnable)
LUAFimany		(glEnableClientState)

LUAFstart(glEvalCoord)
	if (argc == 1) {
		if (lua_istable(L,1)) {
			int len = lua_objlen(L, 1);
			if (len == 1) glEvalCoord1dv(luaReadVec1d(L));
			if (len == 2) glEvalCoord2dv(luaReadVec2d(L));
			else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
		} else {
			glEvalCoord1d(LUAPd(1));
		}
	} else 
	LUAFvarcase(2, glEvalCoord2d(LUAPd(1), LUAPd(2)))
	LUAFvarendcase("table, 1, or 2")
LUAFend()

LUAFstart(glEvalMesh)
	LUAFvarcase(3, glEvalMesh1(LUAPi(1), LUAPi(2), LUAPi(3)))
	LUAFvarcase(5, glEvalMesh2(LUAPi(1), LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5)))
	LUAFvarendcase("3 or 5")
LUAFend()

LUAFstart(glEvalPoint)
	LUAFvarcase(1, glEvalPoint1(LUAPi(1)))
	LUAFvarcase(2, glEvalPoint2(LUAPi(1), LUAPi(2)))
	LUAFvarendcase("1 or 2")
LUAFend()

LUAF			(glEnd)
LUAF			(glEndList)

static vector<GLfloat> feedbackBufferContents;
LUAFstart(glFeedbackBuffer)
	if (argc != 2) luaL_error(L, "expected 2 args");
	int size = LUAPi(1);
	int type = LUAPi(2);
	feedbackBufferContents.resize(size);
	glFeedbackBuffer(size, type, &(feedbackBufferContents[0]));
LUAFend()

LUAF			(glFinish)
LUAF6d			(glFrustum)
LUAF			(glFlush)

LUAFstart(glFog)
	LUAFbeginTableLoopCase()
		LUAFloopCaseBegin()
		LUAFloopCaseString()
			if (!strcasecmp(name, "mode"))			glFogi(GL_FOG_MODE, (int)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "density"))	glFogf(GL_FOG_DENSITY, (float)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "start"))	glFogf(GL_FOG_START, (float)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "end"))		glFogf(GL_FOG_END, (float)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "index"))	glFogi(GL_FOG_INDEX, (int)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "color"))	glFogfv(GL_FOG_COLOR, luaReadVec4f(L));
			else luaL_error(L, "found an unknown fog param: %s", safestr(name));
		LUAFloopCaseEnd()
	LUAFendTableLoopCase()
	if (argc == 2) {
		if (lua_istable(L, 2)) {
			int len = lua_objlen(L, 2);
			if (len == 4) glFogfv(LUAPi(1), luaReadVec4f(L));
			else luaL_error(L, "param got an unknown sized table, size %d", len);
		} else {
			glFogf(LUAPi(1), LUAPf(2));
		}
	} else
	LUAFvarcase(5, glFogfv(LUAPi(1), vec4f(LUAPf(2), LUAPf(3), LUAPf(4), LUAPf(5))))
	LUAFvarendcase("table, 2, or 5")
LUAFend()

LUAFi			(glFrontFace)

LUAFstart(glGenList)
	int lists = (argc == 0) ? 1 : LUAPi(1);
	lua_pushinteger(L, glGenLists(lists));
	return 1;
LUAFend()

LUAFuip			(glGenTextures)

LUAFstart(glGet)
	if (argc != 1) luaL_error(L, "expected one parameter");
	int param = (int)lua_tonumber(L,1);
	switch (param) {
	//bool
	case GL_ALPHA_TEST:
	case GL_AUTO_NORMAL:
	case GL_BLEND:
	case GL_CLIP_PLANE0:
	case GL_CLIP_PLANE1:
	case GL_CLIP_PLANE2:
	case GL_CLIP_PLANE3:
	case GL_CLIP_PLANE4:
	case GL_CLIP_PLANE5:
	case GL_COLOR_ARRAY:
	case GL_COLOR_LOGIC_OP:
	case GL_COLOR_MATERIAL:
	case GL_COLOR_TABLE:
	case GL_CULL_FACE:
	case GL_CURRENT_RASTER_POSITION_VALID:
	case GL_DEPTH_TEST:
	case GL_DEPTH_WRITEMASK:
	case GL_DITHER:
	case GL_DOUBLEBUFFER:
	case GL_EDGE_FLAG:
	case GL_EDGE_FLAG_ARRAY:
	case GL_FOG:
	case GL_INDEX_ARRAY:
	case GL_INDEX_LOGIC_OP:
	case GL_INDEX_MODE:
	case GL_LIGHTING:
	case GL_LIGHT_MODEL_LOCAL_VIEWER:
	case GL_LIGHT_MODEL_TWO_SIDE:
	case GL_LINE_SMOOTH:
	case GL_LINE_STIPPLE:
	case GL_MAP1_COLOR_4:
	case GL_MAP1_NORMAL:
	case GL_MAP1_TEXTURE_COORD_1:
	case GL_MAP1_TEXTURE_COORD_2:
	case GL_MAP1_TEXTURE_COORD_3:
	case GL_MAP1_TEXTURE_COORD_4:
	case GL_MAP1_VERTEX_3:
	case GL_MAP1_VERTEX_4:
	case GL_MAP2_COLOR_4:
	case GL_MAP2_NORMAL:
	case GL_MAP2_TEXTURE_COORD_1:
	case GL_MAP2_TEXTURE_COORD_2:
	case GL_MAP2_TEXTURE_COORD_3:
	case GL_MAP2_TEXTURE_COORD_4:
	case GL_MAP2_VERTEX_3:
	case GL_MAP2_VERTEX_4:
	case GL_MAP_COLOR:
	case GL_MAP_STENCIL:
	case GL_NORMAL_ARRAY:
	case GL_NORMALIZE:
	case GL_PACK_LSB_FIRST:
	case GL_POINT_SMOOTH:
	case GL_POLYGON_OFFSET_FILL:
	case GL_POLYGON_OFFSET_LINE:
	case GL_POLYGON_OFFSET_POINT:
	case GL_POLYGON_SMOOTH:
	case GL_POLYGON_STIPPLE:
	case GL_RGBA_MODE:
	case GL_SCISSOR_TEST:
	case GL_STENCIL_TEST:
	case GL_STEREO:
	case GL_TEXTURE_1D:
	case GL_TEXTURE_2D:
	case GL_TEXTURE_3D:
	case GL_TEXTURE_COORD_ARRAY:
	case GL_TEXTURE_CUBE_MAP:
	case GL_TEXTURE_GEN_Q:
	case GL_TEXTURE_GEN_R:
	case GL_TEXTURE_GEN_S:
	case GL_TEXTURE_GEN_T:
	case GL_UNPACK_LSB_FIRST:
	case GL_UNPACK_SWAP_BYTES:
	case GL_VERTEX_ARRAY:
		{
			GLboolean v = GL_FALSE;
			glGetBooleanv(param, &v);
			lua_pushboolean(L,v);
			return 1;
		}
		break;
	//int
	case GL_ACCUM_ALPHA_BITS:
	case GL_ACCUM_BLUE_BITS:
	case GL_ACCUM_GREEN_BITS:
	case GL_ACCUM_RED_BITS:
	case GL_ACTIVE_TEXTURE:
	case GL_ALPHA_BITS:
	case GL_ALPHA_TEST_FUNC:
	case GL_ATTRIB_STACK_DEPTH:
	case GL_AUX_BUFFERS:
	case GL_BLEND_DST:
	case GL_BLEND_SRC:
	case GL_BLUE_BITS:
	case GL_CLIENT_ATTRIB_STACK_DEPTH:
	case GL_COLOR_ARRAY_SIZE:
	case GL_COLOR_ARRAY_STRIDE:
	case GL_COLOR_ARRAY_TYPE:
	case GL_COLOR_MATERIAL_FACE:
	case GL_COLOR_MATERIAL_PARAMETER:	
	case GL_CULL_FACE_MODE:
	case GL_DEPTH_BITS:
	case GL_DEPTH_FUNC:
	case GL_DRAW_BUFFER:
	case GL_EDGE_FLAG_ARRAY_STRIDE:
	case GL_FOG_HINT:
	case GL_FOG_INDEX:
	case GL_FOG_MODE:
	case GL_FRONT_FACE:
	case GL_GREEN_BITS:
	case GL_INDEX_ARRAY_STRIDE:
	case GL_INDEX_ARRAY_TYPE:
	case GL_INDEX_BITS:
	case GL_INDEX_CLEAR_VALUE:
	case GL_INDEX_OFFSET:
	case GL_INDEX_SHIFT:
	case GL_INDEX_WRITEMASK:
	case GL_LINE_SMOOTH_HINT:
	case GL_LINE_STIPPLE_PATTERN:
	case GL_LINE_STIPPLE_REPEAT:
	case GL_LIST_BASE:
	case GL_LIST_INDEX:
	case GL_LIST_MODE:
	case GL_LOGIC_OP_MODE:
	case GL_MAP1_GRID_SEGMENTS:
	case GL_MAP1_INDEX:
	case GL_MAP2_GRID_SEGMENTS:
	case GL_MAP2_INDEX:
	case GL_MATRIX_MODE:
	case GL_MAX_CLIENT_ATTRIB_STACK_DEPTH:
	case GL_MAX_ATTRIB_STACK_DEPTH:
	case GL_MAX_CLIP_PLANES:
	case GL_MAX_EVAL_ORDER:
	case GL_MAX_LIGHTS:
	case GL_MAX_LIST_NESTING:
	case GL_MAX_MODELVIEW_STACK_DEPTH:
	case GL_MAX_NAME_STACK_DEPTH:
	case GL_MAX_PIXEL_MAP_TABLE:
	case GL_MAX_PROJECTION_STACK_DEPTH:
	case GL_MAX_TEXTURE_SIZE:
	case GL_MAX_TEXTURE_STACK_DEPTH:
	case GL_MAX_VIEWPORT_DIMS:
	case GL_MODELVIEW_STACK_DEPTH:
	case GL_NAME_STACK_DEPTH:
	case GL_NORMAL_ARRAY_STRIDE:
	case GL_NORMAL_ARRAY_TYPE:
	case GL_PACK_ALIGNMENT:
	case GL_PACK_ROW_LENGTH:
	case GL_PACK_SKIP_PIXELS:
	case GL_PACK_SKIP_ROWS:
	case GL_PACK_SWAP_BYTES:
	case GL_PERSPECTIVE_CORRECTION_HINT:
	case GL_PIXEL_MAP_A_TO_A_SIZE:
	case GL_PIXEL_MAP_B_TO_B_SIZE:
	case GL_PIXEL_MAP_G_TO_G_SIZE:
	case GL_PIXEL_MAP_I_TO_A_SIZE:
	case GL_PIXEL_MAP_I_TO_B_SIZE:
	case GL_PIXEL_MAP_I_TO_G_SIZE:
	case GL_PIXEL_MAP_I_TO_I_SIZE:
	case GL_PIXEL_MAP_I_TO_R_SIZE:
	case GL_PIXEL_MAP_R_TO_R_SIZE:
	case GL_PIXEL_MAP_S_TO_S_SIZE:
	case GL_POINT_SMOOTH_HINT:
	case GL_POLYGON_SMOOTH_HINT:
	case GL_PROJECTION_STACK_DEPTH:
	case GL_READ_BUFFER:
	case GL_RED_BITS:
	case GL_RENDER_MODE:
	case GL_SHADE_MODEL:
	case GL_STENCIL_BITS:
	case GL_STENCIL_CLEAR_VALUE:
	case GL_STENCIL_FAIL:
	case GL_STENCIL_FUNC:
	case GL_STENCIL_PASS_DEPTH_FAIL:
	case GL_STENCIL_PASS_DEPTH_PASS:
	case GL_STENCIL_REF:
	case GL_STENCIL_VALUE_MASK:
	case GL_STENCIL_WRITEMASK:
	case GL_SUBPIXEL_BITS:
	case GL_TEXTURE_BINDING_1D:
	case GL_TEXTURE_BINDING_2D:
	case GL_TEXTURE_BINDING_CUBE_MAP:
	case GL_TEXTURE_BINDING_3D:
	case GL_TEXTURE_COORD_ARRAY_SIZE:
	case GL_TEXTURE_COORD_ARRAY_STRIDE:
	case GL_TEXTURE_COORD_ARRAY_TYPE:
	case GL_TEXTURE_STACK_DEPTH:
	case GL_UNPACK_ALIGNMENT:
	case GL_UNPACK_ROW_LENGTH:
	case GL_UNPACK_SKIP_PIXELS:
	case GL_UNPACK_SKIP_ROWS:
	case GL_VERTEX_ARRAY_SIZE:
	case GL_VERTEX_ARRAY_STRIDE:
	case GL_VERTEX_ARRAY_TYPE:
		//multitexture
	case GL_MAX_TEXTURE_UNITS:
	case GL_TEXTURE0:
	case GL_TEXTURE1:
	case GL_TEXTURE2:
	case GL_TEXTURE3:
	case GL_TEXTURE4:
	case GL_TEXTURE5:
	case GL_TEXTURE6:
	case GL_TEXTURE7:
	case GL_TEXTURE8:
	case GL_TEXTURE9:
	case GL_TEXTURE10:
	case GL_TEXTURE11:
	case GL_TEXTURE12:
	case GL_TEXTURE13:
	case GL_TEXTURE14:
	case GL_TEXTURE15:
	case GL_TEXTURE16:
	case GL_TEXTURE17:
	case GL_TEXTURE18:
	case GL_TEXTURE19:
	case GL_TEXTURE20:
	case GL_TEXTURE21:
	case GL_TEXTURE22:
	case GL_TEXTURE23:
	case GL_TEXTURE24:
	case GL_TEXTURE25:
	case GL_TEXTURE26:
	case GL_TEXTURE27:
	case GL_TEXTURE28:
	case GL_TEXTURE29:
	case GL_TEXTURE30:
	case GL_TEXTURE31:
		//GL_EXT_draw_range_elements
	case GL_MAX_ELEMENTS_VERTICES:
	case GL_MAX_ELEMENTS_INDICES:
		//GL_EXT_framebuffer_object
	case GL_FRAMEBUFFER_BINDING_EXT:
	case GL_RENDERBUFFER_BINDING_EXT:
	case GL_MAX_COLOR_ATTACHMENTS_EXT:
	case GL_MAX_RENDERBUFFER_SIZE_EXT:
		{
			GLint v = 0;
			glGetIntegerv(param, &v);
			lua_pushinteger(L,v);
			return 1;
		}
		break;
	//double
	case GL_ALPHA_SCALE:
	case GL_ALPHA_TEST_REF:
	case GL_ALPHA_BIAS:
	case GL_BLUE_BIAS:
	case GL_BLUE_SCALE:
	case GL_CURRENT_INDEX:
	case GL_CURRENT_RASTER_DISTANCE:
	case GL_CURRENT_RASTER_INDEX:
	case GL_DEPTH_BIAS:
	case GL_DEPTH_CLEAR_VALUE:
	case GL_DEPTH_SCALE:
	case GL_FOG_DENSITY:
	case GL_FOG_END:
	case GL_FOG_START:
	case GL_GREEN_BIAS:
	case GL_GREEN_SCALE:
	case GL_LINE_WIDTH:
	case GL_LINE_WIDTH_GRANULARITY:
	case GL_POINT_SIZE:
	case GL_POINT_SIZE_GRANULARITY:
	case GL_POLYGON_OFFSET_FACTOR:
	case GL_POLYGON_OFFSET_UNITS:
	case GL_RED_BIAS:
	case GL_RED_SCALE:
	case GL_ZOOM_X:
	case GL_ZOOM_Y:
		{
			GLdouble v = 0;
			glGetDoublev(param, &v);
			lua_pushnumber(L,v);
			return 1;
		}
		break;
	//int[2]
	case GL_POLYGON_MODE:
		{
			GLint v[2];
			glGetIntegerv(param, v);
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 2; i++) {
				lua_pushinteger(L, i+1);
				lua_pushinteger(L, v[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	//double[2]
	case GL_DEPTH_RANGE:
	case GL_LINE_WIDTH_RANGE:
	case GL_MAP1_GRID_DOMAIN:
	case GL_MAP2_GRID_DOMAIN:
	case GL_POINT_SIZE_RANGE:
		{
			GLdouble v[2];
			glGetDoublev(param, v);
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 2; i++) {
				lua_pushinteger(L, i+1);
				lua_pushnumber(L, v[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	//double[3]
	case GL_CURRENT_NORMAL:
		{
			GLdouble v[3];
			glGetDoublev(param, v);
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 3; i++) {
				lua_pushinteger(L, i+1);
				lua_pushnumber(L, v[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	//bool[4]
	case GL_COLOR_WRITEMASK:
		{
			GLboolean v[4];
			glGetBooleanv(param, v);
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 4; i++) {
				lua_pushinteger(L, i+1);
				lua_pushboolean(L, v[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	//int[4]
	case GL_VIEWPORT:
	case GL_SCISSOR_BOX:
		{
			GLint v[4];
			glGetIntegerv(param, v);
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 4; i++) {
				lua_pushinteger(L, i+1);
				lua_pushinteger(L, v[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	//double[4]
	case GL_ACCUM_CLEAR_VALUE:
	case GL_COLOR_CLEAR_VALUE:
	case GL_CURRENT_COLOR:
	case GL_CURRENT_RASTER_COLOR:
	case GL_CURRENT_RASTER_POSITION:
	case GL_CURRENT_RASTER_TEXTURE_COORDS:
	case GL_CURRENT_TEXTURE_COORDS:
	case GL_FOG_COLOR:
	case GL_LIGHT_MODEL_AMBIENT:
		{
			GLdouble v[4];
			glGetDoublev(param, v);
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 4; i++) {
				lua_pushinteger(L, i+1);
				lua_pushnumber(L, v[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	//double[16]
	case GL_COLOR_MATRIX:
	case GL_MODELVIEW_MATRIX:
	case GL_PROJECTION_MATRIX:
	case GL_TEXTURE_MATRIX:
	case GL_TRANSPOSE_COLOR_MATRIX:
	case GL_TRANSPOSE_MODELVIEW_MATRIX:
	case GL_TRANSPOSE_PROJECTION_MATRIX:
	case GL_TRANSPOSE_TEXTURE_MATRIX:
		{
			double matrix[16];
			glGetDoublev(param, matrix);
			//now to put it back in OpenGL
			lua_newtable(L);
			int table = lua_gettop(L);
			for (int i = 0; i < 16; i++) {
				lua_pushinteger(L, i+1);
				lua_pushnumber(L, matrix[i]);
				lua_settable(L,table);
			}
			return 1;
		}
		break;
	}
	
	GLint maxLights = 8;
	glGetIntegerv(GL_MAX_LIGHTS, &maxLights);
	if (param >= GL_LIGHT0 && param < GL_LIGHT0 + maxLights) {
		GLboolean v = GL_FALSE;
		glGetBooleanv(param, &v);
		lua_pushboolean(L,v);
		return 1;
	}
LUAFend()

LUAFstart(glGetClipPlane)
	double eqn[4];
	glGetClipPlane(LUAPi(1), eqn);
	lua_newtable(L);
	int table = lua_gettop(L);
	for (int i = 0; i < 4; i++) {
		lua_pushinteger(L, i+1);
		lua_pushnumber(L, eqn[i]);
		lua_settable(L, table);
	}
	return 1;
LUAFend()

LUAFri			(glGetError)

LUAFstart(glGetPointer)
	void *ptr = NULL;
	glGetPointerv(LUAPi(1), &ptr);
	//lightuserdata or int?
	//lightuserdata preserves pointer resolution (on any system where it'd take up more bits than an int)
	//but int would be more reusable with other functions (glReadPixels when a framebuffer is bound)
	lua_pushlightuserdata(L, ptr);
	//stick with the highest resolution
	//somehow provide conversion inside lua if you really want it
	//or change functions like glReadPixels to resolve both
	return 1;
LUAFend()

LUAFstart(glGetPolygonStipple)
	int mask[32];
	glGetPolygonStipple((GLubyte*)mask);
	lua_newtable(L);
	int table = lua_gettop(L);
	for (int i = 0; i < 32; i++) {
		lua_pushinteger(L,i+1);
		lua_pushinteger(L,mask[i]);
		lua_settable(L, table);
	}
	return 1;
LUAFend()

LUAFstart(glGetString)
	lua_pushstring(L,(const char *)glGetString(LUAPi(1)));
	return 1;
LUAFend()

LUAFstart(glGetTexImage)
	if (argc < 2) luaL_error(L, "expected at least two parameters");
	
	int err = glGetError();
	if (err) luaL_error(L, "glGetError entering function is %d", err);
	
	int target = LUAPi(1);
	int level = LUAPi(2);
	//default for lua-gen'd
	int format = GL_RGBA;//LUAPi(3)
	int type = GL_FLOAT;//LUAPi(4)
	
	//explicit buffer means vbo
	bool usevbo = argc >= 5;

	float *buffer = NULL;
	int buffersize = 0;	//only used internally with non-vbo
	if (usevbo) {
		format = LUAPi(3);
		type = LUAPi(4);
		buffer = (float*)((char*)NULL + LUAPi(5));
	} else {		
		GLint width;
		glGetTexLevelParameteriv(target, level, GL_TEXTURE_WIDTH, &width);
		if (err = glGetError()) luaL_error(L, "glGetTexLevelParameteriv width raised gl error %d", err);

		GLint height;
		glGetTexLevelParameteriv(target, level, GL_TEXTURE_HEIGHT, &height);
		if (err = glGetError()) luaL_error(L, "glGetTexLevelParameteriv height raised gl error %d", err);

		int components = 4;
		//glGetTexLevelParameteriv(target, level, GL_TEXTURE_COMPONENTS, &components);

		buffersize = width*height*components;
		buffer = new float[buffersize];
		if (!buffer) luaL_error(L, "failed to alloc buffer for size=%d: w=%d h=%d components=%d", buffersize, width, height), components;
	}
	
	//gl error in here ...
	glGetTexImage(target, level, format, type, buffer);
	
	if (!usevbo) {
		if (err = glGetError()) {
			delete[] buffer;
			luaL_error(L, "glGetTexImage raised gl error %d", err);
		}

		lua_newtable(L);
		int table = lua_gettop(L);
		for (int i = 0; i < buffersize; i++) {
			lua_pushinteger(L, i+1);
			lua_pushnumber(L, buffer[i]);
			lua_settable(L, table);
		}
		
		delete[] buffer;
		
		return 1;
	}
LUAFend()

LUAFstart(glGetTexLevelParameter)
	GLint ret = 0;
	glGetTexLevelParameteriv(LUAPi(1), LUAPi(2), LUAPi(3), &ret);
	lua_pushinteger(L, ret);
	return 1;
LUAFend()

LUAF2i			(glHint)
LUAFi			(glIndexMask)

//not mt safe
LUAFstart(glIndexPointer)
	if (argc == 1) {
		static vector<int> buffer;
		buffer.resize(0);

		int t = lua_gettop(L);
		if (!lua_istable(L,t)) luaL_error(L, "expected a table as the single param");

		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			buffer.push_back((int)lua_tointeger(L,-1));
			lua_pop(L,1);
		}
		
		glIndexPointer(GL_INT, 0, &(buffer[0]));
		
	} else if (argc == 3) {
		glIndexPointer(LUAPi(1), LUAPi(2), (char*)NULL+LUAPi(3));
	} else {
		luaL_error(L, "expected either 1 or 3 args");
	}
LUAFend()

LUAFd			(glIndexd)
LUAF			(glInitNames)
LUAFirb			(glIsEnabled)
LUAFirb			(glIsList)
LUAFirb			(glIsTexture)

#define LUAFloop_glLightf(str, gl)	} else if (!strcasecmp(name, #str)) { glLightf(index, gl, lua_tonumber(L,-1));
#define LUAFloop_glLight4f(str, gl)	} else if (!strcasecmp(name, #str)) { glLightfv(index, gl, luaReadVec4f(L));
#define LUAFloop_glLight3f(str, gl)	} else if (!strcasecmp(name, #str)) { glLightfv(index, gl, quatf(luaReadVec3f(L), 0.f));

static int glLight_index(lua_State *L, int index) {
	LUAFbeginTableLoop()
		LUAFloopCaseBegin()
		LUAFloopCaseString()
			LUAFloopCaseBegin()
			LUAFloop_glLightf(spotExponent, GL_SPOT_EXPONENT)
			LUAFloop_glLightf(spotCutoff, GL_SPOT_CUTOFF)
			LUAFloop_glLightf(constantAttenuation, GL_CONSTANT_ATTENUATION)
			LUAFloop_glLightf(linearAttenuation, GL_LINEAR_ATTENUATION)
			LUAFloop_glLightf(quadraticAttenuation, GL_QUADRATIC_ATTENUATION)
			LUAFloop_glLight4f(ambient, GL_AMBIENT)
			LUAFloop_glLight4f(diffuse, GL_DIFFUSE)
			LUAFloop_glLight4f(specular, GL_SPECULAR)
			LUAFloop_glLight4f(position, GL_POSITION)
			LUAFloop_glLight3f(spotDirection, GL_SPOT_DIRECTION)
			} else if (!strcasecmp(name, "enable")) { (lua_toboolean(L,-1) ? glEnable : glDisable)(index);
			LUAFloopCaseStringEnd()
		LUAFloopCaseEnd()
	LUAFendTableLoop()
	return 0;
}

LUAFstart(glLight)
	LUAFbeginTableLoopCase()
		LUAFloopCaseBegin()
		LUAFloopCaseNumber()
			glLight_index(L, GL_LIGHT0+(int)num);
		LUAFloopCaseEnd()
	LUAFendTableLoopCase()
	LUAFvarcase(3, glLightf(LUAPi(1), LUAPi(2), LUAPf(3)))
	LUAFvarcase(5, glLightfv(LUAPi(1), LUAPi(2), vec4f(LUAPf(3), LUAPf(4), LUAPf(5), 0.f)))
	LUAFvarcase(6, glLightfv(LUAPi(1), LUAPi(2), vec4f(LUAPf(3), LUAPf(4), LUAPf(5), LUAPf(6))))
	LUAFvarendcase("table, 3, 5, or 6")
LUAFend()

LUAFstart(glLightModel)
	LUAFbeginTableLoopCase()
		LUAFloopCaseBegin()
		LUAFloopCaseString()
			if (!strcasecmp(name, "ambient"))			glLightModelfv(GL_LIGHT_MODEL_AMBIENT, luaReadVec4f(L));
			else if (!strcasecmp(name, "localViewer"))	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, (float)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "twoSide"))		glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, (float)lua_tonumber(L,-1));
			else if (!strcasecmp(name, "colorControl"))	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, (float)lua_tonumber(L,-1));
			else luaL_error(L, "found an unknown material param: %s", safestr(name));
		LUAFloopCaseEnd()
	LUAFendTableLoopCase()
	LUAFvarcase(2, glLightModelf(LUAPi(1), LUAPf(2)))
	LUAFvarcase(5, glLightModelfv(LUAPi(1), vec4f(LUAPf(2), LUAPf(3), LUAPf(4), LUAPf(5))))
	LUAFvarendcase("table, 3, or 5")
LUAFend()

LUAF2i			(glLineStipple)
LUAFd			(glLineWidth)
LUAFi			(glListBase)
LUAF			(glLoadIdentity)

LUAFstart(glLoadMatrix)
	double matrix[16];
	for (int i = 0; i < 16; i++) {
		lua_pushinteger(L,i+1);
		lua_gettable(L,1);
		matrix[i] = lua_tonumber(L,-1);
		lua_pop(L,1);
	}
	glLoadMatrixd(matrix);
LUAFend()

LUAFi			(glLoadName)
LUAFi			(glLogicOp)

#define LUAFloop_glMaterialf(str, gl)	} else if (!strcasecmp(name, #str)) { glMaterialf(face, gl, lua_tonumber(L,-1));
#define LUAFloop_glMaterial4f(str, gl)	} else if (!strcasecmp(name, #str)) { glMaterialfv(face, gl, luaReadVec4f(L));
#define LUAFloop_glMaterial3f(str, gl)	} else if (!strcasecmp(name, #str)) { glMaterialfv(face, gl, quatf(luaReadVec3f(L), 0.f));

static int glMaterial_face(lua_State *L, int face) {
	LUAFbeginTableLoop()
		LUAFloopCaseBegin()
		LUAFloopCaseString()
			LUAFloopCaseBegin()
			LUAFloop_glMaterial4f(ambient, GL_AMBIENT)
			LUAFloop_glMaterial4f(diffuse, GL_DIFFUSE)
			LUAFloop_glMaterial4f(specular, GL_SPECULAR)
			LUAFloop_glMaterial4f(emission, GL_EMISSION)
			LUAFloop_glMaterialf(shininess, GL_SHININESS)
			LUAFloop_glMaterial4f(ambientAndDiffuse, GL_AMBIENT_AND_DIFFUSE)
			LUAFloop_glMaterial3f(colorIndexes, GL_COLOR_INDEXES)
			LUAFloopCaseStringEnd()
		LUAFloopCaseEnd()
	LUAFendTableLoop()
	return 0;
}

LUAFstart(glMaterial)
	LUAFbeginTableLoopCase()
		LUAFloopCaseBegin()
		LUAFloopCaseString()
			if (!strcasecmp(name, "front")) glMaterial_face(L, GL_FRONT);
			else if (!strcasecmp(name, "back")) glMaterial_face(L, GL_BACK);
			else if (!strcasecmp(name, "frontAndBack")) glMaterial_face(L, GL_FRONT_AND_BACK);
			else luaL_error(L, "found an unknown material face string param: %s", safestr(name));
		LUAFloopCaseNumber()
			if (num == 1) glMaterial_face(L, GL_FRONT);
			else if (num == 2) glMaterial_face(L, GL_BACK);
			else luaL_error(L, "found an unknown material face number param: %d", num);
		LUAFloopCaseEnd()
	LUAFendTableLoopCase()
	LUAFvarcase(3, glMaterialf(LUAPi(1), LUAPi(2), LUAPf(3)))
	LUAFvarcase(5, glMaterialfv(LUAPi(1), LUAPi(2), vec4f(LUAPf(3),LUAPf(4),LUAPf(5), 0.f)) )	//GL_COLOR_INDEXES only
	LUAFvarcase(6, glMaterialfv(LUAPi(1), LUAPi(2), vec4f(LUAPf(3),LUAPf(4),LUAPf(5),LUAPf(6))) )	//all else
	LUAFvarendcase("table, 4 or 6")
LUAFend()

LUAFi(glMatrixMode)

LUAFstart(glMultiTexCoord1dvARB)
	if (argc == 2) {
		if (lua_istable(L,2)) {
			int len = lua_objlen(L, 2);
			if (len == 1) glMultiTexCoord1dvARB( LUAPi(1), luaReadVec1d(L) );
			if (len == 2) glMultiTexCoord2dvARB( LUAPi(1), luaReadVec2d(L) );
			if (len == 3) glMultiTexCoord3dvARB( LUAPi(1), luaReadVec3d(L) );
			if (len == 4) glMultiTexCoord4dvARB( LUAPi(1), luaReadVec4d(L) );
			else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
		} else {
			glMultiTexCoord1dARB(LUAPi(1), LUAPd(2) );
		}
	} else 
	LUAFvarcase(3, glMultiTexCoord2dARB(LUAPi(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarcase(4, glMultiTexCoord3dARB(LUAPi(1), LUAPd(2), LUAPd(3), LUAPd(4) ) )
	LUAFvarcase(5, glMultiTexCoord4dARB(LUAPi(1), LUAPd(2), LUAPd(3), LUAPd(4), LUAPd(5) ) )
	LUAFvarendcase("1 and table, 2, 3, 4, or 5")
LUAFend()

LUAFstart(glMultMatrix)
	double matrix[16];
	for (int i = 0; i < 16; i++) {
		lua_pushinteger(L,i+1);
		lua_gettable(L,1);
		matrix[i] = lua_tonumber(L,-1);
		lua_pop(L,1);
	}
	glMultMatrixd(matrix);
LUAFend()

LUAF2i			(glNewList)

LUAFstart(glNormal)
	if (argc == 1) {
		if (!lua_istable(L,1)) {
			luaL_error(L, "expected a table");
			return 0;
		}
		int len = lua_objlen(L, 1);
		if (len == 3) glNormal3dv(luaReadVec3d(L));
		else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
	} else 
	LUAFvarcase(3, glNormal3d( LUAPd(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarendcase("table or 3")
LUAFend()

//not mt safe
LUAFstart(glNormalPointer)
	if (argc == 1) {
		static vector<float> buffer;
		buffer.resize(0);

		int t = lua_gettop(L);
		if (!lua_istable(L,t)) luaL_error(L, "expected a table as the single param");

		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			//fprintf(stderr, "%s - %s\n", lua_typename(L, lua_type(L, -2)), lua_typename(L, lua_type(L, -1)));
				
			int t2 = lua_gettop(L);
			//TODO won't this leak if the key isnt a string?
			if (!lua_istable(L,t2)) luaL_error(L, "found vector elem %s that wasnt a table", lua_tostring(L,-2));

			int newlen = lua_objlen(L,t2);	//make sure the length of the subtables are all the same
			if (newlen != 3) luaL_error(L, "i was reading vecs of length 3 when i found one of differing length %d", newlen);

			lua_pushnil(L);
			while (lua_next(L,t2) != 0) {
				buffer.push_back((float)lua_tonumber(L,-1));
				lua_pop(L,1);
			}

			lua_pop(L,1);
		}
		
		//done!
		glNormalPointer(GL_FLOAT, 0, &(buffer[0]));
		
	} else if (argc == 3) {
		glNormalPointer(LUAPi(1), LUAPi(2), (char*)NULL+LUAPi(3));
	} else {
		luaL_error(L, "expected either 1 or 3 args");
	}
LUAFend()

LUAF6d			(glOrtho)
LUAFd			(glPassThrough)
LUAF1i1d		(glPixelStoref)
LUAF1i1d		(glPixelTransferf)
LUAF2d			(glPixelZoom)
LUAFd			(glPointSize)
LUAF2i			(glPolygonMode)
LUAF2d			(glPolygonOffset)

LUAFstart(glPolygonStipple)
	int mask[32];
	for (int i = 0; i < 32; i++) {
		lua_pushinteger(L,i+1);
		lua_gettable(L,1);
		mask[i] = lua_tointeger(L,-1);
		lua_pop(L,1);
	}
	glPolygonStipple((GLubyte*)mask);
LUAFend()

LUAF			(glPopAttrib)
LUAF			(glPopClientAttrib)
LUAF			(glPopMatrix)
LUAF			(glPopName)

LUAFstart(glPrioritizeTextures)
	size_t n = LUAPi(1);
	GLuint *textures = new GLuint[n];
	GLclampf *priorities = new GLclampf[n];
	for (int i = 0; i < n; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, 2);
		textures[i] = (GLuint)lua_tointeger(L, -1);
		lua_pop(L,1);
		
		lua_pushinteger(L, i+1);
		lua_gettable(L, 3);
		priorities[i] = (GLclampf)lua_tonumber(L, -1);
		lua_pop(L,1);
	}
	glPrioritizeTextures(n, textures, priorities);
	delete textures;
	delete priorities;
LUAFend()

LUAFi			(glPushAttrib)
LUAFi			(glPushClientAttrib)
LUAF			(glPushMatrix)
LUAFi			(glPushName)

LUAFstart(glRasterPos)
	if (argc == 1) {
		if (!lua_istable(L,1)) {
			luaL_error(L, "expected a table");
		}
		int len = lua_objlen(L, 1);
		if (len == 2) glRasterPos2dv( luaReadVec2d(L));
		else if (len == 3) glRasterPos3dv(luaReadVec3d(L));
		else if (len == 4) glRasterPos4dv( luaReadVec4d(L));
		else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
	} else 
	LUAFvarcase(2, glRasterPos2d(LUAPd(1), LUAPd(2) ) )
	LUAFvarcase(3, glRasterPos3d(LUAPd(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarcase(4, glRasterPos4d(LUAPd(1), LUAPd(2), LUAPd(3), LUAPd(4) ) )
	LUAFvarendcase("table, 2, 3, or 4")
LUAFend()

LUAFi			(glReadBuffer)

LUAFstart(glReadPixels)
	//if we have less than seven arguments then we should return the results as a lua array ...
	GLint x = LUAPi(1);
	GLint y = LUAPi(2);
	GLsizei width = LUAPi(3);
	GLsizei height = LUAPi(4);
	GLenum format = LUAPi(5);
	GLenum type = LUAPi(6);
	if (argc >= 7) {
		int ofs = LUAPi(7);
		GLvoid* buffer = (char*)NULL + ofs;
		glReadPixels(x,y,width,height,format,type,buffer);
		
#if 0	//see what we read
		//try to read back that data and see what you get...
		float *ptr = (float *)glMapBufferARB(GL_PIXEL_PACK_BUFFER_ARB, GL_READ_ONLY_ARB);
		GLint size = 0;
		glGetBufferParameterivARB(GL_PIXEL_PACK_BUFFER_ARB, GL_BUFFER_SIZE_ARB, &size);
		fprintf(stderr, "read size as %d\n", size);
		assert(ptr);
//		for (int i = 0; i < size/sizeof(float); i++) {
//			fprintf(stderr, "reading %d is %f\n", i, ptr[i]);
//		}
		ptr = (float*)((char*)ptr + ofs);
		fprintf(stderr, "reading offset %d as %f %f %f %f\n", ofs, ptr[0],ptr[1],ptr[2],ptr[3]);
		glUnmapBufferARB(GL_PIXEL_PACK_BUFFER_ARB);
#endif
		
	} else {
		int components = GL_Format_to_Components(L, format);
		type = GL_FLOAT;

		int size = width*height*components;
		float *buffer = new float[size];
		if (!buffer) luaL_error(L, "failed to alloc buffer for size=%d: w=%d h=%d components=%d", size, width, height, components);
		
		glReadPixels(x,y,width,height,format,type,buffer);
		
		lua_newtable(L);
		int table = lua_gettop(L);
		for (int i = 0; i < size; i++) {
			lua_pushinteger(L, i+1);
			lua_pushnumber(L, buffer[i]);
			lua_settable(L, table);
		}

		delete[] buffer;
		return 1;
	}
LUAFend()

LUAF4d			(glRectd)

//not mt safe
static vector<GLuint> selectBufferContents;
LUAFstart(glRenderMode)
	GLint oldMode = 0;
	glGetIntegerv(GL_RENDER_MODE, &oldMode);
	int ret = glRenderMode(LUAPi(1));
	lua_pushinteger(L, ret);
	if (!ret) return 1;
	
	/*
	bleh, the ret doesn't directly measure the length of the array to return
	it does if you pick through it
	this means i might as well return the whole array
	and that in itself means i might as well store it somewhere else rather than re-create it every time
	*/
	
	lua_newtable(L);
	int t = lua_gettop(L);

	switch (oldMode) {
	case GL_FEEDBACK:
		{
			for (int i = 0; i < feedbackBufferContents.size(); i++) {
				lua_pushinteger(L, i+1);
				lua_pushnumber(L, feedbackBufferContents[i]);
				lua_settable(L, t);
			}
		}
		break;
	case GL_SELECT: 
		{
			for (int i = 0; i < selectBufferContents.size(); i++) {
				lua_pushinteger(L, i+1);
				lua_pushinteger(L, selectBufferContents[i]);
				lua_settable(L, t);
			}
		}
		break;
	default:
		luaL_error(L, "got an unknown prior render mode type %d", oldMode);
	}
	
	return 2;
LUAFend()

LUAF4d			(glRotated)

LUAFstart(glScale)
	if (argc == 1) {
		if (!lua_istable(L,1)) {
			luaL_error(L, "expected a table");
		}
		int len = lua_objlen(L, 1);
		if (len == 3) {
			vec3d v = luaReadVec3d(L);
			glScaled(VEC3ELEM(v));
		} else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
	} else 
	LUAFvarcase(3, glScaled(LUAPd(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarendcase("table or 3")
LUAFend()

LUAF4i			(glScissor)

LUAFstart(glSelectBuffer)
	/*
	this will be a tough one
	because we need to pass an object to gl
	that gl fills upon a glRenderMode call
	and we can subsequently access...
	... my vote is have the user pass a size
	. then create a C int array of that size
	exposed via lua table w/userdata field
	with a metatable that allows [] and # operations
	... but what about iterating?
	
	we could always have glRenderMode return a table ...
	or glSelectBuffer pass a reference to it,
		and glRenderMode fill the ref when switching away from GL_SELECT...
	*/
	if (argc < 1) luaL_error(L, "expected param: size");
	int size = lua_tointeger(L, 1);
	selectBufferContents.resize(size);
	glSelectBuffer(selectBufferContents.size(), &selectBufferContents[0]);
LUAFend()

LUAFi			(glShadeModel)
LUAF3i			(glStencilFunc)
LUAFi			(glStencilMask)
LUAF3i			(glStencilOp)

LUAFstart(glTexCoord)
	if (argc == 1) {
		if (lua_istable(L,1)) {
			int len = lua_objlen(L, 1);
			if (len == 1) glTexCoord1dv( luaReadVec1d(L) );
			else if (len == 2) glTexCoord2dv( luaReadVec2d(L) );
			else if (len == 3) glTexCoord3dv( luaReadVec3d(L) );
			else if (len == 4) glTexCoord4dv( luaReadVec4d(L) );
			else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
		} else {
			glTexCoord1d(LUAPd(1));
		}
	} else 
	LUAFvarcase(2, glTexCoord2d(LUAPd(1), LUAPd(2) ) )
	LUAFvarcase(3, glTexCoord3d(LUAPd(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarcase(4, glTexCoord4d(LUAPd(1), LUAPd(2), LUAPd(3), LUAPd(4) ) )
	LUAFvarendcase("table, 1, 2, 3, or 4")
LUAFend()

//not mt safe.
LUAFstart(glTexCoordPointer)

	if (argc == 1) {
		static vector<float> buffer;
		buffer.resize(0);

		int t = lua_gettop(L);
		if (!lua_istable(L,t)) luaL_error(L, "expected a table as the single param");

		int plen = -1;	//persistant length of the member tables
		
		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			
			int t2 = lua_gettop(L);
			if (!lua_istable(L,t2)) luaL_error(L, "found vector elem %s that wasnt a table", lua_tostring(L,-2));

			int newlen = lua_objlen(L,t2);	//make sure the length of the subtables are all the same
			if (plen == -1) plen = newlen;	//save the first length as persistent
			if (plen != newlen) luaL_error(L, "i was reading vecs of length %d when i found one of differing length %d", plen, newlen);

			lua_pushnil(L);
			while (lua_next(L,t2) != 0) {
				buffer.push_back((float)lua_tonumber(L,-1));
				lua_pop(L,1);
			}
			
			lua_pop(L,1);
		}
		
		//done!
		glTexCoordPointer(plen, GL_FLOAT, 0, &(buffer[0]));
		
	} else if (argc == 4) {
		glTexCoordPointer(LUAPi(1), LUAPi(2), LUAPi(3), (char*)NULL+LUAPi(4));
	} else {
		luaL_error(L, "expected either 1 or 4 args");
	}
LUAFend()

LUAF2ivar1f4f	(glTexEnv)
LUAF2ivar1i4d	(glTexGen)

//lua error on fail
typedef float buftype;
buftype *getTextureBuffer(lua_State *L, int stackpos, int width, int height, int depth, int format, GLenum *type) {
	buftype *buffer = NULL;
	if (lua_isfunction(L,stackpos)) {
		//buffer type and format override
		*type = GL_FLOAT;
		//todo - should we maybe consult this one over with the calling function?
		//if (format != GL_RGBA) format = GL_RGB;
		int components = GL_Format_to_Components(L, format);//(format == GL_RGBA) ? 4 : 3;
		buffer = new buftype[width * height * depth * components];
		assert(buffer);
		
		buftype *bp = buffer;
		for (int k = 0; k < depth; k++) {
			for (int j = 0; j < height; j++) {
				for (int i = 0; i < width; i++) {

					lua_pushvalue(L,stackpos);	//copy the function onto the stack
					//push the two variables
					lua_pushinteger(L,i);	//lua_pushnumber(L, ((double)i+.5)/(double)width);
					lua_pushinteger(L,j);	//lua_pushnumber(L, ((double)j+.5)/(double)height);
					lua_pushinteger(L,k);	//lua_pushnumber(L, ((double)k+.5)/(double)depth);
					//and run the callback
					if (!luaSafeCall(L, 3, components)) {
						delete[] buffer;
						luaL_error(L, "failed to run buffer pixel calblack");
					}
					for (int k = 0; k < components; k++ ) {
						*bp = (buftype)(lua_tonumber(L, k - components)); bp++;
					}
					lua_pop(L,components);
				}
			}
		}
	} else if (lua_istable(L, stackpos)) {
		//buffer type and format override
		*type = GL_FLOAT;
		//todo - should we maybe consult this one over with the calling function?
		//if (format != GL_RGBA) format = GL_RGB;
		int components = GL_Format_to_Components(L, format);//(format == GL_RGBA) ? 4 : 3;
		buffer = new buftype[width * height * depth * components];
		assert(buffer);
	
	
		//todo - table iterate across here and pick values into the buffer
		buftype *bp = buffer;

		lua_pushnil(L);  // first key
		while (lua_next(L, stackpos) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			*bp = lua_tonumber(L, -1);
			bp++;
			lua_pop(L,1);
			if (bp >= buffer + width * height * depth * components) break;
		}
		lua_pop(L,1);
	} else if (lua_isstring(L, stackpos)) {
		int bitsPerPixel = GL_Type_to_BPP(*type);
		int components = GL_Format_to_Components(L, format);
		int bufsize = (width * height * depth * bitsPerPixel * components + (sizeof(buftype)-1)) / sizeof(buftype);
		buffer = new buftype[bufsize];
		assert(buffer);

		size_t len = 0;
		const char *strbuf = lua_tolstring(L, stackpos, &len);
		
		int minsize = len < bufsize ? len : bufsize;
		memset(buffer, 0, bufsize);
		memcpy(buffer, strbuf, minsize);
	} else if (lua_isuserdata(L, stackpos)) {
		int bitsPerPixel = GL_Type_to_BPP(*type);
		int components = GL_Format_to_Components(L, format);
		int bufsize = (width * height * depth * bitsPerPixel * components + (sizeof(buftype)-1)) / sizeof(buftype);
		buffer = new buftype[bufsize];
		assert(buffer);
		
		void *data = lua_touserdata(L, stackpos);
		memcpy(buffer, data, bufsize);
	} else {
		//buffer type and format override
		*type = GL_FLOAT;
		//todo - should we maybe consult this one over with the calling function?
		//if (format != GL_RGBA) format = GL_RGB;
		int components = GL_Format_to_Components(L, format);//(format == GL_RGBA) ? 4 : 3;
		buffer = new buftype[width * height * depth * components];
		assert(buffer);

		memset(buffer, 0, sizeof(*buffer) * width * height * depth * components);
	}

	//I guess glTexSubImage2D can't accept null for its buffer?
	return buffer;
}

LUAFstart(glTexImage1D)
	//see if we have a generate function
	int target = LUAPi(1);
	int level = LUAPi(2);
	int internalFormat = LUAPi(3);
	int width = LUAPi(4);
	int border = LUAPi(5);
	int format = LUAPi(6);
	GLenum type = LUAPi(7);
	buftype *pixels = getTextureBuffer(L,8, width, 1, 1, format, &type);
	glTexImage1D(target, level, internalFormat, width, border, format, type, pixels);
	//cleanup
	if (pixels) delete[] pixels;
LUAFend()

LUAFstart(glTexImage2D)
	//see if we have a generate function
	int target = LUAPi(1);
	int level = LUAPi(2);
	int internalFormat = LUAPi(3);
	int width = LUAPi(4);
	int height = LUAPi(5);
	int border = LUAPi(6);
	int format = LUAPi(7);
	GLenum type = LUAPi(8);
	buftype *pixels = getTextureBuffer(L, 9, width, height, 1, format, &type);
	glTexImage2D(target, level, internalFormat, width, height, border, format, type, pixels);
	//cleanup
	if (pixels) delete[] pixels;
LUAFend()

#ifdef __APPLE__
LUAFstart(glTexImage3D)
#else
LUAFstart(glTexImage3DEXT)
#endif
	//see if we have a generate function
	int target = LUAPi(1);
	int level = LUAPi(2);
	int internalFormat = LUAPi(3);
	int width = LUAPi(4);
	int height = LUAPi(5);
	int depth = LUAPi(6);
	int border = LUAPi(7);
	int format = LUAPi(8);
	GLenum type = LUAPi(9);
	buftype *pixels = getTextureBuffer(L, 10, width, height, depth, format, &type);
//I have to do something about this...
#ifdef __APPLE__
	glTexImage3D(target, level, internalFormat, width, height, depth, border, format, type, pixels);
#else
	glTexImage3DEXT(target, level, internalFormat, width, height, depth, border, format, type, pixels);
#endif
	//cleanup
	if (pixels) delete[] pixels;
LUAFend()

LUAFstart(glTexSubImage1D)
	//see if we have a generate function
	int target = LUAPi(1);
	int level = LUAPi(2);
	int xofs = LUAPi(3);
	int width = LUAPi(4);
	int format = LUAPi(5);
	GLenum type = LUAPi(6);
	buftype *pixels = getTextureBuffer(L, 7, width, 1, 1, format, &type);
	glTexSubImage1D(target, level, xofs, width, format, type, pixels);
	//cleanup
	if (pixels) delete[] pixels;
LUAFend()

LUAFstart(glTexSubImage2D)
	//see if we have a generate function
	int target = LUAPi(1);
	int level = LUAPi(2);
	int xofs = LUAPi(3);
	int yofs = LUAPi(4);
	int width = LUAPi(5);
	int height = LUAPi(6);
	int format = LUAPi(7);
	GLenum type = LUAPi(8);
	buftype *pixels = getTextureBuffer(L, 9, width, height, 1, format, &type);
	glTexSubImage2D(target, level, xofs, yofs, width, height, format, type, pixels);
	//cleanup
	if (pixels) delete[] pixels;
LUAFend()

LUAF2ivar1f4f	(glTexParameter)

LUAFstart(glTranslate)
	if (argc == 1) {
		if (!lua_istable(L,1)) {
			luaL_error(L, "expected a table");
		}
		int len = lua_objlen(L, 1);
		if (len == 3) {
			vec3d v = luaReadVec3d(L);
			glTranslated(VEC3ELEM(v));
		} else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
	} else 
	LUAFvarcase(3, glTranslated(LUAPd(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarendcase("table or 3")
LUAFend()

LUAFstart(glVertex)
	if (argc == 1) {
		if (!lua_istable(L,1)) {
			luaL_error(L, "expected a table");
		}
		int len = lua_objlen(L, 1);
		if (len == 2) glVertex2dv(luaReadVec2d(L));
		else if (len == 3) glVertex3dv(luaReadVec3d(L));
		else if (len == 4) glVertex4dv( luaReadVec4d(L));
		else luaL_error(L, "single parameter got an unknown sized table, size %d", len);
	} else 
	LUAFvarcase(2, glVertex2d(LUAPd(1), LUAPd(2) ) )
	LUAFvarcase(3, glVertex3d(LUAPd(1), LUAPd(2), LUAPd(3) ) )
	LUAFvarcase(4, glVertex4d(LUAPd(1), LUAPd(2), LUAPd(3), LUAPd(4) ) )
	LUAFvarendcase("table, 2, 3, or 4")
LUAFend()

//not mt safe.
LUAFstart(glVertexPointer)
	if (argc == 1) {
		static vector<float> buffer;
		buffer.resize(0);

		int t = lua_gettop(L);
		if (!lua_istable(L,t)) luaL_error(L, "expected a table as the single param");

		int plen = -1;	//persistant length of the member tables
		
		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			
			int t2 = lua_gettop(L);
			//TODO memleak if key is not a string?
			if (!lua_istable(L,t2)) luaL_error(L, "found vector elem %s that wasnt a table", lua_tostring(L,-2));

			int newlen = lua_objlen(L,t2);	//make sure the length of the subtables are all the same
			if (plen == -1) plen = newlen;	//save the first length as persistent
			if (plen != newlen) luaL_error(L, "i was reading vecs of length %d when i found one of differing length ", plen, newlen);

			lua_pushnil(L);
			while (lua_next(L,t2) != 0) {
				buffer.push_back((float)lua_tonumber(L,-1));
				lua_pop(L,1);
			}
			
			lua_pop(L,1);
		}
		
		//done!
		glVertexPointer(plen, GL_FLOAT, 0, &(buffer[0]));
		
	} else if (argc == 4) {
		glVertexPointer(LUAPi(1), LUAPi(2), LUAPi(3), (char*)NULL+LUAPi(4));
	} else {
		luaL_error(L, "expected either 1 or 4 args");
	}
LUAFend()

LUAF4i			(glViewport)


//OpenGL 1.2 Imaging Subset

#if 0
LUAFstart(glColorTable)
	GLenum target = LUAPi(1);
	GLenum internalFormat = LUAPi(2);
	GLsizei width = LUAPi(3);
	GLenum format = LUAPi(4);
	GLenum type = LUAPi(5);
	buftype *table = getTextureBuffer(L, 6, width, 1, 1, format, &type);
	glColorTable(target, internalFormat, width, format, type, table);
	if (table) delete[] table;
LUAFend()
#endif

//TODO - split off opengl versions
//even if they're all put in the gl table
//at least keep them in separate arrays
static const luaL_reg luaBindings_GL[] = {
	//OpenGL 1.0 ... was there such a thing?  everything that doesn't have an associated extension I'm guessing

	//OpenGL 1.1
	LUA_GLUE(glAccum)
	LUA_GLUE(glAlphaFunc)
	//glAreTexturesResident			//EXT_texture_object
	LUA_GLUE(glArrayElement)		//EXT_vertex_array
	LUA_GLUE(glBegin)
	LUA_GLUE(glBindTexture)			//EXT_texture_object
	//LUA_GLUE(glBitmap)	//how should I represent bitmaps?  tables of bools? ints? strings of bytes?  strings of 1's and 0's?
	LUA_GLUE(glBlendFunc)
	LUA_GLUE(glCallList)
	LUA_GLUE(glClear)
	LUA_GLUE(glClearAccum)
	LUA_GLUE(glClearColor)
	LUA_GLUE(glClearDepth)
	LUA_GLUE(glClearIndex)
	LUA_GLUE(glClearStencil)
	LUA_GLUE(glClipPlane)
	LUA_GLUE(glColor)
	LUA_GLUE(glColorMask)
	LUA_GLUE(glColorMaterial)
	LUA_GLUE(glColorPointer)		//EXT_vertex_array, INTEL_parallel_arrays
	LUA_GLUE(glCopyPixels)
	LUA_GLUE(glCopyTexImage1D)		//EXT_copy_texture
	LUA_GLUE(glCopyTexImage2D)		//EXT_copy_texture
	LUA_GLUE(glCopyTexSubImage1D)	//EXT_copy_texture
	LUA_GLUE(glCopyTexSubImage2D)	//EXT_copy_texture
	LUA_GLUE(glCullFace)
	LUA_GLUE(glDeleteLists)
	LUA_GLUE(glDeleteTexture)		//EXT_texture_object
	LUA_GLUE(glDepthFunc)
	LUA_GLUE(glDepthMask)
	LUA_GLUE(glDepthRange)
	LUA_GLUE(glDisable)
	LUA_GLUE(glDisableClientState)
	LUA_GLUE(glDrawArrays)			//EXT_vertex_array
	LUA_GLUE(glDrawBuffer)
	LUA_GLUE(glDrawElements)
	//LUA_GLUE(glDrawPixels) someday ... could be useful for fbo/vbo/pbo stuff
	LUA_GLUE(glEdgeFlag)
	LUA_GLUE(glEdgeFlagPointer)		//EXT_vertex_array
	LUA_GLUE(glEnable)
	LUA_GLUE(glEnableClientState)
	LUA_GLUE(glEnd)
	LUA_GLUE(glEndList)
	LUA_GLUE(glEvalCoord)
	LUA_GLUE(glEvalMesh)
	LUA_GLUE(glEvalPoint)
	LUA_GLUE(glFeedbackBuffer)
	LUA_GLUE(glFinish)
	LUA_GLUE(glFlush)
	LUA_GLUE(glFog)
	LUA_GLUE(glFrontFace)
	LUA_GLUE(glFrustum)
	LUA_GLUE(glGenList)
	LUA_GLUEs(glGenTexture)			//EXT_texture_object
	LUA_GLUE(glGet)
	LUA_GLUE(glGetClipPlane)
	LUA_GLUE(glGetError)
	//LUA_GLUE(glGetLight)
	//LUA_GLUE(glGetMap)
	//LUA_GLUE(glGetMaterial)
	//LUA_GLUE(glGetPixelMap)
	LUA_GLUE(glGetPointer)			//EXT_vertex_array
	LUA_GLUE(glGetPolygonStipple)
	LUA_GLUE(glGetString)		// merge with glGet too?
	//LUA_GLUE(glGetTexEnv)
	//LUA_GLUE(glGetTexGen)
	LUA_GLUE(glGetTexImage)
	LUA_GLUE(glGetTexLevelParameter)
	//LUA_GLUE(glGetTexParameter)
	LUA_GLUE(glHint)				//PGI_misc_hints
	LUA_GLUE(glIndexMask)
	LUA_GLUE(glIndexPointer)		//EXT_vertex_array
	LUA_GLUEd(glIndex)
	LUA_GLUE(glInitNames)
	//glInterleavedArrays
	LUA_GLUE(glIsEnabled)			//EXT_texture_object
	LUA_GLUE(glIsList)
	LUA_GLUE(glIsTexture)
	LUA_GLUE(glLight)
	LUA_GLUE(glLightModel)
	LUA_GLUE(glLineStipple)
	LUA_GLUE(glLineWidth)
	LUA_GLUE(glListBase)
	LUA_GLUE(glLoadIdentity)
	LUA_GLUE(glLoadMatrix)
	LUA_GLUE(glLoadName)
	LUA_GLUE(glLogicOp)
	//LUA_GLUE(glMap)
	//LUA_GLUE(glMapGrid)
	LUA_GLUE(glMaterial)
	LUA_GLUE(glMatrixMode)
	LUA_GLUE(glMultMatrix)
	LUA_GLUE(glNewList)
	LUA_GLUE(glNormal)
	LUA_GLUE(glNormalPointer)		//EXT_vertex_array, INTEL_parallel_arrays
	LUA_GLUE(glOrtho)
	LUA_GLUE(glPassThrough)
	//LUA_GLUE(glPixelMap)
	LUA_GLUEf(glPixelStore)
	LUA_GLUEf(glPixelTransfer)
	LUA_GLUE(glPixelZoom)
	LUA_GLUE(glPointSize)
	LUA_GLUE(glPolygonMode)
	LUA_GLUE(glPolygonOffset)		//EXT_polygon_offset
	LUA_GLUE(glPolygonStipple)
	LUA_GLUE(glPopAttrib)
	LUA_GLUE(glPopClientAttrib)
	LUA_GLUE(glPopMatrix)
	LUA_GLUE(glPopName)
	LUA_GLUE(glPrioritizeTextures)	//EXT_texture_object
	LUA_GLUE(glPushAttrib)
	LUA_GLUE(glPushClientAttrib)
	LUA_GLUE(glPushMatrix)
	LUA_GLUE(glPushName)
	LUA_GLUE(glRasterPos)
	LUA_GLUE(glReadBuffer)
	LUA_GLUE(glReadPixels)
	LUA_GLUEd(glRect)
	LUA_GLUE(glRenderMode)
	LUA_GLUEd(glRotate)
	LUA_GLUE(glScale)
	LUA_GLUE(glScissor)
	LUA_GLUE(glSelectBuffer)
	LUA_GLUE(glShadeModel)
	LUA_GLUE(glStencilFunc)
	LUA_GLUE(glStencilMask)
	LUA_GLUE(glStencilOp)
	LUA_GLUE(glTexCoord)
	LUA_GLUE(glTexCoordPointer)		//EXT_vertex_array, INTEL_parallel_arrays
	LUA_GLUE(glTexEnv)
	LUA_GLUE(glTexGen)
	LUA_GLUE(glTexImage1D)
	LUA_GLUE(glTexImage2D)
	LUA_GLUE(glTexParameter)
	LUA_GLUE(glTexSubImage1D)		//EXT_subtexture
	LUA_GLUE(glTexSubImage2D)		//EXT_subtexture
	LUA_GLUE(glTranslate)
	LUA_GLUE(glVertex)
	LUA_GLUE(glVertexPointer)		//EXT_vertex_array, INTEL_parallel_arrays
	LUA_GLUE(glViewport)

	//OpenGL 1.2
	//glBlendColor					//EXT_blend_color
	//glBlendEquation				//EXT_blend_minmax
	//glDrawRangeElements			//EXT_draw_range_elements
	//LUA_GLUE(glTexImage3D)			//EXT_texture3D
	//glTexSubImage3D				//EXT_texture3D
	//glCopyTexSubImage3D			//EXT_copy_texture
	
	//OpenGL 1.2 Imaging Subset
	//glColorSubTable				//EXT_paletted_texture, EXT_color_subtable
	//LUA_GLUE(glColorTable)			//EXT_paletted_texture, SGI_color_table
	//glColorTableParameter			//SGI_color_table
	//glConvolutionFilter1D			//EXT_convolution
	//glConvolutionFilter2D			//EXT_convolution
	//glConvolutionParameter		//EXT_convolution
	//glCopyColorSubTable			//EXT_color_subtable
	//glCopyColorTable				//SGI_color_table
	//glCopyConvolutionFilter1D		//EXT_convolution
	//glCopyConvolutionFilter2D		//EXT_convolution
	//glGetColorTable				//EXT_paletted_texture, SGI_color_table
	//glGetColorTableParameter		//EXT_paletted_texture, SGI_color_table
	//glGetConvolutionFilter		//EXT_convolution
	//glGetConvolutionParameter		//EXT_convolution
	//glGetHistogram				//EXT_histogram
	//glGetHistogramParameter		//EXT_histogram
	//glGetMinmax					//EXT_histogram
	//glGetMinmaxParameter			//EXT_histogram
	//glGetSeparableFilter			//EXT_convolution
	//glHistogram					//EXT_histogram
	//glMinmax						//EXT_histogram
	//glResetHistogram				//EXT_histogram
	//glResetMinmax					//EXT_histogram
	//glSeparableFilter2D			//EXT_convolution
	
	//OpenGL 1.3
	//LUA_GLUEARB(glActiveTexture)	//ARB_multitexture
	//glClientActiveTexture			//ARB_multitexture
	//glCompressedTexImage1D		//ARB_texture_compression
	//glCompressedTexImage2D		//ARB_texture_compression
	//glCompressedTexImage3D		//ARB_texture_compression
	//glCompressedTexSubImage1D		//ARB_texture_compression
	//glCompressedTexSubImage2D		//ARB_texture_compression
	//glCompressedTexSubImage3D		//ARB_texture_compression
	//glGetCompressedTexImage		//ARB_texture_compression
	//LUA_GLUE(glMultiTexCoord)		//ARB_multitexture
	//glSampleCoverage				//ARB_multisample
	//glSamplePass					//ARB_multisample
	//glLoadTransposeMatrix			//ARB_transpose_matrix
	//glMultTransposeMatrix			//ARB_transpose_matrix
	
	//OpenGL 1.4
	//glBlendFuncSeparate			//EXT_blend_func_separate
	//glFogCoord					//EXT_fog_coord
	//glFogCoordPointer				//EXT_fog_coord
	//glMultiDrawArrays				//EXT_multi_draw_arrays
	//glMultiDrawElements			//EXT_multi_draw_arrays
	//glPointParameter				//NV_point_sprite, ARB_point_parameters
	//glSecondaryColor				//EXT_secondary_color
	//glSecondaryColorPointer		//EXT_secondary_color
	//glWindowPos					//ARB_window_pos
	
	//OpenGL 1.5
	//glBeginQuery					//ARB_occlusion_query
	//glBindBuffer					//ARB_vertex_buffer_object
	//glBufferData					//ARB_vertex_buffer_object
	//glBufferSubData				//ARB_vertex_buffer_object
	//glDeleteBuffers				//ARB_vertex_buffer_object
	//glDeleteQueries				//ARB_occlusion_query
	//glEndQuery					//ARB_occlusion_query
	//glGenBuffers					//ARB_vertex_buffer_object
	//glGenQueries					//ARB_occlusion_query
	//glGetBufferParameter			//ARB_vertex_buffer_object
	//glGetBufferPointer			//ARB_vertex_buffer_object
	//glGetBufferSubData			//ARB_vertex_buffer_object
	//glGetQuery					//ARB_occlusion_query
	//glGetQueryObject				//ARB_occlusion_query
	//glIsBuffer					//ARB_vertex_buffer_object
	//glIsQuery						//ARB_occlusion_query
	//glMapBuffer					//ARB_vertex_buffer_object
	//glUnmapBuffer					//ARB_vertex_buffer_object
	
	//OpenGL 2.0
	//glAttachShader
	//glBindAttribLocation			//ARB_vertex_shader
	//glBlendEquationSeparate		//ATI_blend_equation_separate, EXT_blend_equation_separate
	//glCompileShader				//ARB_shader_objects
	//glCreateProgram				//ARB_shader_objects has glCreateProgramObject
	//glCreateShader				//ARB_shader_objects has glCreateShaderObject
	//glDeleteProgram				//ARB_vertex_program
	//glDeleteShader				//ARB_shader_objects has glDeleteObject
	//glDetachShader				//ARB_shader_objects has glDetachObject
	//glDisableVertexAttribArray	//ARB_vertex_program
	//glDrawBuffers					//ARB_draw_buffers
	//glEnableVertexAttribArray		//ARB_vertex_program
	//glGetActiveAttrib				//ARB_vertex_shader
	//glGetActiveUniform			//ARB_shader_objects
	//glGetAttachedShaders			//ARB_shader_objects has glGetAttachedObjects
	//glGetProgram					//ARB_vertex_program
	//glGetProgramInfoLog			//ARB_shader_objects has glGetInfoLog
	//glGetShader
	//glGetShaderInfoLog			//ARB_shader_objects has glGetInfoLog
	//glGetShaderSource				//ARB_shader_objects
	//glGetUniform					//ARB_shader_objects
	//glGetUniformLocation			//ARB_shader_objects
	//glGetVertexAttrib				//ARB_vertex_program
	//glGetVertexAttribPointer		//ARB_vertex_program
	//glIsProgram					//ARB_vertex_program
	//glIsShader
	//glLinkShader					//ARB_shader_objects has glLinkProgram
	//glShaderSource				//ARB_shader_objects
	//glStencilFuncSeparate			//ATI_separate_stencil
	//glStencilMaskSeparate
	//glStencilOpSeparate			//ATI_separate_stencil
	//glUniform						//ARB_shader_objects
	//glUniformMatrix				//ARB_shader_objects
	//glUseProgram					//ARB_shader_objects has glUseProgramObject
	//glValidateProgram				//ARB_shader_objects
	//glVertexAttrib				//ARB_vertex_program
	//glVertexAttribPointer			//ARB_vertex_program

//terminator
	{NULL, NULL}
};

void luaBindGL(lua_State *L) {
	lua_getglobal(L, "gl");
	luaAddBindings(L, luaBindings_GL);

	//special case for extension-loaded functions
	//only load them if we got their pointers
#ifdef __APPLE__
	LUA_GLUE_FUNC_SAFE(glTexImage3D)
#else
	LUA_GLUE_FUNC_SAFE_EXT(glTexImage3D)
#endif
	LUA_GLUE_FUNC_SAFE_ARB(glActiveTexture)	//ARB_multitexture
	LUA_GLUE_FUNC_SAFE_SUFFIX(glMultiTexCoord, 1dvARB)		//ARB_multitexture

//gl.h
	LUAGi(GL_VERSION_1_1)
	LUAGi(GL_VERSION_1_2)
	LUAGi(GL_VERSION_1_3)
	LUAGi(GL_VERSION_1_4)
	LUAGi(GL_VERSION_1_5)
	LUAGi(GL_VERSION_2_0)
	LUAGi(GL_ACCUM)
	LUAGi(GL_LOAD)
	LUAGi(GL_RETURN)
	LUAGi(GL_MULT)
	LUAGi(GL_ADD)
	LUAGi(GL_NEVER)
	LUAGi(GL_LESS)
	LUAGi(GL_EQUAL)
	LUAGi(GL_LEQUAL)
	LUAGi(GL_GREATER)
	LUAGi(GL_NOTEQUAL)
	LUAGi(GL_GEQUAL)
	LUAGi(GL_ALWAYS)
	LUAGi(GL_CURRENT_BIT)
	LUAGi(GL_POINT_BIT)
	LUAGi(GL_LINE_BIT)
	LUAGi(GL_POLYGON_BIT)
	LUAGi(GL_POLYGON_STIPPLE_BIT)
	LUAGi(GL_PIXEL_MODE_BIT)
	LUAGi(GL_LIGHTING_BIT)
	LUAGi(GL_FOG_BIT)
	LUAGi(GL_DEPTH_BUFFER_BIT)
	LUAGi(GL_ACCUM_BUFFER_BIT)
	LUAGi(GL_STENCIL_BUFFER_BIT)
	LUAGi(GL_VIEWPORT_BIT)
	LUAGi(GL_TRANSFORM_BIT)
	LUAGi(GL_ENABLE_BIT)
	LUAGi(GL_COLOR_BUFFER_BIT)
	LUAGi(GL_HINT_BIT)
	LUAGi(GL_EVAL_BIT)
	LUAGi(GL_LIST_BIT)
	LUAGi(GL_TEXTURE_BIT)
	LUAGi(GL_SCISSOR_BIT)
	LUAGi(GL_ALL_ATTRIB_BITS)
	LUAGi(GL_POINTS)
	LUAGi(GL_LINES)
	LUAGi(GL_LINE_LOOP)
	LUAGi(GL_LINE_STRIP)
	LUAGi(GL_TRIANGLES)
	LUAGi(GL_TRIANGLE_STRIP)
	LUAGi(GL_TRIANGLE_FAN)
	LUAGi(GL_QUADS)
	LUAGi(GL_QUAD_STRIP)
	LUAGi(GL_POLYGON)
	LUAGi(GL_ZERO)
	LUAGi(GL_ONE)
	LUAGi(GL_SRC_COLOR)
	LUAGi(GL_ONE_MINUS_SRC_COLOR)
	LUAGi(GL_SRC_ALPHA)
	LUAGi(GL_ONE_MINUS_SRC_ALPHA)
	LUAGi(GL_DST_ALPHA)
	LUAGi(GL_ONE_MINUS_DST_ALPHA)
	LUAGi(GL_DST_COLOR)
	LUAGi(GL_ONE_MINUS_DST_COLOR)
	LUAGi(GL_SRC_ALPHA_SATURATE)
	LUAGi(GL_TRUE)
	LUAGi(GL_FALSE)
	LUAGi(GL_CLIP_PLANE0)
	LUAGi(GL_CLIP_PLANE1)
	LUAGi(GL_CLIP_PLANE2)
	LUAGi(GL_CLIP_PLANE3)
	LUAGi(GL_CLIP_PLANE4)
	LUAGi(GL_CLIP_PLANE5)
	LUAGi(GL_BYTE)
	LUAGi(GL_UNSIGNED_BYTE)
	LUAGi(GL_SHORT)
	LUAGi(GL_UNSIGNED_SHORT)
	LUAGi(GL_INT)
	LUAGi(GL_UNSIGNED_INT)
	LUAGi(GL_FLOAT)
	LUAGi(GL_2_BYTES)
	LUAGi(GL_3_BYTES)
	LUAGi(GL_4_BYTES)
	LUAGi(GL_DOUBLE)
	LUAGi(GL_NONE)
	LUAGi(GL_FRONT_LEFT)
	LUAGi(GL_FRONT_RIGHT)
	LUAGi(GL_BACK_LEFT)
	LUAGi(GL_BACK_RIGHT)
	LUAGi(GL_FRONT)
	LUAGi(GL_BACK)
	LUAGi(GL_LEFT)
	LUAGi(GL_RIGHT)
	LUAGi(GL_FRONT_AND_BACK)
	LUAGi(GL_AUX0)
	LUAGi(GL_AUX1)
	LUAGi(GL_AUX2)
	LUAGi(GL_AUX3)
	LUAGi(GL_NO_ERROR)
	LUAGi(GL_INVALID_ENUM)
	LUAGi(GL_INVALID_VALUE)
	LUAGi(GL_INVALID_OPERATION)
	LUAGi(GL_STACK_OVERFLOW)
	LUAGi(GL_STACK_UNDERFLOW)
	LUAGi(GL_OUT_OF_MEMORY)
	LUAGi(GL_2D)
	LUAGi(GL_3D)
	LUAGi(GL_3D_COLOR)
	LUAGi(GL_3D_COLOR_TEXTURE)
	LUAGi(GL_4D_COLOR_TEXTURE)
	LUAGi(GL_PASS_THROUGH_TOKEN)
	LUAGi(GL_POINT_TOKEN)
	LUAGi(GL_LINE_TOKEN)
	LUAGi(GL_POLYGON_TOKEN)
	LUAGi(GL_BITMAP_TOKEN)
	LUAGi(GL_DRAW_PIXEL_TOKEN)
	LUAGi(GL_COPY_PIXEL_TOKEN)
	LUAGi(GL_LINE_RESET_TOKEN)
	LUAGi(GL_EXP)
	LUAGi(GL_EXP2)
	LUAGi(GL_CW)
	LUAGi(GL_CCW)
	LUAGi(GL_COEFF)
	LUAGi(GL_ORDER)
	LUAGi(GL_DOMAIN)
	LUAGi(GL_CURRENT_COLOR)
	LUAGi(GL_CURRENT_INDEX)
	LUAGi(GL_CURRENT_NORMAL)
	LUAGi(GL_CURRENT_TEXTURE_COORDS)
	LUAGi(GL_CURRENT_RASTER_COLOR)
	LUAGi(GL_CURRENT_RASTER_INDEX)
	LUAGi(GL_CURRENT_RASTER_TEXTURE_COORDS)
	LUAGi(GL_CURRENT_RASTER_POSITION)
	LUAGi(GL_CURRENT_RASTER_POSITION_VALID)
	LUAGi(GL_CURRENT_RASTER_DISTANCE)
	LUAGi(GL_POINT_SMOOTH)
	LUAGi(GL_POINT_SIZE)
	LUAGi(GL_POINT_SIZE_RANGE)
	LUAGi(GL_POINT_SIZE_GRANULARITY)
	LUAGi(GL_LINE_SMOOTH)
	LUAGi(GL_LINE_WIDTH)
	LUAGi(GL_LINE_WIDTH_RANGE)
	LUAGi(GL_LINE_WIDTH_GRANULARITY)
	LUAGi(GL_LINE_STIPPLE)
	LUAGi(GL_LINE_STIPPLE_PATTERN)
	LUAGi(GL_LINE_STIPPLE_REPEAT)
	LUAGi(GL_LIST_MODE)
	LUAGi(GL_MAX_LIST_NESTING)
	LUAGi(GL_LIST_BASE)
	LUAGi(GL_LIST_INDEX)
	LUAGi(GL_POLYGON_MODE)
	LUAGi(GL_POLYGON_SMOOTH)
	LUAGi(GL_POLYGON_STIPPLE)
	LUAGi(GL_EDGE_FLAG)
	LUAGi(GL_CULL_FACE)
	LUAGi(GL_CULL_FACE_MODE)
	LUAGi(GL_FRONT_FACE)
	LUAGi(GL_LIGHTING)
	LUAGi(GL_LIGHT_MODEL_LOCAL_VIEWER)
	LUAGi(GL_LIGHT_MODEL_TWO_SIDE)
	LUAGi(GL_LIGHT_MODEL_AMBIENT)
	LUAGi(GL_SHADE_MODEL)
	LUAGi(GL_COLOR_MATERIAL_FACE)
	LUAGi(GL_COLOR_MATERIAL_PARAMETER)
	LUAGi(GL_COLOR_MATERIAL)
	LUAGi(GL_FOG)
	LUAGi(GL_FOG_INDEX)
	LUAGi(GL_FOG_DENSITY)
	LUAGi(GL_FOG_START)
	LUAGi(GL_FOG_END)
	LUAGi(GL_FOG_MODE)
	LUAGi(GL_FOG_COLOR)
	LUAGi(GL_DEPTH_RANGE)
	LUAGi(GL_DEPTH_TEST)
	LUAGi(GL_DEPTH_WRITEMASK)
	LUAGi(GL_DEPTH_CLEAR_VALUE)
	LUAGi(GL_DEPTH_FUNC)
	LUAGi(GL_ACCUM_CLEAR_VALUE)
	LUAGi(GL_STENCIL_TEST)
	LUAGi(GL_STENCIL_CLEAR_VALUE)
	LUAGi(GL_STENCIL_FUNC)
	LUAGi(GL_STENCIL_VALUE_MASK)
	LUAGi(GL_STENCIL_FAIL)
	LUAGi(GL_STENCIL_PASS_DEPTH_FAIL)
	LUAGi(GL_STENCIL_PASS_DEPTH_PASS)
	LUAGi(GL_STENCIL_REF)
	LUAGi(GL_STENCIL_WRITEMASK)
	LUAGi(GL_MATRIX_MODE)
	LUAGi(GL_NORMALIZE)
	LUAGi(GL_VIEWPORT)
	LUAGi(GL_MODELVIEW_STACK_DEPTH)
	LUAGi(GL_PROJECTION_STACK_DEPTH)
	LUAGi(GL_TEXTURE_STACK_DEPTH)
	LUAGi(GL_MODELVIEW_MATRIX)
	LUAGi(GL_PROJECTION_MATRIX)
	LUAGi(GL_TEXTURE_MATRIX)
	LUAGi(GL_ATTRIB_STACK_DEPTH)
	LUAGi(GL_CLIENT_ATTRIB_STACK_DEPTH)
	LUAGi(GL_ALPHA_TEST)
	LUAGi(GL_ALPHA_TEST_FUNC)
	LUAGi(GL_ALPHA_TEST_REF)
	LUAGi(GL_DITHER)
	LUAGi(GL_BLEND_DST)
	LUAGi(GL_BLEND_SRC)
	LUAGi(GL_BLEND)
	LUAGi(GL_LOGIC_OP_MODE)
	LUAGi(GL_INDEX_LOGIC_OP)
	LUAGi(GL_COLOR_LOGIC_OP)
	LUAGi(GL_AUX_BUFFERS)
	LUAGi(GL_DRAW_BUFFER)
	LUAGi(GL_READ_BUFFER)
	LUAGi(GL_SCISSOR_BOX)
	LUAGi(GL_SCISSOR_TEST)
	LUAGi(GL_INDEX_CLEAR_VALUE)
	LUAGi(GL_INDEX_WRITEMASK)
	LUAGi(GL_COLOR_CLEAR_VALUE)
	LUAGi(GL_COLOR_WRITEMASK)
	LUAGi(GL_INDEX_MODE)
	LUAGi(GL_RGBA_MODE)
	LUAGi(GL_DOUBLEBUFFER)
	LUAGi(GL_STEREO)
	LUAGi(GL_RENDER_MODE)
	LUAGi(GL_PERSPECTIVE_CORRECTION_HINT)
	LUAGi(GL_POINT_SMOOTH_HINT)
	LUAGi(GL_LINE_SMOOTH_HINT)
	LUAGi(GL_POLYGON_SMOOTH_HINT)
	LUAGi(GL_FOG_HINT)
	LUAGi(GL_TEXTURE_GEN_S)
	LUAGi(GL_TEXTURE_GEN_T)
	LUAGi(GL_TEXTURE_GEN_R)
	LUAGi(GL_TEXTURE_GEN_Q)
	LUAGi(GL_PIXEL_MAP_I_TO_I)
	LUAGi(GL_PIXEL_MAP_S_TO_S)
	LUAGi(GL_PIXEL_MAP_I_TO_R)
	LUAGi(GL_PIXEL_MAP_I_TO_G)
	LUAGi(GL_PIXEL_MAP_I_TO_B)
	LUAGi(GL_PIXEL_MAP_I_TO_A)
	LUAGi(GL_PIXEL_MAP_R_TO_R)
	LUAGi(GL_PIXEL_MAP_G_TO_G)
	LUAGi(GL_PIXEL_MAP_B_TO_B)
	LUAGi(GL_PIXEL_MAP_A_TO_A)
	LUAGi(GL_PIXEL_MAP_I_TO_I_SIZE)
	LUAGi(GL_PIXEL_MAP_S_TO_S_SIZE)
	LUAGi(GL_PIXEL_MAP_I_TO_R_SIZE)
	LUAGi(GL_PIXEL_MAP_I_TO_G_SIZE)
	LUAGi(GL_PIXEL_MAP_I_TO_B_SIZE)
	LUAGi(GL_PIXEL_MAP_I_TO_A_SIZE)
	LUAGi(GL_PIXEL_MAP_R_TO_R_SIZE)
	LUAGi(GL_PIXEL_MAP_G_TO_G_SIZE)
	LUAGi(GL_PIXEL_MAP_B_TO_B_SIZE)
	LUAGi(GL_PIXEL_MAP_A_TO_A_SIZE)
	LUAGi(GL_UNPACK_SWAP_BYTES)
	LUAGi(GL_UNPACK_LSB_FIRST)
	LUAGi(GL_UNPACK_ROW_LENGTH)
	LUAGi(GL_UNPACK_SKIP_ROWS)
	LUAGi(GL_UNPACK_SKIP_PIXELS)
	LUAGi(GL_UNPACK_ALIGNMENT)
	LUAGi(GL_PACK_SWAP_BYTES)
	LUAGi(GL_PACK_LSB_FIRST)
	LUAGi(GL_PACK_ROW_LENGTH)
	LUAGi(GL_PACK_SKIP_ROWS)
	LUAGi(GL_PACK_SKIP_PIXELS)
	LUAGi(GL_PACK_ALIGNMENT)
	LUAGi(GL_MAP_COLOR)
	LUAGi(GL_MAP_STENCIL)
	LUAGi(GL_INDEX_SHIFT)
	LUAGi(GL_INDEX_OFFSET)
	LUAGi(GL_RED_SCALE)
	LUAGi(GL_RED_BIAS)
	LUAGi(GL_ZOOM_X)
	LUAGi(GL_ZOOM_Y)
	LUAGi(GL_GREEN_SCALE)
	LUAGi(GL_GREEN_BIAS)
	LUAGi(GL_BLUE_SCALE)
	LUAGi(GL_BLUE_BIAS)
	LUAGi(GL_ALPHA_SCALE)
	LUAGi(GL_ALPHA_BIAS)
	LUAGi(GL_DEPTH_SCALE)
	LUAGi(GL_DEPTH_BIAS)
	LUAGi(GL_MAX_EVAL_ORDER)
	LUAGi(GL_MAX_LIGHTS)
	LUAGi(GL_MAX_CLIP_PLANES)
	LUAGi(GL_MAX_TEXTURE_SIZE)
	LUAGi(GL_MAX_PIXEL_MAP_TABLE)
	LUAGi(GL_MAX_ATTRIB_STACK_DEPTH)
	LUAGi(GL_MAX_MODELVIEW_STACK_DEPTH)
	LUAGi(GL_MAX_NAME_STACK_DEPTH)
	LUAGi(GL_MAX_PROJECTION_STACK_DEPTH)
	LUAGi(GL_MAX_TEXTURE_STACK_DEPTH)
	LUAGi(GL_MAX_VIEWPORT_DIMS)
	LUAGi(GL_MAX_CLIENT_ATTRIB_STACK_DEPTH)
	LUAGi(GL_SUBPIXEL_BITS)
	LUAGi(GL_INDEX_BITS)
	LUAGi(GL_RED_BITS)
	LUAGi(GL_GREEN_BITS)
	LUAGi(GL_BLUE_BITS)
	LUAGi(GL_ALPHA_BITS)
	LUAGi(GL_DEPTH_BITS)
	LUAGi(GL_STENCIL_BITS)
	LUAGi(GL_ACCUM_RED_BITS)
	LUAGi(GL_ACCUM_GREEN_BITS)
	LUAGi(GL_ACCUM_BLUE_BITS)
	LUAGi(GL_ACCUM_ALPHA_BITS)
	LUAGi(GL_NAME_STACK_DEPTH)
	LUAGi(GL_AUTO_NORMAL)
	LUAGi(GL_MAP1_COLOR_4)
	LUAGi(GL_MAP1_INDEX)
	LUAGi(GL_MAP1_NORMAL)
	LUAGi(GL_MAP1_TEXTURE_COORD_1)
	LUAGi(GL_MAP1_TEXTURE_COORD_2)
	LUAGi(GL_MAP1_TEXTURE_COORD_3)
	LUAGi(GL_MAP1_TEXTURE_COORD_4)
	LUAGi(GL_MAP1_VERTEX_3)
	LUAGi(GL_MAP1_VERTEX_4)
	LUAGi(GL_MAP2_COLOR_4)
	LUAGi(GL_MAP2_INDEX)
	LUAGi(GL_MAP2_NORMAL)
	LUAGi(GL_MAP2_TEXTURE_COORD_1)
	LUAGi(GL_MAP2_TEXTURE_COORD_2)
	LUAGi(GL_MAP2_TEXTURE_COORD_3)
	LUAGi(GL_MAP2_TEXTURE_COORD_4)
	LUAGi(GL_MAP2_VERTEX_3)
	LUAGi(GL_MAP2_VERTEX_4)
	LUAGi(GL_MAP1_GRID_DOMAIN)
	LUAGi(GL_MAP1_GRID_SEGMENTS)
	LUAGi(GL_MAP2_GRID_DOMAIN)
	LUAGi(GL_MAP2_GRID_SEGMENTS)
	LUAGi(GL_TEXTURE_1D)
	LUAGi(GL_TEXTURE_2D)
	LUAGi(GL_FEEDBACK_BUFFER_POINTER)
	LUAGi(GL_FEEDBACK_BUFFER_SIZE)
	LUAGi(GL_FEEDBACK_BUFFER_TYPE)
	LUAGi(GL_SELECTION_BUFFER_POINTER)
	LUAGi(GL_SELECTION_BUFFER_SIZE)
	LUAGi(GL_TEXTURE_WIDTH)
	LUAGi(GL_TEXTURE_HEIGHT)
	LUAGi(GL_TEXTURE_INTERNAL_FORMAT)
	LUAGi(GL_TEXTURE_BORDER_COLOR)
	LUAGi(GL_TEXTURE_BORDER)
	LUAGi(GL_DONT_CARE)
	LUAGi(GL_FASTEST)
	LUAGi(GL_NICEST)
	LUAGi(GL_LIGHT0)
	LUAGi(GL_LIGHT1)
	LUAGi(GL_LIGHT2)
	LUAGi(GL_LIGHT3)
	LUAGi(GL_LIGHT4)
	LUAGi(GL_LIGHT5)
	LUAGi(GL_LIGHT6)
	LUAGi(GL_LIGHT7)
	LUAGi(GL_AMBIENT)
	LUAGi(GL_DIFFUSE)
	LUAGi(GL_SPECULAR)
	LUAGi(GL_POSITION)
	LUAGi(GL_SPOT_DIRECTION)
	LUAGi(GL_SPOT_EXPONENT)
	LUAGi(GL_SPOT_CUTOFF)
	LUAGi(GL_CONSTANT_ATTENUATION)
	LUAGi(GL_LINEAR_ATTENUATION)
	LUAGi(GL_QUADRATIC_ATTENUATION)
	LUAGi(GL_COMPILE)
	LUAGi(GL_COMPILE_AND_EXECUTE)
	LUAGi(GL_CLEAR)
	LUAGi(GL_AND)
	LUAGi(GL_AND_REVERSE)
	LUAGi(GL_COPY)
	LUAGi(GL_AND_INVERTED)
	LUAGi(GL_NOOP)
	LUAGi(GL_XOR)
	LUAGi(GL_OR)
	LUAGi(GL_NOR)
	LUAGi(GL_EQUIV)
	LUAGi(GL_INVERT)
	LUAGi(GL_OR_REVERSE)
	LUAGi(GL_COPY_INVERTED)
	LUAGi(GL_OR_INVERTED)
	LUAGi(GL_NAND)
	LUAGi(GL_SET)
	LUAGi(GL_EMISSION)
	LUAGi(GL_SHININESS)
	LUAGi(GL_AMBIENT_AND_DIFFUSE)
	LUAGi(GL_COLOR_INDEXES)
	LUAGi(GL_MODELVIEW)
	LUAGi(GL_PROJECTION)
	LUAGi(GL_TEXTURE)
	LUAGi(GL_COLOR)
	LUAGi(GL_DEPTH)
	LUAGi(GL_STENCIL)
	LUAGi(GL_COLOR_INDEX)
	LUAGi(GL_STENCIL_INDEX)
	LUAGi(GL_DEPTH_COMPONENT)
	LUAGi(GL_RED)
	LUAGi(GL_GREEN)
	LUAGi(GL_BLUE)
	LUAGi(GL_ALPHA)
	LUAGi(GL_RGB)
	LUAGi(GL_RGBA)
	LUAGi(GL_LUMINANCE)
	LUAGi(GL_LUMINANCE_ALPHA)
	LUAGi(GL_BITMAP)
	LUAGi(GL_POINT)
	LUAGi(GL_LINE)
	LUAGi(GL_FILL)
	LUAGi(GL_RENDER)
	LUAGi(GL_FEEDBACK)
	LUAGi(GL_SELECT)
	LUAGi(GL_FLAT)
	LUAGi(GL_SMOOTH)
	LUAGi(GL_KEEP)
	LUAGi(GL_REPLACE)
	LUAGi(GL_INCR)
	LUAGi(GL_DECR)
	LUAGi(GL_VENDOR)
	LUAGi(GL_RENDERER)
	LUAGi(GL_VERSION)
	LUAGi(GL_EXTENSIONS)
	LUAGi(GL_S)
	LUAGi(GL_T)
	LUAGi(GL_R)
	LUAGi(GL_Q)
	LUAGi(GL_MODULATE)
	LUAGi(GL_DECAL)
	LUAGi(GL_TEXTURE_ENV_MODE)
	LUAGi(GL_TEXTURE_ENV_COLOR)
	LUAGi(GL_TEXTURE_ENV)
	LUAGi(GL_EYE_LINEAR)
	LUAGi(GL_OBJECT_LINEAR)
	LUAGi(GL_SPHERE_MAP)
	LUAGi(GL_TEXTURE_GEN_MODE)
	LUAGi(GL_OBJECT_PLANE)
	LUAGi(GL_EYE_PLANE)
	LUAGi(GL_NEAREST)
	LUAGi(GL_LINEAR)
	LUAGi(GL_NEAREST_MIPMAP_NEAREST)
	LUAGi(GL_LINEAR_MIPMAP_NEAREST)
	LUAGi(GL_NEAREST_MIPMAP_LINEAR)
	LUAGi(GL_LINEAR_MIPMAP_LINEAR)
	LUAGi(GL_TEXTURE_MAG_FILTER)
	LUAGi(GL_TEXTURE_MIN_FILTER)
	LUAGi(GL_TEXTURE_WRAP_S)
	LUAGi(GL_TEXTURE_WRAP_T)
	LUAGi(GL_CLAMP)
	LUAGi(GL_REPEAT)
	LUAGi(GL_CLIENT_PIXEL_STORE_BIT)
	LUAGi(GL_CLIENT_VERTEX_ARRAY_BIT)
	LUAGi(GL_CLIENT_ALL_ATTRIB_BITS)
	LUAGi(GL_POLYGON_OFFSET_FACTOR)
	LUAGi(GL_POLYGON_OFFSET_UNITS)
	LUAGi(GL_POLYGON_OFFSET_POINT)
	LUAGi(GL_POLYGON_OFFSET_LINE)
	LUAGi(GL_POLYGON_OFFSET_FILL)
	LUAGi(GL_ALPHA4)
	LUAGi(GL_ALPHA8)
	LUAGi(GL_ALPHA12)
	LUAGi(GL_ALPHA16)
	LUAGi(GL_LUMINANCE4)
	LUAGi(GL_LUMINANCE8)
	LUAGi(GL_LUMINANCE12)
	LUAGi(GL_LUMINANCE16)
	LUAGi(GL_LUMINANCE4_ALPHA4)
	LUAGi(GL_LUMINANCE6_ALPHA2)
	LUAGi(GL_LUMINANCE8_ALPHA8)
	LUAGi(GL_LUMINANCE12_ALPHA4)
	LUAGi(GL_LUMINANCE12_ALPHA12)
	LUAGi(GL_LUMINANCE16_ALPHA16)
	LUAGi(GL_INTENSITY)
	LUAGi(GL_INTENSITY4)
	LUAGi(GL_INTENSITY8)
	LUAGi(GL_INTENSITY12)
	LUAGi(GL_INTENSITY16)
	LUAGi(GL_R3_G3_B2)
	LUAGi(GL_RGB4)
	LUAGi(GL_RGB5)
	LUAGi(GL_RGB8)
	LUAGi(GL_RGB10)
	LUAGi(GL_RGB12)
	LUAGi(GL_RGB16)
	LUAGi(GL_RGBA2)
	LUAGi(GL_RGBA4)
	LUAGi(GL_RGB5_A1)
	LUAGi(GL_RGBA8)
	LUAGi(GL_RGB10_A2)
	LUAGi(GL_RGBA12)
	LUAGi(GL_RGBA16)
	LUAGi(GL_TEXTURE_RED_SIZE)
	LUAGi(GL_TEXTURE_GREEN_SIZE)
	LUAGi(GL_TEXTURE_BLUE_SIZE)
	LUAGi(GL_TEXTURE_ALPHA_SIZE)
	LUAGi(GL_TEXTURE_LUMINANCE_SIZE)
	LUAGi(GL_TEXTURE_INTENSITY_SIZE)
	LUAGi(GL_PROXY_TEXTURE_1D)
	LUAGi(GL_PROXY_TEXTURE_2D)
	LUAGi(GL_TEXTURE_PRIORITY)
	LUAGi(GL_TEXTURE_RESIDENT)
	LUAGi(GL_TEXTURE_BINDING_1D)
	LUAGi(GL_TEXTURE_BINDING_2D)
	LUAGi(GL_TEXTURE_BINDING_3D)
	LUAGi(GL_VERTEX_ARRAY)
	LUAGi(GL_NORMAL_ARRAY)
	LUAGi(GL_COLOR_ARRAY)
	LUAGi(GL_INDEX_ARRAY)
	LUAGi(GL_TEXTURE_COORD_ARRAY)
	LUAGi(GL_EDGE_FLAG_ARRAY)
	LUAGi(GL_VERTEX_ARRAY_SIZE)
	LUAGi(GL_VERTEX_ARRAY_TYPE)
	LUAGi(GL_VERTEX_ARRAY_STRIDE)
	LUAGi(GL_NORMAL_ARRAY_TYPE)
	LUAGi(GL_NORMAL_ARRAY_STRIDE)
	LUAGi(GL_COLOR_ARRAY_SIZE)
	LUAGi(GL_COLOR_ARRAY_TYPE)
	LUAGi(GL_COLOR_ARRAY_STRIDE)
	LUAGi(GL_INDEX_ARRAY_TYPE)
	LUAGi(GL_INDEX_ARRAY_STRIDE)
	LUAGi(GL_TEXTURE_COORD_ARRAY_SIZE)
	LUAGi(GL_TEXTURE_COORD_ARRAY_TYPE)
	LUAGi(GL_TEXTURE_COORD_ARRAY_STRIDE)
	LUAGi(GL_EDGE_FLAG_ARRAY_STRIDE)
	LUAGi(GL_VERTEX_ARRAY_POINTER)
	LUAGi(GL_NORMAL_ARRAY_POINTER)
	LUAGi(GL_COLOR_ARRAY_POINTER)
	LUAGi(GL_INDEX_ARRAY_POINTER)
	LUAGi(GL_TEXTURE_COORD_ARRAY_POINTER)
	LUAGi(GL_EDGE_FLAG_ARRAY_POINTER)
	LUAGi(GL_V2F)
	LUAGi(GL_V3F)
	LUAGi(GL_C4UB_V2F)
	LUAGi(GL_C4UB_V3F)
	LUAGi(GL_C3F_V3F)
	LUAGi(GL_N3F_V3F)
	LUAGi(GL_C4F_N3F_V3F)
	LUAGi(GL_T2F_V3F)
	LUAGi(GL_T4F_V4F)
	LUAGi(GL_T2F_C4UB_V3F)
	LUAGi(GL_T2F_C3F_V3F)
	LUAGi(GL_T2F_N3F_V3F)
	LUAGi(GL_T2F_C4F_N3F_V3F)
	LUAGi(GL_T4F_C4F_N3F_V4F)
	LUAGi(GL_BGR)
	LUAGi(GL_BGRA)
	LUAGi(GL_CONSTANT_COLOR)
	LUAGi(GL_ONE_MINUS_CONSTANT_COLOR)
	LUAGi(GL_CONSTANT_ALPHA)
	LUAGi(GL_ONE_MINUS_CONSTANT_ALPHA)
	LUAGi(GL_BLEND_COLOR)
	LUAGi(GL_FUNC_ADD)
	LUAGi(GL_MIN)
	LUAGi(GL_MAX)
	LUAGi(GL_BLEND_EQUATION)
	LUAGi(GL_BLEND_EQUATION_RGB)
	LUAGi(GL_BLEND_EQUATION_ALPHA)
	LUAGi(GL_FUNC_SUBTRACT)
	LUAGi(GL_FUNC_REVERSE_SUBTRACT)
	LUAGi(GL_COLOR_MATRIX)
	LUAGi(GL_COLOR_MATRIX_STACK_DEPTH)
	LUAGi(GL_MAX_COLOR_MATRIX_STACK_DEPTH)
	LUAGi(GL_POST_COLOR_MATRIX_RED_SCALE)
	LUAGi(GL_POST_COLOR_MATRIX_GREEN_SCALE)
	LUAGi(GL_POST_COLOR_MATRIX_BLUE_SCALE)
	LUAGi(GL_POST_COLOR_MATRIX_ALPHA_SCALE)
	LUAGi(GL_POST_COLOR_MATRIX_RED_BIAS)
	LUAGi(GL_POST_COLOR_MATRIX_GREEN_BIAS)
	LUAGi(GL_POST_COLOR_MATRIX_BLUE_BIAS)
	LUAGi(GL_POST_COLOR_MATRIX_ALPHA_BIAS)
	LUAGi(GL_COLOR_TABLE)
	LUAGi(GL_POST_CONVOLUTION_COLOR_TABLE)
	LUAGi(GL_POST_COLOR_MATRIX_COLOR_TABLE)
	LUAGi(GL_PROXY_COLOR_TABLE)
	LUAGi(GL_PROXY_POST_CONVOLUTION_COLOR_TABLE)
	LUAGi(GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE)
	LUAGi(GL_COLOR_TABLE_SCALE)
	LUAGi(GL_COLOR_TABLE_BIAS)
	LUAGi(GL_COLOR_TABLE_FORMAT)
	LUAGi(GL_COLOR_TABLE_WIDTH)
	LUAGi(GL_COLOR_TABLE_RED_SIZE)
	LUAGi(GL_COLOR_TABLE_GREEN_SIZE)
	LUAGi(GL_COLOR_TABLE_BLUE_SIZE)
	LUAGi(GL_COLOR_TABLE_ALPHA_SIZE)
	LUAGi(GL_COLOR_TABLE_LUMINANCE_SIZE)
	LUAGi(GL_COLOR_TABLE_INTENSITY_SIZE)
	LUAGi(GL_CONVOLUTION_1D)
	LUAGi(GL_CONVOLUTION_2D)
	LUAGi(GL_SEPARABLE_2D)
	LUAGi(GL_CONVOLUTION_BORDER_MODE)
	LUAGi(GL_CONVOLUTION_FILTER_SCALE)
	LUAGi(GL_CONVOLUTION_FILTER_BIAS)
	LUAGi(GL_REDUCE)
	LUAGi(GL_CONVOLUTION_FORMAT)
	LUAGi(GL_CONVOLUTION_WIDTH)
	LUAGi(GL_CONVOLUTION_HEIGHT)
	LUAGi(GL_MAX_CONVOLUTION_WIDTH)
	LUAGi(GL_MAX_CONVOLUTION_HEIGHT)
	LUAGi(GL_POST_CONVOLUTION_RED_SCALE)
	LUAGi(GL_POST_CONVOLUTION_GREEN_SCALE)
	LUAGi(GL_POST_CONVOLUTION_BLUE_SCALE)
	LUAGi(GL_POST_CONVOLUTION_ALPHA_SCALE)
	LUAGi(GL_POST_CONVOLUTION_RED_BIAS)
	LUAGi(GL_POST_CONVOLUTION_GREEN_BIAS)
	LUAGi(GL_POST_CONVOLUTION_BLUE_BIAS)
	LUAGi(GL_POST_CONVOLUTION_ALPHA_BIAS)
	LUAGi(GL_CONSTANT_BORDER)
	LUAGi(GL_REPLICATE_BORDER)
	LUAGi(GL_CONVOLUTION_BORDER_COLOR)
	LUAGi(GL_MAX_ELEMENTS_VERTICES)
	LUAGi(GL_MAX_ELEMENTS_INDICES)
	LUAGi(GL_HISTOGRAM)
	LUAGi(GL_PROXY_HISTOGRAM)
	LUAGi(GL_HISTOGRAM_WIDTH)
	LUAGi(GL_HISTOGRAM_FORMAT)
	LUAGi(GL_HISTOGRAM_RED_SIZE)
	LUAGi(GL_HISTOGRAM_GREEN_SIZE)
	LUAGi(GL_HISTOGRAM_BLUE_SIZE)
	LUAGi(GL_HISTOGRAM_ALPHA_SIZE)
	LUAGi(GL_HISTOGRAM_LUMINANCE_SIZE)
	LUAGi(GL_HISTOGRAM_SINK)
	LUAGi(GL_MINMAX)
	LUAGi(GL_MINMAX_FORMAT)
	LUAGi(GL_MINMAX_SINK)
	LUAGi(GL_TABLE_TOO_LARGE)
	LUAGi(GL_UNSIGNED_BYTE_3_3_2)
	LUAGi(GL_UNSIGNED_SHORT_4_4_4_4)
	LUAGi(GL_UNSIGNED_SHORT_5_5_5_1)
	LUAGi(GL_UNSIGNED_INT_8_8_8_8)
	LUAGi(GL_UNSIGNED_INT_10_10_10_2)
	LUAGi(GL_UNSIGNED_BYTE_2_3_3_REV)
	LUAGi(GL_UNSIGNED_SHORT_5_6_5)
	LUAGi(GL_UNSIGNED_SHORT_5_6_5_REV)
	LUAGi(GL_UNSIGNED_SHORT_4_4_4_4_REV)
	LUAGi(GL_UNSIGNED_SHORT_1_5_5_5_REV)
	LUAGi(GL_UNSIGNED_INT_8_8_8_8_REV)
	LUAGi(GL_UNSIGNED_INT_2_10_10_10_REV)
	LUAGi(GL_RESCALE_NORMAL)
	LUAGi(GL_LIGHT_MODEL_COLOR_CONTROL)
	LUAGi(GL_SINGLE_COLOR)
	LUAGi(GL_SEPARATE_SPECULAR_COLOR)
	LUAGi(GL_PACK_SKIP_IMAGES)
	LUAGi(GL_PACK_IMAGE_HEIGHT)
	LUAGi(GL_UNPACK_SKIP_IMAGES)
	LUAGi(GL_UNPACK_IMAGE_HEIGHT)
	LUAGi(GL_TEXTURE_3D)
	LUAGi(GL_PROXY_TEXTURE_3D)
	LUAGi(GL_TEXTURE_DEPTH)
	LUAGi(GL_TEXTURE_WRAP_R)
	LUAGi(GL_MAX_3D_TEXTURE_SIZE)
	LUAGi(GL_CLAMP_TO_EDGE)
	LUAGi(GL_CLAMP_TO_BORDER)
	LUAGi(GL_TEXTURE_MIN_LOD)
	LUAGi(GL_TEXTURE_MAX_LOD)
	LUAGi(GL_TEXTURE_BASE_LEVEL)
	LUAGi(GL_TEXTURE_MAX_LEVEL)
	LUAGi(GL_SMOOTH_POINT_SIZE_RANGE)
	LUAGi(GL_SMOOTH_POINT_SIZE_GRANULARITY)
	LUAGi(GL_SMOOTH_LINE_WIDTH_RANGE)
	LUAGi(GL_SMOOTH_LINE_WIDTH_GRANULARITY)
	LUAGi(GL_ALIASED_POINT_SIZE_RANGE)
	LUAGi(GL_ALIASED_LINE_WIDTH_RANGE)
	LUAGi(GL_TEXTURE0)
	LUAGi(GL_TEXTURE1)
	LUAGi(GL_TEXTURE2)
	LUAGi(GL_TEXTURE3)
	LUAGi(GL_TEXTURE4)
	LUAGi(GL_TEXTURE5)
	LUAGi(GL_TEXTURE6)
	LUAGi(GL_TEXTURE7)
	LUAGi(GL_TEXTURE8)
	LUAGi(GL_TEXTURE9)
	LUAGi(GL_TEXTURE10)
	LUAGi(GL_TEXTURE11)
	LUAGi(GL_TEXTURE12)
	LUAGi(GL_TEXTURE13)
	LUAGi(GL_TEXTURE14)
	LUAGi(GL_TEXTURE15)
	LUAGi(GL_TEXTURE16)
	LUAGi(GL_TEXTURE17)
	LUAGi(GL_TEXTURE18)
	LUAGi(GL_TEXTURE19)
	LUAGi(GL_TEXTURE20)
	LUAGi(GL_TEXTURE21)
	LUAGi(GL_TEXTURE22)
	LUAGi(GL_TEXTURE23)
	LUAGi(GL_TEXTURE24)
	LUAGi(GL_TEXTURE25)
	LUAGi(GL_TEXTURE26)
	LUAGi(GL_TEXTURE27)
	LUAGi(GL_TEXTURE28)
	LUAGi(GL_TEXTURE29)
	LUAGi(GL_TEXTURE30)
	LUAGi(GL_TEXTURE31)
	LUAGi(GL_ACTIVE_TEXTURE)
	LUAGi(GL_CLIENT_ACTIVE_TEXTURE)
	LUAGi(GL_MAX_TEXTURE_UNITS)
	LUAGi(GL_COMBINE)
	LUAGi(GL_COMBINE_RGB)
	LUAGi(GL_COMBINE_ALPHA)
	LUAGi(GL_RGB_SCALE)
	LUAGi(GL_ADD_SIGNED)
	LUAGi(GL_INTERPOLATE)
	LUAGi(GL_CONSTANT)
	LUAGi(GL_PRIMARY_COLOR)
	LUAGi(GL_PREVIOUS)
	LUAGi(GL_SUBTRACT)
	LUAGi(GL_SRC0_RGB)
	LUAGi(GL_SRC1_RGB)
	LUAGi(GL_SRC2_RGB)
#ifndef WIN32
	LUAGi(GL_SRC3_RGB)
	LUAGi(GL_SRC4_RGB)
	LUAGi(GL_SRC5_RGB)
	LUAGi(GL_SRC6_RGB)
	LUAGi(GL_SRC7_RGB)
#endif
	LUAGi(GL_SRC0_ALPHA)
	LUAGi(GL_SRC1_ALPHA)
	LUAGi(GL_SRC2_ALPHA)
#ifndef WIN32
	LUAGi(GL_SRC3_ALPHA)
	LUAGi(GL_SRC4_ALPHA)
	LUAGi(GL_SRC5_ALPHA)
	LUAGi(GL_SRC6_ALPHA)
	LUAGi(GL_SRC7_ALPHA)
#endif
	LUAGi(GL_SOURCE0_RGB)
	LUAGi(GL_SOURCE1_RGB)
	LUAGi(GL_SOURCE2_RGB)
#ifndef WIN32
	LUAGi(GL_SOURCE3_RGB)
	LUAGi(GL_SOURCE4_RGB)
	LUAGi(GL_SOURCE5_RGB)
	LUAGi(GL_SOURCE6_RGB)
	LUAGi(GL_SOURCE7_RGB)
#endif
	LUAGi(GL_SOURCE0_ALPHA)
	LUAGi(GL_SOURCE1_ALPHA)
	LUAGi(GL_SOURCE2_ALPHA)
#ifndef WIN32
	LUAGi(GL_SOURCE3_ALPHA)
	LUAGi(GL_SOURCE4_ALPHA)
	LUAGi(GL_SOURCE5_ALPHA)
	LUAGi(GL_SOURCE6_ALPHA)
	LUAGi(GL_SOURCE7_ALPHA)
#endif
	LUAGi(GL_OPERAND0_RGB)
	LUAGi(GL_OPERAND1_RGB)
	LUAGi(GL_OPERAND2_RGB)
#ifndef WIN32
	LUAGi(GL_OPERAND3_RGB)
	LUAGi(GL_OPERAND4_RGB)
	LUAGi(GL_OPERAND5_RGB)
	LUAGi(GL_OPERAND6_RGB)
	LUAGi(GL_OPERAND7_RGB)
#endif
	LUAGi(GL_OPERAND0_ALPHA)
	LUAGi(GL_OPERAND1_ALPHA)
	LUAGi(GL_OPERAND2_ALPHA)
#ifndef WIN32
	LUAGi(GL_OPERAND3_ALPHA)
	LUAGi(GL_OPERAND4_ALPHA)
	LUAGi(GL_OPERAND5_ALPHA)
	LUAGi(GL_OPERAND6_ALPHA)
	LUAGi(GL_OPERAND7_ALPHA)
#endif
	LUAGi(GL_DOT3_RGB)
	LUAGi(GL_DOT3_RGBA)
	LUAGi(GL_TRANSPOSE_MODELVIEW_MATRIX)
	LUAGi(GL_TRANSPOSE_PROJECTION_MATRIX)
	LUAGi(GL_TRANSPOSE_TEXTURE_MATRIX)
	LUAGi(GL_TRANSPOSE_COLOR_MATRIX)
	LUAGi(GL_NORMAL_MAP)
	LUAGi(GL_REFLECTION_MAP)
	LUAGi(GL_TEXTURE_CUBE_MAP)
	LUAGi(GL_TEXTURE_BINDING_CUBE_MAP)
	LUAGi(GL_TEXTURE_CUBE_MAP_POSITIVE_X)
	LUAGi(GL_TEXTURE_CUBE_MAP_NEGATIVE_X)
	LUAGi(GL_TEXTURE_CUBE_MAP_POSITIVE_Y)
	LUAGi(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y)
	LUAGi(GL_TEXTURE_CUBE_MAP_POSITIVE_Z)
	LUAGi(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z)
	LUAGi(GL_PROXY_TEXTURE_CUBE_MAP)
	LUAGi(GL_MAX_CUBE_MAP_TEXTURE_SIZE)
	LUAGi(GL_COMPRESSED_ALPHA)
	LUAGi(GL_COMPRESSED_LUMINANCE)
	LUAGi(GL_COMPRESSED_LUMINANCE_ALPHA)
	LUAGi(GL_COMPRESSED_INTENSITY)
	LUAGi(GL_COMPRESSED_RGB)
	LUAGi(GL_COMPRESSED_RGBA)
	LUAGi(GL_TEXTURE_COMPRESSION_HINT)
	LUAGi(GL_TEXTURE_COMPRESSED_IMAGE_SIZE)
	LUAGi(GL_TEXTURE_COMPRESSED)
	LUAGi(GL_NUM_COMPRESSED_TEXTURE_FORMATS)
	LUAGi(GL_COMPRESSED_TEXTURE_FORMATS)
	LUAGi(GL_MULTISAMPLE)
	LUAGi(GL_SAMPLE_ALPHA_TO_COVERAGE)
	LUAGi(GL_SAMPLE_ALPHA_TO_ONE)
	LUAGi(GL_SAMPLE_COVERAGE)
	LUAGi(GL_SAMPLE_BUFFERS)
	LUAGi(GL_SAMPLES)
	LUAGi(GL_SAMPLE_COVERAGE_VALUE)
	LUAGi(GL_SAMPLE_COVERAGE_INVERT)
	LUAGi(GL_MULTISAMPLE_BIT)
	LUAGi(GL_DEPTH_COMPONENT16)
	LUAGi(GL_DEPTH_COMPONENT24)
	LUAGi(GL_DEPTH_COMPONENT32)
	LUAGi(GL_TEXTURE_DEPTH_SIZE)
	LUAGi(GL_DEPTH_TEXTURE_MODE)
	LUAGi(GL_TEXTURE_COMPARE_MODE)
	LUAGi(GL_TEXTURE_COMPARE_FUNC)
	LUAGi(GL_COMPARE_R_TO_TEXTURE)
	LUAGi(GL_QUERY_COUNTER_BITS)
	LUAGi(GL_CURRENT_QUERY)
	LUAGi(GL_QUERY_RESULT)
	LUAGi(GL_QUERY_RESULT_AVAILABLE)
	LUAGi(GL_SAMPLES_PASSED)
	LUAGi(GL_FOG_COORD_SRC)
	LUAGi(GL_FOG_COORD)
	LUAGi(GL_FRAGMENT_DEPTH)
	LUAGi(GL_CURRENT_FOG_COORD)
	LUAGi(GL_FOG_COORD_ARRAY_TYPE)
	LUAGi(GL_FOG_COORD_ARRAY_STRIDE)
	LUAGi(GL_FOG_COORD_ARRAY_POINTER)
	LUAGi(GL_FOG_COORD_ARRAY)
	LUAGi(GL_FOG_COORDINATE_SOURCE)
	LUAGi(GL_FOG_COORDINATE)
	LUAGi(GL_CURRENT_FOG_COORDINATE)
	LUAGi(GL_FOG_COORDINATE_ARRAY_TYPE)
	LUAGi(GL_FOG_COORDINATE_ARRAY_STRIDE)
	LUAGi(GL_FOG_COORDINATE_ARRAY_POINTER)
	LUAGi(GL_FOG_COORDINATE_ARRAY)
	LUAGi(GL_COLOR_SUM)
	LUAGi(GL_CURRENT_SECONDARY_COLOR)
	LUAGi(GL_SECONDARY_COLOR_ARRAY_SIZE)
	LUAGi(GL_SECONDARY_COLOR_ARRAY_TYPE)
	LUAGi(GL_SECONDARY_COLOR_ARRAY_STRIDE)
	LUAGi(GL_SECONDARY_COLOR_ARRAY_POINTER)
	LUAGi(GL_SECONDARY_COLOR_ARRAY)
	LUAGi(GL_POINT_SIZE_MIN)
	LUAGi(GL_POINT_SIZE_MAX)
	LUAGi(GL_POINT_FADE_THRESHOLD_SIZE)
	LUAGi(GL_POINT_DISTANCE_ATTENUATION)
	LUAGi(GL_BLEND_DST_RGB)
	LUAGi(GL_BLEND_SRC_RGB)
	LUAGi(GL_BLEND_DST_ALPHA)
	LUAGi(GL_BLEND_SRC_ALPHA)
	LUAGi(GL_GENERATE_MIPMAP)
	LUAGi(GL_GENERATE_MIPMAP_HINT)
	LUAGi(GL_INCR_WRAP)
	LUAGi(GL_DECR_WRAP)
	LUAGi(GL_MIRRORED_REPEAT)
	LUAGi(GL_MAX_TEXTURE_LOD_BIAS)
	LUAGi(GL_TEXTURE_FILTER_CONTROL)
	LUAGi(GL_TEXTURE_LOD_BIAS)
	LUAGi(GL_ARRAY_BUFFER)
	LUAGi(GL_ELEMENT_ARRAY_BUFFER)
	LUAGi(GL_ARRAY_BUFFER_BINDING)
	LUAGi(GL_ELEMENT_ARRAY_BUFFER_BINDING)
	LUAGi(GL_VERTEX_ARRAY_BUFFER_BINDING)
	LUAGi(GL_NORMAL_ARRAY_BUFFER_BINDING)
	LUAGi(GL_COLOR_ARRAY_BUFFER_BINDING)
	LUAGi(GL_INDEX_ARRAY_BUFFER_BINDING)
	LUAGi(GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING)
	LUAGi(GL_EDGE_FLAG_ARRAY_BUFFER_BINDING)
	LUAGi(GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING)
	LUAGi(GL_FOG_COORD_ARRAY_BUFFER_BINDING)
	LUAGi(GL_WEIGHT_ARRAY_BUFFER_BINDING)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING)
	LUAGi(GL_STREAM_DRAW)
	LUAGi(GL_STREAM_READ)
	LUAGi(GL_STREAM_COPY)
	LUAGi(GL_STATIC_DRAW)
	LUAGi(GL_STATIC_READ)
	LUAGi(GL_STATIC_COPY)
	LUAGi(GL_DYNAMIC_DRAW)
	LUAGi(GL_DYNAMIC_READ)
	LUAGi(GL_DYNAMIC_COPY)
	LUAGi(GL_READ_ONLY)
	LUAGi(GL_WRITE_ONLY)
	LUAGi(GL_READ_WRITE)
	LUAGi(GL_BUFFER_SIZE)
	LUAGi(GL_BUFFER_USAGE)
	LUAGi(GL_BUFFER_ACCESS)
	LUAGi(GL_BUFFER_MAPPED)
	LUAGi(GL_BUFFER_MAP_POINTER)
	LUAGi(GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING)
	LUAGi(GL_CURRENT_PROGRAM)
	LUAGi(GL_SHADER_TYPE)
	LUAGi(GL_DELETE_STATUS)
	LUAGi(GL_COMPILE_STATUS)
	LUAGi(GL_LINK_STATUS)
	LUAGi(GL_VALIDATE_STATUS)
	LUAGi(GL_INFO_LOG_LENGTH)
	LUAGi(GL_ATTACHED_SHADERS)
	LUAGi(GL_ACTIVE_UNIFORMS)
	LUAGi(GL_ACTIVE_UNIFORM_MAX_LENGTH)
	LUAGi(GL_SHADER_SOURCE_LENGTH)
	LUAGi(GL_FLOAT_VEC2)
	LUAGi(GL_FLOAT_VEC3)
	LUAGi(GL_FLOAT_VEC4)
	LUAGi(GL_INT_VEC2)
	LUAGi(GL_INT_VEC3)
	LUAGi(GL_INT_VEC4)
	LUAGi(GL_BOOL)
	LUAGi(GL_BOOL_VEC2)
	LUAGi(GL_BOOL_VEC3)
	LUAGi(GL_BOOL_VEC4)
	LUAGi(GL_FLOAT_MAT2)
	LUAGi(GL_FLOAT_MAT3)
	LUAGi(GL_FLOAT_MAT4)
	LUAGi(GL_SAMPLER_1D)
	LUAGi(GL_SAMPLER_2D)
	LUAGi(GL_SAMPLER_3D)
	LUAGi(GL_SAMPLER_CUBE)
	LUAGi(GL_SAMPLER_1D_SHADOW)
	LUAGi(GL_SAMPLER_2D_SHADOW)
	LUAGi(GL_SHADING_LANGUAGE_VERSION)
	LUAGi(GL_VERTEX_SHADER)
	LUAGi(GL_MAX_VERTEX_UNIFORM_COMPONENTS)
	LUAGi(GL_MAX_VARYING_FLOATS)
	LUAGi(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS)
	LUAGi(GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS)
	LUAGi(GL_ACTIVE_ATTRIBUTES)
	LUAGi(GL_ACTIVE_ATTRIBUTE_MAX_LENGTH)
	LUAGi(GL_FRAGMENT_SHADER)
	LUAGi(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS)
	LUAGi(GL_FRAGMENT_SHADER_DERIVATIVE_HINT)
	LUAGi(GL_MAX_VERTEX_ATTRIBS)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_ENABLED)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_SIZE)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_STRIDE)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_TYPE)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_NORMALIZED)
	LUAGi(GL_CURRENT_VERTEX_ATTRIB)
	LUAGi(GL_VERTEX_ATTRIB_ARRAY_POINTER)
	LUAGi(GL_VERTEX_PROGRAM_POINT_SIZE)
	LUAGi(GL_VERTEX_PROGRAM_TWO_SIDE)
	LUAGi(GL_MAX_TEXTURE_COORDS)
	LUAGi(GL_MAX_TEXTURE_IMAGE_UNITS)
	LUAGi(GL_MAX_DRAW_BUFFERS)
	LUAGi(GL_DRAW_BUFFER0)
	LUAGi(GL_DRAW_BUFFER1)
	LUAGi(GL_DRAW_BUFFER2)
	LUAGi(GL_DRAW_BUFFER3)
	LUAGi(GL_DRAW_BUFFER4)
	LUAGi(GL_DRAW_BUFFER5)
	LUAGi(GL_DRAW_BUFFER6)
	LUAGi(GL_DRAW_BUFFER7)
	LUAGi(GL_DRAW_BUFFER8)
	LUAGi(GL_DRAW_BUFFER9)
	LUAGi(GL_DRAW_BUFFER10)
	LUAGi(GL_DRAW_BUFFER11)
	LUAGi(GL_DRAW_BUFFER12)
	LUAGi(GL_DRAW_BUFFER13)
	LUAGi(GL_DRAW_BUFFER14)
	LUAGi(GL_DRAW_BUFFER15)
	LUAGi(GL_POINT_SPRITE)
	LUAGi(GL_COORD_REPLACE)
	LUAGi(GL_POINT_SPRITE_COORD_ORIGIN)
	LUAGi(GL_LOWER_LEFT)
	LUAGi(GL_UPPER_LEFT)
	LUAGi(GL_STENCIL_BACK_FUNC)
	LUAGi(GL_STENCIL_BACK_VALUE_MASK)
	LUAGi(GL_STENCIL_BACK_REF)
	LUAGi(GL_STENCIL_BACK_FAIL)
	LUAGi(GL_STENCIL_BACK_PASS_DEPTH_FAIL)
	LUAGi(GL_STENCIL_BACK_PASS_DEPTH_PASS)
	LUAGi(GL_STENCIL_BACK_WRITEMASK)
	
	//apple special
	//parallels the ati and arb specials as well
	LUAGiAPPLE(GL_RGBA_FLOAT32)
	LUAGiAPPLE(GL_RGB_FLOAT32)
	LUAGiAPPLE(GL_ALPHA_FLOAT32)
	LUAGiAPPLE(GL_INTENSITY_FLOAT32)
	LUAGiAPPLE(GL_LUMINANCE_FLOAT32)

	lua_pop(L,1);
}



void luaBind(lua_State *L) {
	luaBindGL(L);						//gl bindings
	luaBindGL_ARB_vertex_shader(L);
	luaBindGL_ARB_shader_objects(L);
	luaBindGL_ARB_vertex_program(L);
	luaBindGL_EXT_framebuffer_object(L);
	luaBindGL_ARB_vertex_buffer_object(L);
	luaBind_Util(L);
}
