/*

    File: main.cpp

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <assert.h>
#include <iostream>

#include "math/vec.h"
#include "system/singletons.h"
#include "system/exception.h"
#include "system/subsystem.h"
#include "base/opengl.h"
#include "base/image.h"
#include "shader/shaderext.h"
#include "main.h"

#ifdef WIN32
#define strcasecmp _stricmp
#endif

using namespace std;

//statics that everyone gets to see
ProgSys progSys;

int clearFlags = GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT;
quatf sceneRot(0,0,0,1);		//object rotation, not camera rotation
vec3f scenePos(0,0,5);		//camera distance (so its inverse transformed)
float sceneLogScale = 0.f;		//object scale.  used for ortho scenes
bool sceneOrtho = false;
bool menuInputResponse = false;
static bool allowSpecialKeys = true;

static int PF_clearFlags(lua_State *L) {
	int old = clearFlags;
	int argc = lua_gettop(L);
	if (argc > 0) {
		clearFlags = (int)lua_tonumber(L,1);
	}
	lua_pushinteger(L, old);
	return 1;
}

//todo - make this a table with a meta that calls the real vars
static int PF_mouse(lua_State *L) {
	lua_pushnumber(L, mouse.getFPos().x);
	lua_pushnumber(L, mouse.getFPos().y);
	int b = 0;
	if (mouse.getButtonState(mouse.BUTTON_LEFT) == BUTTON_STATE_DOWN) b |= 1;
	if (mouse.getButtonState(mouse.BUTTON_RIGHT) == BUTTON_STATE_DOWN) b |= 2;
	if (mouse.getButtonState(mouse.BUTTON_MIDDLE) == BUTTON_STATE_DOWN) b |= 4;
	lua_pushnumber(L, b);
	lua_pushnumber(L, mouse.getDZ());
	return 4;
}

// takes in a string or something
// poops out a bitflag for whether it's set or not
static int PF_kbd(lua_State *L) {
	if (!lua_gettop(L)) return 0;
	const char *name = lua_tostring(L, 1);
	//map plz?
	for (int i = 0; i < numberof(Key::name); i++) {
		if (!strcasecmp(name, Key::name[i])) {
			lua_pushboolean(L, kbd.getKeyState(i) & BUTTON_STATE_DOWN);
			return 1;
		}
	}
	return 0;
}

static int PF_timer(lua_State *L) {
	lua_pushnumber(L, timer.getTime());
	return 1;
}

static int PF_dispSize(lua_State *L) {
	lua_pushinteger(L, disp.getSize().x);
	lua_pushinteger(L, disp.getSize().y);
	return 2;
}

static int PF_exit(lua_State *L) {
	sys.requestExit();
	return 0;
}

static int PF_ortho(lua_State *L) {
	bool oldOrtho = sceneOrtho;
	if (lua_gettop(L)) {
		sceneOrtho = lua_toboolean(L, 1);
	}
	lua_pushboolean(L, oldOrtho);
	return 1;
}

static int PF_menuInputResponse(lua_State *L) {
	lua_pushboolean(L, menuInputResponse);
	return 1;
}

static int PF_allowSpecialKeys(lua_State *L) {
	bool oldAllowGlushKeys = allowSpecialKeys;
	if (lua_gettop(L)) allowSpecialKeys = lua_toboolean(L, 1);
	lua_pushboolean(L, oldAllowGlushKeys);
	return 1;
}

static const luaL_reg mainGlueFuncs[] = {
	{"clearFlags", PF_clearFlags},
	{"mouse", PF_mouse},
	{"kbd", PF_kbd},
	{"timer", PF_timer},
	{"dispSize", PF_dispSize},
	{"exit", PF_exit},
	{"ortho", PF_ortho},
	{"menuInputResponse", PF_menuInputResponse},
	{"allowSpecialKeys", PF_allowSpecialKeys},
	{NULL, NULL}
};

//my fix for the time being on letting the user get access to these:
void luaStateAddMainCalls(lua_State *L) {
	lua_getglobal(L, "glush");
	luaAddBindings(L, mainGlueFuncs);
	lua_pop(L,1);
}

void GlushModule::displayFunc() {
	//default unless otherwise specified
	if (clearFlags) glClear(clearFlags);

	//this is doing bad things ...
	//even when we use our own lua state, and not a passed-in one
	lua_State *L = progSys.getState();
	if (L) {
		bool ran = false;
		int top = lua_gettop(L);
		lua_getglobal(L, "glush");
		lua_getfield(L, -1, "display");
		if (lua_isfunction(L, -1)) {
			if (!luaSafeCall(L, 0, 0)) {
				con << "lua_pcall got error during GlushModule::displayFunc running glush.display" << endl;
				lua_pop(L, 1);	//pop the error
			} else {
				ran = true;
			}
		}
		lua_settop(L, top);
		if (!ran) {
			progSys.kill();
		}
	}
	
	//manual swap buffers ... hmm ...
}

void GlushModule::initCore() {

	//it would be nice to put some override in our script
	//this could easily be done if we had a sys.resize function ...
	//sys.initSize.set(512,512);
	sys.setInitSize(vec2i(800,600));
	
	//if you feel like it
	//run the initial script before running this
	//(which means setting up lua beforehand)
	//(and might run into some circular definition issues)
	//and in that determine whethr we should start single or double buffered / init size / etc
	super::init();
	
	//todo - put this in base?
#ifdef WIN32
	cvarSys->exec("log");
#endif

	base.clearAllViewports();
	
	base.views[VIEW_3D].setPos(vec3f(0,0,0));
	
	//do this before progSys init
	glushMenu.init();
	if (!shaderSystem.init()) con << "Shader System failed to init";

	/*
	init string ...
	rather than loading flame/flame or radiosity every time within the hardcoded
	how about something like init.lua?
	and from there we can redirect it (config-file-style)
	*/
}

void GlushModule::initWithGoods(lua_State *L) {
	initCore();
	progSys.initWithGoods(L);
}

void GlushModule::init() {	
	initCore();
	initProgSys();
}

void GlushModule::initProgSys() {	
	string init = "init.lua";	//default
//	string extra = "";
	vector<string> args;
	if (sys.getArgCount() > 1) {
		string param = sys.getArg(1);
		if (param.length() > 4) {
			string ext = param.substr(param.length()-4, 4);
			if (ext != ".lua") con << "warning: expected .lua extension but i got a " << ext << "\n";
			//if its a .lua file then clip the end and use that as our require() param
			//else param = param.substr(0, param.length()-4);
			init = param;
		}
		for (size_t j = 2; j < sys.getArgCount(); j++) {
			args.push_back(sys.getArg(j));
		}
	}
	progSys.init(init, args);
}

void GlushModule::reset() {
	progSys.kill();
	glushMenu.shutdown();
	glushMenu.init();
	initProgSys();
}

	
void GlushModule::shutdown() {
	progSys.kill();
	//shaderSystem.shutdown();	//no shutdown eh?
	glushMenu.shutdown();
	
	super::shutdown();
}

#ifdef WINSCREENSAVERSYSTEM
static bool menuOpen = false;
#else
static bool menuOpen = true;
#endif

static float zNear = .1f;
static float zFar = 100.f;
static float trackScale = 1.f;
static float distMoveScale = 0.25f;
static float panMoveScale = 25.f;

static CVarw cv_zNear(&zNear, "zNear");
static CVarw cv_zFar(&zFar, "zFar");	
static CVarw cv_trackScale(&trackScale, "trackScale");
static CVarw cv_distMoveScale(&distMoveScale, "distMoveScale");
static CVarw cv_panMoveScale(&panMoveScale, "panMoveScale");

void GlushModule::update() {

	//lua mt update
	progSys.updateThreads();

	//resize code:
	glViewport(0,0,VEC2ELEM(disp.getSize()));
	
	base.views[VIEW_3D].setupView(
		bitflag(VIEW_FLAG_CLEAR_PROJECTION) |
		bitflag(VIEW_FLAG_APPLY_PROJECTION),
		(double)disp.getSize().y / (double)disp.getSize().x);
	
	float aspectRatio = (float)disp.getSize().x / (float)disp.getSize().y;
	if (sceneOrtho) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-aspectRatio*.5, aspectRatio*.5, -.5, .5, -1, 1);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		float es = exp(sceneLogScale);
		glScalef(es,es,es);
		glTranslatef(-scenePos.x,-scenePos.y,-scenePos.z);
	} else {
		//projection is already set up from the underlying framework
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(-scenePos.x,-scenePos.y,-scenePos.z);
		vec4f aa = quatToAngleAxis(sceneRot);
		glRotatef(aa.w, VEC3ELEM(aa));
		//TODO - zoom focus rather than scale scene in non-ortho
		//otherwise scale wont work well with translating to recenter the scene
		float es = exp(sceneLogScale);
		glScalef(es,es,es);
	}
	
	//render code
	displayFunc();
	
	if (kbd.keystate[Key::ESC] == BUTTON_STATE_PRESS && allowSpecialKeys) {
		menuOpen = !menuOpen;
	}
	
	//TODO - glDelete all the textures, shaders, lists, and whatever else you can think of in reset()
	//or else they will leak here
	if (kbd.keystate[Key::F2] == BUTTON_STATE_PRESS && allowSpecialKeys) {
		reset();
	}
	
	//if menu isnt open then display the menu
	if (menuOpen) glushMenu.display();
	//if menu is closed then do the other input.
	//else... if menu sys said we didnt get anything then do the other input
	if (!menuOpen || !(menuInputResponse = glushMenu.getInput())) {
		//if we got no menu input then do the other input

		//how should our controls work?
		//no modifiers <=> rotate
		//something <=> translate xy
		//something <=> translate z
		//something <=> scale (if you really don't like z translation that much ... )

		//trackball code
		if (mouse.getMove().x || mouse.getMove().y) {
			if (mouse.getButtonState(mouse.BUTTON_LEFT)) {
				if (sceneOrtho) {
					scenePos.x -= aspectRatio * mouse.getFMove().x * exp(-sceneLogScale);
					scenePos.y -= mouse.getFMove().y * exp(-sceneLogScale);
				} else {
					if (kbd.keystate[Key::LSHIFT] || kbd.keystate[Key::RSHIFT]) {
						scenePos.z += distMoveScale * (float)mouse.getMove().y;
					} else {
						//cross product of yz screenspace movement
						//save the magnitude in the w component
						quatf aa(-mouse.getMove().y, mouse.getMove().x, 0, mouse.getMove().len());
						//normalize axis
						aa.axis() /= aa.w;
						//scale angle
						aa.w *= trackScale;
						//convert to quat space and append
						sceneRot.leftMul(angleAxisToQuat(aa)).normalize();
					}
				}
			} else if (mouse.getButtonState(mouse.BUTTON_RIGHT)) {
				if (sceneOrtho) {
					sceneLogScale += mouse.getFMove().y;
				} else {
					if (kbd.keystate[Key::LSHIFT] || kbd.keystate[Key::RSHIFT]) {
						sceneLogScale += mouse.getFMove().y;
					} else {
						scenePos.x -= panMoveScale * mouse.getFMove().x;
						scenePos.y -= panMoveScale * mouse.getFMove().y;
					}
				}
			}
		}
	}
	
	//for dbuf this runs the appropriate swapbuffers call
	//sys.updateDisplayBuffers();
	//seems this gets called already.  probably by whatever was gonna call GlushModule::render()
}
	
void GlushModule::render(const Viewport *vp, const View *v) {
	//doing it manually for now - so dont go here
	assert(false);
}

GlushModule glushModule;
BaseModule *RegisterApplication() { return &glushModule; }

//gl
//glu
//glui	<- hacked my own (not very oop was it?)
//glut	<- hacked my own (shift,ctrl,alt had bad read key state code)
//glew	<- hacked my own (dunno why ... i bet glew is better than my code)
//lua
//=
//glutwaia
