/*

    File: menu.cpp

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <cassert>
#include <string>
#include <sstream>
#include "shader/shaderext.h"
#include "system/exception.h"
#include "system/system.h"
#include "system/singletons.h"
#include "base/main.h"
#include "menu/sys.h"
#include "menu/lib.h"	//menu
#include "menu.h"
#include "menu/luabindings.h"
#include "gc/gc.h"

#include "system/subsystem.h"

#ifdef WINSYSTEM
#include "winsystem/wstring.h"
#include "winsystem/system.h"
#include "winsystem/singletons.h"
#endif

#ifdef WINSYSTEM
#define WIN32_ALLOW_DRAG_AND_DROP
#endif

using namespace std;
using namespace pwn;

#ifdef WIN32_ALLOW_DRAG_AND_DROP
static LRESULT WINAPI DropFilesProc(int nCode, WPARAM wParam, LPARAM lParam);
static HHOOK dropFilesHookHandle;
static HWND dragDropWnd;
#endif

class MenuTex2D;
class GlushMenu : public IGlushMenu {
public:
	class RootResizeListener : public ResizeListener {
		GlushMenu *owner;
	public:
		RootResizeListener(GlushMenu *_owner) : owner(_owner) {}
		virtual void callback_onResize(Object *o) {
			Object *area = owner->area;
			owner->area->setSize(vec2f(owner->area->getSize().x, menuSys.getSize().y-2));
			Object *vert = owner->area->getVert();
			vert->setPos(vec2f(0,0));
			Object *horz = owner->area->getHorz();
			horz->setVisible(false);
		}
	};

	vec2f row;
	vector<MenuTex2D*> menuTex2Ds;

	RootResizeListener rootResizeListener;

	gc::GlobalPointer<ScrollArea> area;
	gc::GlobalPointer<ScrollContainer> container;
	
	GlushMenu()
	:	area(gc::sys->createPtr<ScrollArea>()),
		container(gc::sys->createPtr<ScrollContainer>()),
		rootResizeListener(this)
	{}

	virtual void init() {
		row = vec2f(2,1);	//last good menu position
	
		menuSys.init();
		menuSys.getRoot()->setScale(vec2f(12.f));
		
		container = new ScrollContainer(vec2f(0,0), vec2f(14, 1));
		container->setBorderColor(0);
		container->setBackgroundColor(0);
		area = new ScrollArea(vec2f(1,1), vec2f(14, menuSys.getSize().y-2), container);
		area->setBorderColor(0);
		area->setBackgroundColor(0);
		
		/*
		should listeners also be smart-ptr'd?
		should they have dtor code that unlinks them from all their added listeners?
		otherwise if they get deleted and the object still persist... there could be trouble?
		but doing so would mean holding 
		*/
		menuSys.getRoot()->addResizeListener(&rootResizeListener);
		menuSys.getRoot()->addChild(area);
		rootResizeListener.callback_onResize(menuSys.getRoot());	//init resize

		//now create a base menu for all else to attach to
		//make it a scroll pane and put the scrollbar on the left side please...

#ifdef WIN32_ALLOW_DRAG_AND_DROP
		try {
			//init drag and drop
#if 1	//TODO - only if we went through the win system ...
			dragDropWnd = winSys.getHWnd();
#else	//else .. glut?
			const char *windowName = "GLUT Window";
			dragDropWnd = FindWindow(NULL, str2wstr(windowName).c_str());
			if (!dragDropWnd) throw Exception() << "couldn't find window " << windowName;
#endif
			DragAcceptFiles(dragDropWnd, true);
			dropFilesHookHandle = SetWindowsHookEx(WH_GETMESSAGE, DropFilesProc, (HINSTANCE)NULL, GetCurrentThreadId());
			if (!dropFilesHookHandle) throw Exception() << "SetWindowsHookEx failed!";
		} catch (Exception &t) {
			throw Exception() << "GlushMenu.init failed: " << t;
		}
#endif //WIN32_ALLOW_DRAG_AND_DROP
	}
	
	virtual void shutdown() {
		menuSys.getRoot()->removeResizeListener(&rootResizeListener);
		menuSys.shutdown();
		menuTex2Ds.resize(0);
	}
	
	virtual void bind(lua_State *L);
	
	virtual void display() {
		menuSys.display();
	}

	virtual bool getInput() { 
		return menuSys.getInput();
	}
	
	void addChild(Object *o) {
		container->addChild(o);
		vec2f maxpos = o->getPos() + o->getSize();
		if (container->getSize().y < maxpos.y) {
			container->setSize( vec2f( container->getSize().x, maxpos.y) );
		}
		rootResizeListener.callback_onResize(menuSys.getRoot());	//manual resize
	}
};
static GlushMenu g_glushMenu;
IGlushMenu &glushMenu = g_glushMenu;


class MenuTex2D : public Object {
	typedef Object super;
protected:
	lua_State *L;
	string varname;
public:
	MenuTex2D(const vec2f &_pos, lua_State *_L, const char *_varname) 
		: super(_pos, vec2f(3.f)), L(_L), varname(_varname) 
	{
		backgroundColor = 0xBFFFFFFF;	//white yo
		lua_getglobal(L, varname.c_str());
		textureID = (int)lua_tonumber(L,-1);
		lua_pop(L,1);
	}

	virtual void getDropFile(const char *filename) {
		int top = lua_gettop(L);
		Image *img = imageSystem.load(filename);
		if (!img) throw Exception() << "couldnt load the image file " << filename;

		assert(img->data);
		int bytesPP = img->bitsPerPixel >> 3;	//all we have is 24 or 32 bitsPerPixel atm
		assert(!(img->bitsPerPixel & 7));		//no <8 bit images

		//first we get the value of our variable
		//see if its a texture
		//if its not then generate it
		lua_getglobal(L, varname.c_str());
		GLuint texID = (int)lua_tonumber(L,-1);
		lua_pop(L,1);
		if (!glIsTexture(texID)) {
			glGenTextures(1, &texID);
			lua_pushinteger(L, texID);
			lua_setglobal(L, varname.c_str());
		}

		//from there we upload the junk
		//in the junk
		const int target = GL_TEXTURE_2D;
		glActiveTextureARB(GL_TEXTURE0);		//does the active unit matter?
		glBindTexture(target, texID);
		gluBuild2DMipmaps(
			target,
			bytesPP == 4 ? 4 : 3,
			VEC2ELEM(img->size),			//auto-adjusts up a power of two, does the proxy testing and everything
			bytesPP == 4 ? GL_RGBA : GL_RGB,
			GL_UNSIGNED_BYTE,
			img->data);

		lua_settop(L,top);
		//no other defaults plz
		delete img;
	}
};

//win32 drag and drop support
#ifdef WIN32_ALLOW_DRAG_AND_DROP

static LRESULT WINAPI DropFilesProc(int nCode, WPARAM wParam, LPARAM lParam)  {
	MSG *msg = (MSG *)lParam;
	if (msg->message == WM_DROPFILES) {
//		mainGui->processDropMsg(msg);
		con << " got a drop!\n";
		DWORD lpos = GetMessagePos();
		POINT point;
		point.x = LOWORD(lpos);
		point.y = HIWORD(lpos);
		ScreenToClient(dragDropWnd, &point);
		vec2f mpos((float)point.x / (float)disp.getSize().x,
					(float)point.y / (float)disp.getSize().y);
		con << " mouse coords: " << mpos << "\n";
		con << " mouse obj coords: " << mouse.getFPos() << "\n";
		mpos.x *= menuSys.getSize().x;
		mpos.y *= menuSys.getSize().y;
		con << "scaled: " << mpos << endl;
		//now cycle through all our MenuTex2D objects ... check their rel pos ... do the load ... 
		for (int i = 0; i < g_glushMenu.menuTex2Ds.size(); i++) {
			MenuTex2D *button = g_glushMenu.menuTex2Ds[i];
			box2f bbox;
			bbox.min = button->getAbsPos();
			bbox.max = bbox.min + button->getSize();
			con << "box min: " << bbox.min << " max: " << bbox.max << endl;
			if (!bbox.contains(mpos)) continue;
			con << "button " << i << " got the drop!" << endl;

			//extract the filename
			wchar_t wfilename[256];
			{
				HDROP hDrop = (HDROP)msg->wParam;
				int fileCount = DragQueryFile(hDrop, 0xFFFFFFFF, 0, 0);
				for (int i = 0; i < fileCount; i++) {
					memset(wfilename, 0, sizeof(wfilename));
					DragQueryFile(hDrop, i, wfilename, sizeof(wfilename));
					break;
				}
			}
			DragFinish((HDROP)msg->wParam);

			button->getDropFile(wstr2str(wfilename).c_str());
		}
	}
	return CallNextHookEx(dropFilesHookHandle, nCode, wParam, lParam);
}
#endif //WIN32_ALLOW_DRAG_AND_DROP

class MyMenuCheckBox : public Checkbox {
	typedef Checkbox super;
protected:
	lua_State *L;
	string varname;
public:
	
	MyMenuCheckBox(const vec2f &_pos, string _name, bool _checked, lua_State *_L, string _varname)
	: super(_pos,_name.c_str(),_checked), L(_L), varname(_varname) {
		update();
	}
	
	virtual void setChecked(bool ch) {
		super::setChecked(ch);
		update();
	}
	
	void update() {
		lua_pushboolean(L, checked);
		lua_setglobal(L, varname.c_str());
	}
};

LUAFstart(glushgui_checkbox)
	try {
		if (argc < 1) throw Exception() << "params: variable name, [variable state]";
		string name = lua_tostring(L, 1);
		bool checked = lua_toboolean(L, 2);
		
		MyMenuCheckBox *cb = new MyMenuCheckBox(g_glushMenu.row, name, checked, L, name);
		if (!cb) throw Exception() << " couldnt create checkbox " << name;
		g_glushMenu.addChild(cb);
		g_glushMenu.row.y += 1.6f;
		
	} catch (Exception &t) {
		con << __FUNCTION__ << ": " << t << "\n";
	}
LUAFend()

class MyMenuButton : public Text {
	typedef Text super;
protected:
	lua_State *L;
	string varname;
public:
	
	MyMenuButton(const vec2f &_pos, string _name, lua_State *_L, string _varname)
	: super(_pos,_name.c_str()), L(_L), varname(_varname) {}
	
	virtual eventResponse_t mouseEvent(MouseEvent event, const vec2f &v) {
		if (event.lpress) {
			//then attempt to invoke the global function with our varname
			//TODO - provide a function object rather than global lookup
			//just like incorporating the state into the objects rather than writing back to a variable
			lua_getglobal(L, varname.c_str());
			if (!lua_isfunction(L, -1)) {
				con << "MenuButton:mouseEvent expected a global function of name " << varname << endl;
			} else {
				if (!luaSafeCall(L, 0, 0)) {
					con << "lua_pcall got error during ProgSys::runBase" << endl;
					lua_pop(L, 1);	//pop the error
				}
			}
			lua_pop(L,1);
		}
		return RESPONSE_STOP;
	}
};

LUAFstart(glushgui_button)
	try {
		if (argc < 1) throw Exception() << "params: function name";
		string name = lua_tostring(L, 1);
		
		MyMenuButton *btn = new MyMenuButton(g_glushMenu.row, name, L, name);
		if (!btn) throw Exception() << " couldnt create checkbox " << name;
		g_glushMenu.addChild(btn);
		g_glushMenu.row.y += 1.6f;
		
	} catch (Exception &t) {
		con << __FUNCTION__ << ": " << t << "\n";
	}
LUAFend()

class MenuScrollBarText : public ScrollbarListener, public TextListener {
protected:
	lua_State *L;
	string varname;
	float value;
	
	gc::GlobalPointer<Scrollbar> scrollbar;
	gc::GlobalPointer<TextField> text;
public:
	MenuScrollBarText(lua_State *_L, const char *_varname, float _value, float vmin, float vmax) 
	:	L(_L), varname(_varname), value(_value),
		scrollbar(gc::sys->createPtr<Scrollbar>()),
		text(gc::sys->createPtr<TextField>())
	{
	
		//we don't want to go by the value's value
		//because that is supposed to be set by the 3rd param
		//lua_getglobal(L, varname.c_str());
		//double val = lua_tonumber(L,-1);
		//lua_pop(L,1);
		double val = value;

		Text *t = new Text(g_glushMenu.row, varname.c_str());
		t->setBorderColor(0);
		g_glushMenu.addChild(t);
		g_glushMenu.row.y += 1.1f;
		
		//add the scrollbar
		scrollbar = new Scrollbar(g_glushMenu.row, 10, true, vmin, vmax, val);
		scrollbar->addScrollbarListener(this);
		g_glushMenu.addChild(scrollbar);
		g_glushMenu.row.y += 1.1f;

		//add the text
		text = new TextField(g_glushMenu.row, 10);
		text->addTextListener(this);
		ostringstream s;
		s << val;
		text->setText(s.str());
		g_glushMenu.addChild(text);

		g_glushMenu.row.y += 1.6f;
		
		//update lua to our slider variable
		update();
	}
	
	virtual void callback_onSliderChange(Scrollbar* sb) {
		assert(sb == scrollbar);

		value = sb->getSliderPos();
		update();
		
		//inf regress detect plz ...
		//but wait, writing directl to the text wont call the callback
		//which is a big security hole
		//why am i using static arrays anyways
		
		if (atof(text->getText().c_str()) != value) {
			ostringstream s;
			s << value;
			text->setText(s.str());
#if 0	//trimming is sucking it up ... i think
			//if theres a decimal then right-trim it
			bool found = false;
			for (char *ch = text->text; *ch; ch++) {
				if (*ch == '.') {
					found = true;
					break;
				}
			}
			if (found) {
				for (char *ch = text->text + strlen(text->text) - 1; ch >= text->text; ch--) {
					if (*ch != '0') break;
					*ch = '\0';	//change all trailing '0' to '\0'
				}
			}
#endif
		}
	}
	
	virtual void callback_onTextChange(TextField* t) {
		assert(t == text);
		float svalue = atof(t->getText().c_str());

		//update scroll pos without updating lua variable
		//i think theres inf regress detect in there
		//lets see
		if (scrollbar->getSliderPos() != svalue) {
			scrollbar->setSliderPos(svalue);
		}
		
		//and a final text update in case the value is outside the slider range
		value = svalue;
		ostringstream o;
		o << value;
		text->setText(o.str());
		update();
	}
	
	void update() {
		lua_pushnumber(L, value);
		lua_setglobal(L, varname.c_str());
	}
};

LUAFstart(glushgui_slider)
	try {
		if (argc < 4) throw Exception() << "params: variable name, minimum, value, maximum scroll ranges";
		const char *name = lua_tostring(L,1);
		float vmin = (float)lua_tonumber(L,2);
		float value = (float)lua_tonumber(L,3);
		float vmax = (float)lua_tonumber(L,4);
		
		//todo - put these somewhere? and dealloc them eventually?
		new MenuScrollBarText(L, name, value, vmin, vmax);
	} catch (Exception &t) {
		con << __FUNCTION__ << ": " << t << "\n";
	}
LUAFend()

class MenuBasis : public Object {
	typedef Object super;
protected:
	vec2f basis[2];
	lua_State *L;
	string varname;
public:
	MenuBasis(const vec2f &_pos, lua_State *_L, const char *_varname, const vec2f _basis[2])
	: super(_pos, vec2f(3.f)), L(_L), varname(_varname) {
		basis[0] = _basis[0];
		basis[1] = _basis[1];
		update();
	}
	
	virtual bool display() {
		//draws bg, border, text
		if (!super::display()) return false;
		
		glDisable(GL_TEXTURE_2D);

		vec3f colors[] = {
			vec3f(1.f, .3f, .3f),
			vec3f(.3f, .3f, 1.f)
		};
		
		//draw our junk
		vec2f center = size * 0.5f;
		glLineWidth(2.f);
		glPointSize(4.f);
		for (int i = 0; i < 2; i++) {
			vec2f endpos = center + basis[i];
			glColor3fv(colors[i]);
			glBegin(GL_LINES);
			glVertex2fv(center);
			glVertex2fv(endpos);
			glEnd();
			glBegin(GL_POINTS);
			glVertex2fv(endpos);
			glEnd();
		}
		glColor3f(1,1,1);
		glPointSize(1.f);
		glLineWidth(1.f);
		glEnable(GL_TEXTURE_2D);
		
		return true;
	}
	
	int lastbasis;
	vec2f lastv;
	
	virtual eventResponse_t mouseEvent(MouseEvent event, const vec2f &v) {
		if (event.lpress) {
			//see if we're clicking near a basis
			//and move it like so
			float bestdist = fInf;
			for (int i = 0; i < numberof(basis); i++) {
				float dist = (v - (basis[i] + size * 0.5)).lenSq();
				if (dist < bestdist) {
					bestdist = dist;
					lastbasis = i;
				}
			}
		} else if (event.ldown) {
			vec2f delta = v - lastv;
			assert(lastbasis >= 0 && lastbasis < numberof(basis));
			basis[lastbasis] += delta;
			
			update();
		}
		lastv = v;
		return RESPONSE_STOP;
	}
	
	void update() {
		//reset lua junkxxor
		//
		int top = lua_gettop(L);
		lua_newtable(L);
		int table = lua_gettop(L);
		int tableIndex = 1;
		for (int b = 0; b < numberof(basis); b++) {
			for (int v = 0; v < 2; v++) {
				lua_pushnumber(L, tableIndex);
				lua_pushnumber(L, basis[b][v] * (v == 1 ? -1.f : 1.f));
				lua_settable(L, table);
				tableIndex++;
			}
		}
		lua_setglobal(L, varname.c_str());
		lua_settop(L,top);
	}
};

LUAFstart(glushgui_basis)
	try {
		if (argc < 1) throw Exception() << "expected variable name, [{basis xx, xy, yx, yy}]";
		const char *name = lua_tostring(L,1);
		
		vec4f v(1,0,0,1);	//identity matrix
		//get the basis if we can
		if (argc > 1) {
			//put the basis on the top
			lua_settop(L, 2);
			v = luaReadVec4f(L);
		}
		vec2f basis[2];
		basis[0].set(v[0], -v[1]);
		basis[1].set(v[2], -v[3]);	//HACK - store the y axis in negative form so it looks/interacts well with the menu basis
		
		//make our control here
		g_glushMenu.addChild(new Text(g_glushMenu.row, name));
		g_glushMenu.row.y += 1.1f;
		
		g_glushMenu.addChild(new MenuBasis(g_glushMenu.row, L, name, basis));
		g_glushMenu.row.y += 3.6f;
		
	} catch (Exception &t) {
		con << __FUNCTION__ << ": " << t << "\n";
	}
LUAFend()

LUAFstart(glushgui_tex2D)
	//make a 3x3 square
	//texture it with what binding the current unit has
	//upon click do a dialog open
	//open drag onto do a on-drag thing
	if (argc < 1) luaL_error(L, "%s expected a variable name", __FUNCTION__);
	const char *name = lua_tostring(L,1);
	
	//make our control here
	g_glushMenu.addChild(new Text(g_glushMenu.row, name));
	g_glushMenu.row.y += 1.1f;

	MenuTex2D *m = new MenuTex2D(g_glushMenu.row, L, name);
	g_glushMenu.menuTex2Ds.push_back(m);
	g_glushMenu.addChild(m);
	g_glushMenu.row.y += 3.6f;
LUAFend()

class MenuColor : public Object {
	typedef Object super;
protected:
	vec3f color;
	lua_State *L;
	string varname;
	float lum;
public:
	MenuColor(const vec2f &_pos, lua_State *_L, const char *_varname, const vec3f &_color)
	: super(_pos, vec2f(3.f)), L(_L), varname(_varname), color(_color), lum(1.f) {
		update();
	}
	
	virtual bool display() {
		//draws bg, border, text
		if (!super::display()) return false;
		
		glDisable(GL_TEXTURE_2D);
		
		glBegin(GL_TRIANGLE_FAN);
		vec2f center = size*.5f;
		glColor3fv(hsv2rgb(0,0,lum));
		glVertex2fv(center);
		int numdivs = 30;
		for (int div = numdivs; div >= 0; div--) {
			float frac = (float)div / (float)numdivs;
			float x = cos(frac * 2.f * M_PI);
			float y = sin(frac * 2.f * M_PI);
			vec2f v(x,y);
			glColor3fv(hsv2rgb(frac, 1.f, lum));
			glVertex2fv(center + .5f * vecScaleElem(v,size));
		}
		glEnd();
		
		//finally a dot for our current color
		//this would require a rgb->hsv function...
		//is a linear inverse that differs depending on which of the six cube slices you're in
		
		glEnable(GL_TEXTURE_2D);
		
		return true;
	}
	
	vec3f hsv2rgb(float h, float s, float v) {
		vec3f huepts[] = {
			vec3f(1,0,0),
			vec3f(1,1,0),
			vec3f(0,1,0),
			vec3f(0,1,1),
			vec3f(0,0,1),
			vec3f(1,0,1),
		};
		float fseg = h * (float)numberof(huepts);
		int seg = (int)fseg;
		float frac = fseg - (float)seg;
		seg %= numberof(huepts);
		if (seg < 0) seg += numberof(huepts);
		vec3f hc = vecLerp(huepts[seg], huepts[(seg+1)%numberof(huepts)], frac);
		vec3f lc = vec3f(v,v,v);
		return vecLerp(vec3f(1,1,1), hc, s) * lum;
	}
	
	virtual eventResponse_t mouseEvent(MouseEvent event, const vec2f &v) {
		if (event.lpress || event.ldown) {
			vec2f d = v - size * 0.5f;
			float h = atan2(d.y, d.x) / (2.f * M_PI);
			if (h < 0) h++;
			float s = d.len();
			if (s < 0.f) s = 0.f;
			if (s > 1.f) s = 1.f;
			//printf("hsv %f %f %f\n", h, s, lum);
			
			//todo - lum slider below the color box
			color = hsv2rgb(h,s,lum);
			
			//con << "color " << color << endl;
			
			update();
		}
		return RESPONSE_STOP;
	}
	
	void update() {
		//reset lua junkxxor
		int top = lua_gettop(L);
		lua_newtable(L);
		int table = lua_gettop(L);
		for (int v = 0; v < 3; v++) {
 			lua_pushnumber(L, v+1);
			lua_pushnumber(L, color[v]);
			lua_settable(L, table);
		}
		lua_setglobal(L, varname.c_str());
		lua_settop(L,top);
	}
};

LUAFstart(glushgui_color)
	if (argc < 1) luaL_error(L, "%s expected a variable name", __FUNCTION__);
	const char *name = lua_tostring(L,1);

	vec3f color(1.f, 1.f, 1.f);
	if (argc > 1) {
		lua_settop(L, 2);
		color = luaReadVec3f(L);
	}
	
	//make our control here
	g_glushMenu.addChild(new Text(g_glushMenu.row, name));
	g_glushMenu.row.y += 1.1f;

	g_glushMenu.addChild(new MenuColor(g_glushMenu.row, L, name, color));
	g_glushMenu.row.y += 3.6f;
LUAFend()

static const luaL_reg menuBindings[] = {
	{"checkbox", PF_glushgui_checkbox},
	{"button", PF_glushgui_button},
	{"slider", PF_glushgui_slider},
	{"basis", PF_glushgui_basis},
	{"tex2D", PF_glushgui_tex2D},
	{"color", PF_glushgui_color},
//terminator
	{NULL, NULL}
};

void GlushMenu::bind(lua_State *L) {
	int top = lua_gettop(L);
	lua_getglobal(L, "glush");
	lua_newtable(L);
	
	luaAddBindings(L, menuBindings);
	lua_setfield(L, -2, "gui");
	
	luaopen_menu(L);	//pushes the menu table onto the stack
	lua_setfield(L, -2, "menu");	//call it "menu" for now, as to not conflict with "gui".  maybe rename one to the other later?

	lua_pop(L,1);
	assert(lua_gettop(L) == top);
}

