/*

    File: program.cpp

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <math.h>

#include <iostream>
#include <vector>
#include <string>
#include <map>

#include "system/exception.h"
#include "system/singletons.h"
#include "base/main.h"
#include "base/filebuffer.h"
#include "main.h"
#include "bindings.h"
#include "menu.h"
//#include "zip.h"

#ifdef WIN32
#include <direct.h>
#define getcwd _getcwd
#define PATH_MAX MAX_PATH
#endif

//set this flag to disable lua-based multithreading and timeslicing
//so far I've yet to use the feature at all in any of my demos (though it is a useful feature)
//and this feature is the only point at which my modifications to the lua library are needed
//
//If you are linking to an original lua library and don't want to modify it then set this flag
//
//#define DISABLE_LUA_THREADS

using namespace std;

void lua_printbacktrace(lua_State *L) {
	con << "Lua stack trace:\n";
	lua_Debug ar;
	memset(&ar, 0, sizeof(ar));
	for (int level = 0; lua_getstack(L, level, &ar); level++) {  /* check function at level */
		lua_getinfo(L, "Snl", &ar);  /* get info about it */
		con << "level " << level << endl;
		con << "  event:       " << ar.event << endl;
		con << "  name:        " << ar.name << endl;
		con << "  namewhat:    " << ar.namewhat << endl;
		con << "  what:        " << ar.what << endl;
		con << "  source:      " << ar.source << endl;
		con << "  currentline: " << ar.currentline << endl;
		con << "  nups:        " << ar.nups << endl;
		con << "  linedefined: " << ar.linedefined << endl;
		con << "  lastlinedefined: " << ar.lastlinedefined << endl;
		con << "  short_src:   " << ar.short_src << endl;
		//    if (ar.currentline > 0) {  /* is there info? */
		//    con.print(L, "%s:%d: ", ar.short_src, ar.currentline);
		memset(&ar, 0, sizeof(ar));
	}
}

#ifndef DISABLE_LUA_THREADS

static void threadTable_hookFunc(lua_State *L, lua_Debug *ar);

class LuaThread {
public:
	lua_State *L;
	double 	elapsedTime;	//elapsed time in the spotlight - used for callbacks to determine how long we've been running
	double 	lastTime;	//the last read time for this thread.  set this when a thread starts/resumes
	int		hookCount;	//for kicks, the # of hook callbacks we got since running in our timeslice
	
	int index;
	
	//later:
	double sleepEndTime;// - if this is set above our app timer then we're still asleep and dont wake us - used with sleep()
	int waitingOn;// - if this is a valid index of a thread in the table then we're waiting for that thread to die before continuing - used with wait()
	
	LuaThread() : L(NULL), elapsedTime(0), lastTime(0), hookCount(0), sleepEndTime(0), waitingOn(-1), index(0xdeadbeef) {}
	
	//create right off the bat
//	LuaThread(const char *fn)
//	: L(NULL), elapsedTime(0), lastTime(0), hookCount(0), sleepEndTime(0), waitingOn(-1)
//	{ create(fn); }
	
	//creates a thread from the base state
	//and clears that new thread entry from the base state stack (where its pushed by default)
	//inits the new state to a specified function
	//leaves the function sitting on top of the stack
	//sets up hook callback
	//
	//the thread will be the function on top of srcState's stack
	//
	// srcState most likely won't be the mainState (where I make my threads from.  why do i do this? i dont know)
	bool create(lua_State *srcState) {
	
		if (!lua_isfunction(srcState, -1)) {
			con << "LuaThread::create - didn't get a function as a parameter!\n";
			return false;
		}
		
		lua_State *mainState = progSys.getState();
		
		L = lua_newthread(mainState);	//make a new thread
		
		//how are we gonna get rid of threads?
		//I'm gonna move it to the base of the stack 
		//because I think removing it makes GC tweak out
		//lua_insert(mainState, 1);			//and take it off the stack
		//lua_pop(mainState, 1);
		
		{	//[newthread]
			lua_getglobal(mainState, "glush");		//[newthread] [glush]
			lua_getfield(mainState, -1, "thread");	//[newthread] [glush] [glush.thread]
			lua_getfield(mainState, -1, "threads");			//[newthread] [glush] [glush.thread] [glush.thread.threads]
			if (lua_isnil(mainState, -1)) {
				lua_pop(mainState,1);					//[newthread] [glush] [glush.thread]
				lua_newtable(mainState);				//[newthread] [glush] [glush.thread] [glush.thread.threads]
				lua_pushvalue(mainState, lua_gettop(mainState));		//[newthread] [glush] [glush.thread] [glush.thread.threads] [glush.thread.threads]
				lua_setfield(mainState, -3, "threads");	//[newthread] [glush] [glush.thread] [glush.thread.threads]
			}
			//[newthread] [glush] [glush.thread] [glush.thread.threads]
			lua_pushinteger(mainState, lua_objlen(mainState, -1)+1);	//[newthread] [glush] [glush.thread] [glush.thread.threads] [index]
			lua_pushvalue(mainState, -5);	//[newthread] [glush] [glush.thread] [glush.thread.threads] [index] [newthread]
			lua_rawset(mainState, -3);	//[newthread] [glush] [glush.thread] [glush.thread.threads]
			lua_pop(mainState, 4);
		}

		//move the function from srcState to L
		lua_xmove(srcState, L, 1);	//from, to, n (why not to,from like mem/strcpy? )
		
		if (!lua_isfunction(L, -1)) {
			con << "LuaThread::create - couldnt find a function\n";
			lua_pop(L,1);
			return false;
		}
		
		//set hooks - only allow so many ops to run
		//lets think about this
		//if we get a PF_sleep then the thread lua_yield's
		//if we get a PF_wait then the thread lua_yield's
		//for what other reason do we want our thread to yield?
		//either way, currently all our line counter does is count the time
		//so why don't we insert a time counter instead?
		// . . . because lua doesn't have any time-based hooks . . .
		lua_sethook(L, threadTable_hookFunc, LUA_MASKCOUNT, 24);	// 8 instructions / loop
		return true;
	}

	//runs the function - called at the beginning of each timeslice
	//expects the function to be on top of the stack, as create() leaves it
	//prereq: create()
	bool run(void) {
		hookCount = 0;		//clear the callback ticker
		elapsedTime = 0;	//clear its timeslice timer
		lastTime = timer.getTime();	//set its last time for callbacks' sake

		//run it
		int status = lua_resume(L, 0);

		if (status == 0) {	//thread finished
			return false;
		}
		if (status == LUA_YIELD) {	//still going
			return true;
		}
		//else...
		con << "run func: lua_pcall got error " << status << endl;
		string errorstr = lua_tostring(L, -1);
		con << " : " << errorstr << endl;
		lua_pop(L, 1);
		
		lua_printbacktrace(L);
		return false;

	}
};

class LuaThreadTable {
	vector<LuaThread *> threads;
	int lastUpdated;	//the last updated thread
public: 
	
	LuaThreadTable() : lastUpdated(0) {}
	
	int addThread(LuaThread *th) {
		if (!th) return -1;		//me being lazy at my error handling
		threads.push_back(th);
		th->index = threads.size()-1;
		return threads.size()-1;
	}
	
//how long the table is to run
#define SCRIPT_DURATION		0.01		//10ms
//how long each thread in the table can run (max)
#define SCRIPT_THREAD_DURATION	0.003		//3ms
	void run(void) {

		if (isEmpty()) return;
	
		int threadCount = threads.size();
		
		double startTick = timer.getTime();
		double thisTick;
		for (; (thisTick = timer.getTime()) < (startTick + SCRIPT_DURATION);
				lastUpdated = (lastUpdated+1) % threadCount)
		{
			LuaThread *t = threads[lastUpdated];

			//null entries are allowed for uniqueness of index in the thread table
			if (!t) continue;
			if (t->waitingOn != -1) continue;
			if (t->sleepEndTime >= timer.getTime()) continue;
			
			if (!t->run()) {
				
				//destroy the thread
				//there is no official Lua destroy thread function - they just drift away with the rest of GC'd objs
				delete t;
				threads[lastUpdated] = NULL;
				
				//now find all waiting on this thread
				//and wake them
				bool found = false;
				for (int j = 0; j < threadCount; j++) {
					LuaThread *wt = threads[j];
					if (wt) {
						found = true;
						if (wt->waitingOn == lastUpdated) {
							wt->waitingOn = -1;
						}
					}
				}
				//if we finished a thread and didnt find any more, dont bother stay in the table loop
				if (!found) {
					break;
				}
			}
		}
	}
	
	LuaThread *find(lua_State *L) {
		for (int i = 0; i < threads.size(); i++) {
			LuaThread *t = threads[i];
			if (!t) continue;
			if (t->L == L) return t;
		}
		return NULL;
	}
	
	LuaThread *getByIndex(int i) {
		if (i < 0 || i >= threads.size()) return NULL;
		return threads[i];
	}
	
	bool isEmpty(void) {
		for (int i = 0; i < threads.size(); i++) {
			if (threads[i]) return false;
		}
		return true;
	}
	
	void clear(void) {
		VECTOR_DELETE_ALL(threads);
	}
};

static LuaThreadTable threadTable;

static void threadTable_hookFunc(lua_State *L, lua_Debug *ar) {
	//TODO - if this is hit upon any lua code called within C code called from a lua child thread
	//(which happens more often than you'd think)
	//then it will break.
	//SOLUTION: when we run a lua child thread -> c function (any PF_*** funcs) then
	//we should set a semaphore
	//if that semaphore is set here then dont yield!
	//ANOTHER SOLUTION: if lua keeps track of c stack depth then, here, if thats set then dont yield

	//1) lookup the lua state in our thread table
	//2) read its last time, subtract the current time, to calc the elapsed time
	//3) add that to its current time in the spotlight, set its last time
	//4) if its time in the spotlight exceeds our per-thread maximum then
	//		yield this thread
	//		and resume another one.  can we do that from a callback function like this?
	LuaThread *t = threadTable.find(L);
	assert(t);
	t->hookCount++;
	double thisTime = timer.getTime();
	double dt = thisTime - t->lastTime;
	t->elapsedTime += dt;

	//but this is!  this is a direct rip from the lua code
	if (!lua_canyield(L)) {
//		con << "got a L->nCcalls yield stall\n";
		return;
	}

	//if we haven't finished our time then bail
	if (t->elapsedTime < SCRIPT_THREAD_DURATION) return;
	
	//if we've ate up all our time then yield this thread
	//from there, we'll return to the lua_pcall to continue on our merry way
	//otherwise we'll return to the lua code executer
	lua_yield(L, 0);
}
#endif //ALLOW_LUA_THREADS

//refs seem to be a quick way to remove references from tables
//but they're not working for me...
//so I'll fall back on int keys & setting to nil to remove
//#define TIMEOUT_USE_REFS

static const char *timeoutTableName = "timeouts";

//if it needs this for updating threads or what not
void ProgSys::updateThreads() {
	if (!L) return;

	if (!lua_checkstack(L, 100)) {
		//for whatever its worth ...
		con << "lua_checkstack failed!\n";
	}

#ifndef DISABLE_LUA_THREADS
	threadTable.run();
#endif
	
	//now we run the timeout tables, if any are valid
	{
		lua_getglobal(L, "glush");		//glush
		lua_getfield(L, -1, "thread");	//glush thread
		lua_getfield(L, -1, timeoutTableName);		//glush thread threads
		int t = lua_gettop(L);	//get our table's index in the stack
		if (!lua_istable(L, -1)) {
			//con << "timeout table is non-table: " << lua_typename(L, lua_type(L,-1)) << endl;
			lua_pop(L, 1);
		} else {
		
			//keep track of all the functions we gotta delete
			//cap it though - instead of doing dyn alloc per frame
			int rememberToDelete[256];
			int rememberCount = 0;
		
			//[timeout table]
			lua_pushnil(L);					//[timeout table] [nil]
			while (lua_next(L, t) != 0) {
				// `key' is at index -2 and `value' at index -1
				// @t [timeout table] [key] [value]
				
				bool deleteThis = false;
				
				if (!lua_istable(L, -1)) {
					con << "FOUND A NON-TABLE OBJECT IN OUR TIMEOUT TABLE!" << endl;
					deleteThis = true;
				} else {
				
					// now get elem 1 of the current value -- that'll be the next execution time
					lua_rawgeti(L, -1, 1);					// @t [timeout table] [key] [entry] [timeout time]
					lua_Number nexttime = lua_tonumber(L, -1);
					lua_pop(L, 1);							// @t [timeout table] [key] [entry]

					if (nexttime <= timer.getTime()) {
					
						//con << "calling timeout index " << (int)lua_tonumber(L, -2) << endl;
						
						//better execute it
						lua_rawgeti(L, -1, 0);				//@t [timeout] [key] [entry] [timeout function]
						int status = lua_pcall(L, 0, 0, 0);	//@t [timeout] [key] [entry] 
						if (status) {
							con << "timeout execution: lua_pcall got error " << status << endl;
							con << " : " << lua_tostring(L, -1) << endl;	//display the error
							lua_pop(L, 1);	//pop the error
						}
						//@t [timeout] [key] [entry] 
						
						deleteThis = true;
					}
				}
				
				if (deleteThis) {
					
					if (rememberCount < numberof(rememberToDelete)) {
						rememberToDelete[rememberCount] = (int)lua_tonumber(L, -2);	//convert the key to a lua number
						rememberCount++;
					} else {
						con << "timeout delete list overflow!\n";
					}
				}
				
				lua_pop(L, 1);  /* removes `value'; keeps `key' for next iteration */
			}
			
			//now we just have the table on top
			//now we go through and remove what of it we need to
			
			for (int i = 0; i < rememberCount; i++) {
				int id = rememberToDelete[i];
				//con << "removing timeout index " << id << endl;
#ifdef TIMEOUT_USE_REFS
				luaL_unref(L, -1, id);
#else
				//looks like this doesn't remove entries from the table
				lua_pushnil(L);
				lua_rawseti(L, -2, id);
#endif
			}
			
			//if (rememberCount) con << "done removing!\n";
			
			lua_pop(L, 1);	//now get rid of the table
		}
		lua_pop(L,2);
	}
}

//my attempt at a library:
#ifndef DISABLE_LUA_THREADS
static int PF_fork(lua_State *L) {
	if (!lua_gettop(L)) luaL_error(L, "%s expected a parameter", __FUNCTION__);

	lua_settop(L, 1);	//move our first param to the top
	if (!lua_isfunction(L, 1)) luaL_error(L, "%s expected a function", __FUNCTION__);
	
	LuaThread *th = new LuaThread();
	if (!th) luaL_error(L, "%s couldn't create thread object", __FUNCTION__);

	if (!th->create(L)) {
		delete th;
		luaL_error(L, "%s wasn't able to fork to the provided function", __FUNCTION__);
	}
	
	int threadIndex = threadTable.addThread(th);
	lua_pushinteger(L, threadIndex);
	
	return 1;
}

static int PF_wait(lua_State *L) {
	int threadIndex = 0;
	if (lua_gettop(L)) threadIndex = (int)lua_tonumber(L, 1);
	
	//current thread
	LuaThread *waiter = threadTable.find(L);
	if (!waiter) {
		con.print("PF_wait called by an unknown thread\n");
		return 0;
	}

	//to be waited on thread
	LuaThread *waitingOn = threadTable.getByIndex(threadIndex);
	if (!waitingOn) {
		con.print("PF_wait couldnt find thread %d\n", threadIndex);
		return 0;	//error or something?
	}
	
//	con.print("PF_wait %p waiting on index %d\n", waiter, (int)threadIndex);

	//ref by handle
	waiter->waitingOn = threadIndex;
	//tell lua to stop running as well
	return lua_yield(L, 0);	//return 0;
}

static int PF_sleep(lua_State *L) {
	lua_Number endTime = 0;
	if (lua_gettop(L)) endTime = lua_tonumber(L, 1);
	endTime += timer.getTime();
	
	//current thread
	LuaThread *sleeper = threadTable.find(L);
	if (!sleeper) {
		con.print("PF_sleep called by an unknown thread\n");
		return 0;
	}
	
//	con.print("PF_sleep %p until %f\n", sleeper, endTime);

	sleeper->sleepEndTime = endTime;
	//tell lua to stop running as well
	return lua_yield(L, 0);	//return 0;
}
#endif //ALLOW_LUA_THREADS

static int PF_bitand(lua_State *L) {
	int argc = lua_gettop(L);
	int param = (int)lua_tonumber(L,1);
	for (int i = 2; i <= argc; i++) {
		param &= (int)lua_tonumber(L,i);
	}
	lua_pushinteger(L,param);
	return 1;
}

static int PF_bitor(lua_State *L) {
	int argc = lua_gettop(L);
	int param = (int)lua_tonumber(L,1);
	for (int i = 2; i <= argc; i++) {
		param |= (int)lua_tonumber(L,i);
	}
	lua_pushinteger(L,param);
	return 1;
}

static int PF_bitxor(lua_State *L) {
	int argc = lua_gettop(L);
	int param = (int)lua_tonumber(L,1);
	for (int i = 2; i <= argc; i++) {
		param ^= (int)lua_tonumber(L,i);
	}
	lua_pushinteger(L,param);
	return 1;
}

static int PF_bitnot(lua_State *L) {
	lua_pushinteger(L, ~(int)lua_tonumber(L, 1));
	return 1;
}

static int PF_bitshiftleft(lua_State *L) {
	lua_pushinteger(L, (int)lua_tonumber(L,1) << (int)lua_tonumber(L,2));
	return 1;
}
static int PF_bitshiftright(lua_State *L) {
	lua_pushinteger(L, (int)lua_tonumber(L,1) >> (int)lua_tonumber(L,2));
	return 1;
}

static int PF_chdir(lua_State *L) {
	if (lua_gettop(L) == 0) {
		con << "usage: PF_chdir(dir)\n";
		return 0;
	}

	chdir(lua_tostring(L,1));
	return 0;
}

static int PF_getcwd(lua_State *L) {
	char cwd[PATH_MAX];	//windows might want to see MAX_PATH.  be prepared to #ifdef
	getcwd(cwd, sizeof(cwd));
	lua_pushstring(L, cwd);
	return 1;
}

//void print(string)
static int PF_print(lua_State *L) {
	int argc = lua_gettop(L);
	ostringstream s;
	lua_getglobal(L, "tostring");
	for (int i = 1; i <= argc; i++) {
		lua_pushvalue(L, -1);
		lua_pushvalue(L, i);
		lua_pcall(L, 1, 1, 0);
		const char *str = lua_tostring (L, -1);
		if (i>1) s << "\t";
		if (str) s << str;
		lua_pop(L,1);
	}
	s << "\n";
	con << s.str();
	return 0;
}

static int PF_cvarExec(lua_State *L) {
	const char *cmd = lua_tostring(L, 1);
	if (cmd) cvarSys->exec((char *)cmd);
	return 0;
}


//atm we are mapping unique indices into functions to call back
//what we need to do is map unique indices into two-element tables:: 1: the function to callback, & 2: when to call it back

static int PF_setTimeout(lua_State *L) {
	if (lua_gettop(L) != 2) {
		con << "usage: PF_setTimeout(function, delay)\n";
		return 0;
	}
	
	if (!lua_isfunction(L, 1)) {
		con << "PF_setTimeout expected a function\n";
		return 0;
	}
	
	//grab the timeout from the top - for later
	lua_Number runtime = lua_tonumber(L, 2);
	runtime += timer.getTime();
	
	lua_settop(L, 1);	//cut out all else but the function
	//stack: [function]
	
	//create a table w/elem 0 as the function and elem 1 as the timeout
	lua_newtable(L);			// [function] [table]
	lua_insert(L, 1);			// [table] [function]
	lua_rawseti(L, 1, 0);		// [table={0=>function}]
	lua_pushnumber(L, runtime);	// [table={0=>function}] [number:timeout]
	lua_rawseti(L, 1, 1);		// [table={0=>function, 1=>timeout}]
			
	//now get the master timeout table
	//if it doesnt exist then create it
	
	//so we're gonna make a new table off of globals for holding all our timeout function calls
	lua_getglobal(L, "glush");
	lua_getfield(L, -1, "thread");
	lua_getfield(L, -1, timeoutTableName);	// [new entry] [glush] [thread] [timeouts]
	
	//if nothing's there then we create a new one
	if (!lua_istable(L, -1)) {
	
		//get that miscreated table off the stack!
		lua_pop(L, 1);						// [new entry] [glush] [thread]
		
		//and make a new one
		lua_newtable(L);				// [new entry] [glush] [thread] [new table]

		//this method pops our table, so we gotta put it back on the stack
		lua_setfield(L, -2, timeoutTableName);	//[new entry] [glush] [thread]
		
		//is the table still on top? nope - put it back there!
		lua_getfield(L, -1, timeoutTableName);	//[new entry] [glush] [thread] [timeouts]

		//btw, what is the difference between lua_setfield and lua_settable ?
		//maybe setfield is a shortcut for pushing the key string onto the stack?
	}
			
	//now we push the stuff onto the master stack
	
	lua_insert(L, 1);					//[glush] [thread] [timeouts] [new entry] 
	
#ifdef TIMEOUT_USE_REFS
	int id = luaL_ref(L, 1);
#else
	static int id = 0;
	id++;
	lua_rawseti(L, 2, id);
#endif
	
	//con << "pushed a new timeout entry w/id " << id << endl;
			
	//and have faith that that worked

	return 0;
}

static int PF_removeTimeout(lua_State *L) {
	luaL_error(L, "%s not implemented yet!", __FUNCTION__);
	//I hope this isn't being used quite yet ...
	return 0;
}

//courtesy of lua.c
static int luaErrorHandler(lua_State *L) {	//err
	if (!lua_isstring(L, 1))  /* 'message' not a string? */
		return 1;  /* keep it intact */
	lua_getfield(L, LUA_GLOBALSINDEX, "debug");
	if (!lua_istable(L, -1)) {
		lua_pop(L, 1);
		return 1;
	}
	lua_getfield(L, -1, "traceback");
	if (!lua_isfunction(L, -1)) {
		lua_pop(L, 2);
		return 1;
	}
	lua_pushvalue(L, 1);  /* pass error message */
	lua_pushinteger(L, 2);  /* skip this function and traceback */
	lua_pcall(L, 2, 1, 0);  /* call debug.traceback */
	return 1;
}

//luaSafeCall prints by default.  maybe it should just return?
//and maybe each caller should print?
bool luaSafeCall(lua_State *L, int nargs, int nres) {	//func, ...[nargs]
	
	lua_pushcfunction(L, luaErrorHandler);	//func, ...[nargs], errfunc
	int errfuncloc = lua_gettop(L) - nargs - 1;
	lua_insert(L, errfuncloc);				//errfunc, func, ...[nargs]
	bool ret = true;
	if (lua_pcall(L, nargs, nres, errfuncloc)) {	//errfunc, ...[nres] or errmsg
		ret = false;
		if (lua_isstring(L, -1)) {
			con << lua_tostring(L, -1);
		}
	}
	lua_remove(L, errfuncloc);		//...[nres] or errmsg
	return ret;
}


class LuaConsoleInterpreter : public ConsoleInterpreter {
public:
	virtual void exec(const char *_buffer) {
		string buffertmp;	//only assign if necessary -- doesn't initial assignment cause allocation?
		const char *buffer = _buffer;
	
		//eventually i'll turn this into the lua.exe file ...
		if (buffer[0] == '=') {
			buffertmp = string("print(tostring(") + (_buffer+1) + string("))");
			buffer = buffertmp.c_str();
			//keep buffertmp alive 'til the end of the function
		}

		//gotta add quotes for spaces in arguments to cvar system . . .
		lua_State *L = progSys.getState();
		int top = -1;
		
		try {
			if (!L) throw Exception() << "couldn't find the lua state";
			top = lua_gettop(L);
			int err = 0;
			if (err = luaL_loadbuffer(L, buffer, strlen(buffer), "script")) {
				throw Exception() << "for including script code:\n" << buffer << "\n\n"
					<< " we got some errors:\n"
					<< lua_tostring(L, -1) << "\n"
					<< lua_tostring(L, -2) << "\n";
			}
			luaSafeCall(L, 0, 0);
		} catch (Exception &t) {
			con << __FUNCTION__ << " got error: " << t << endl;
		}

		if (top != -1) lua_settop(L, top);
	}
};
static LuaConsoleInterpreter luaInterpreter;


//game-independent
static const luaL_reg frameworkGlueFuncs[] = {
//OS exposure
	{"chdir",				PF_chdir},
	{"getcwd",				PF_getcwd},
//base class stuff
	{"print",				PF_print},					//void print(string)
	{"cvarExec",			PF_cvarExec},
//binary operations
	{"bitand",				PF_bitand},
	{"bitor",				PF_bitor},
	{"bitxor",				PF_bitxor},
	{"bitnot",				PF_bitnot},
	{"bitshiftleft",		PF_bitshiftleft},
	{"bitshiftright",		PF_bitshiftright},

	{NULL, NULL}
};

//game-dependent
static const luaL_reg threadGlueFuncs[] = {
//multithread
#ifndef DISABLE_LUA_THREADS
	{"fork",				PF_fork},					//thread fork( void(*)() )
	{"wait",				PF_wait},					//void wait(thread)
	{"sleep",				PF_sleep},					//void sleep(float)
#endif
//engine calls
	{"setTimeout",			PF_setTimeout},				//void setTimeout( void(*)() )	-- our single-thread fork alternative
	{"removeTimeout",		PF_removeTimeout},			//void removeTimeout( void(*)() ) -- and its alternative
//terminator
	{NULL, NULL}
};

static const luaL_reg luaRegEmpty[] = {
	{NULL, NULL}
};

//currently only referenced by progSys.sessionInit
//this means the mod will have been loaded already
//this means its safe to call a game api function to load its predef functions
void ProgSys::buildPredefined() {	

	//create all modules here
	luaL_register(L, "gl", luaRegEmpty);
	lua_pop(L,1);

	luaL_register(L, "glush", frameworkGlueFuncs);

	lua_newtable(L);
	luaAddBindings(L, threadGlueFuncs);
	lua_setfield(L, -2, "thread");
	lua_pop(L,1);

	//found here and in libglush
	//perhaps they should be put in some common location?
	luaStateAddMainCalls(L);
	luaBind(L);
	glushMenu.bind(L);
	
	//special case for the non lib: rebind os.exit to glush.requestExit
	{
		int top = lua_gettop(L);
		lua_getglobal(L, "os");		//os
		lua_getglobal(L, "glush");	//os glush
		lua_getfield(L, -1, "exit");	//os glush glush.exit
		lua_setfield(L, -3, "exit");	//os glush
		lua_pop(L,2);
		assert(lua_gettop(L) == top);
	}
	
	//another special case: rebind print to glush.print
	{
		int top = lua_gettop(L);
		lua_getglobal(L, "glush");	//glush
		lua_getfield(L, -1, "print");	//glush, print
		lua_setglobal(L, "print");	//glush
		lua_pop(L,1);
		assert(lua_gettop(L) == top);
	}
}

//stop all threads ...
void ProgSys::stopThreads() {
#ifndef DISABLE_LUA_THREADS
	threadTable.clear();
#endif
}

bool ProgSys::loadFile(const string &filename) {
	assert(filename.length());	
	ostringstream ss;

#if 0	//require seemed to work better in 5.1.3 .. or maybe that's me..
		//that and I just changed the extension to be part of the loadFile api param
		//so this code would need to manually validate it and chop it off
		//but as it is this is already done in all the calling functions
	ss << "require '" << filename << "'\n";
#else
	{
		FileBuffer fb((filename).c_str());
		if (!fb.getData()) throw Exception() << "failed to load the file " << filename;
		ss << fb.getData();
	}
#endif

	//for some odd reason without moving our stringstream into a separate stream
	//it tends to get garbled before we can use it ...
	string sstr = ss.str();
	const char *buffer = sstr.c_str();

	int err = 0;
	if (err = luaL_loadbuffer(L, buffer, strlen(buffer), "script")) {
		con << "for including script code:\n" << buffer << endl << endl;
		con << " we got some errors:\n" << lua_tostring(L, -1) << endl << lua_tostring(L, -2) << endl;
		return false;
	}
	
	return true;
}

//all the gamemod needs
lua_State *ProgSys::getState() {
	return L;
}


static int panic(lua_State *L) {
	con.print("Lua Paniced!\n");
	con.print("%s\n", lua_tostring(L, -1));
	con.print("%s\n", lua_tostring(L, -2));
	return 0;
}

//run all base level code, and thus register all functions and what not in the global table
void ProgSys::runBase(const vector<string> &args) {
	assert(L);
	
	//begin rip from lua.exe: - i guess this is for the callback for pcall?
	int base = lua_gettop(L);  /* function index */
	
	lua_pushliteral(L, "_TRACEBACK");
	lua_rawget(L, LUA_GLOBALSINDEX);  /* get traceback function */
	lua_insert(L, base);  /* put it under chunk and args */
	
	for (int i = 0; i < args.size(); i++) {
		lua_pushstring(L, args[i].c_str());
	}
	
	if (!luaSafeCall(L, args.size(), 0)) {
		con << "lua_pcall got error during ProgSys::runBase" << endl;
		lua_pop(L, 1);	//pop the error
	}

	lua_remove(L, base);  /* remove traceback function */
	//end rip
}

ProgSys::ProgSys() : L(NULL), doNotClose(false) {
}

void ProgSys::init(const string &filename, const vector<string> &args) {
	con.setInterpreter(&luaInterpreter);
	stopThreads();
	
	try {
	
		//reconstruct the main state here
		L = lua_open();
		if (!L) throw Exception() << "failed to init lua state";
		//maybe the main state should only be created at the session?
		lua_atpanic(L, panic);

		luaL_openlibs(L);
		buildPredefined();	//2) re-register any predefined functions
		
		//now run our base code
		if (!loadFile(filename)) throw Exception() << "loadFile failed";
		
		//and the ol' faithful, run-base to register all our base-level functions
		runBase(args);
		
		//now capture our display function reference from the global
		//how do you capture a function in lua?
		//make a new state and copy it there?

	} catch (Exception &t) {
		con << "ProgSys::init failed: " << t << endl;
		kill();
	}
}

void ProgSys::initWithGoods(lua_State *_L) {
	L = _L;
	doNotClose = true;

	con.setInterpreter(&luaInterpreter);
	stopThreads();

	buildPredefined();
}

	
void ProgSys::kill() {
	stopThreads();
	
	if (L && !doNotClose) lua_close(L);

	L = NULL;
	doNotClose = false;

	stopThreads();
}
