/*

    File: program.h

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#ifndef GLUSH_PROGRAM_H
#define GLUSH_PROGRAM_H

#include <lua.hpp>
#include <iostream>
#include <vector>
#include <string>
#include "math/vec.h"

#include "base/base.h"

inline void lua_rawgetfield(lua_State *L, int t, const char *k) {
	if (t < 0) t += lua_gettop(L) + 1;
	lua_pushstring(L,k);
	lua_rawget(L,t);
}

inline void lua_rawsetfield(lua_State *L, int t, const char *k) {
	int top = lua_gettop(L);
	if (t < 0) t += top + 1;
	lua_pushstring(L,k);
	lua_insert(L, top);
	lua_rawset(L, t);
}


class ProgSys {
protected:
	lua_State *L;
	bool doNotClose;
public:
	ProgSys();
	
	void initWithGoods(lua_State *L);
	void init(const std::string &filename, const std::vector<std::string> &args);
	void kill();

	void updateThreads();
	lua_State *getState();
protected:
	void stopThreads();
	bool loadFile(const std::string &filename);
	void buildPredefined();
	void runBase(const std::vector<std::string> &args);
};

bool luaSafeCall(lua_State *L, int nargs, int nres);

// some quick macros for exposing constants in lua
// you know, considering the below, maybe I should make a structure for these too?

#define LUAGi(x)		lua_pushinteger(L, x); lua_setfield(L, -2, #x);
#define LUAGiEXT(x)		lua_pushinteger(L, x##_EXT); lua_setfield(L, -2, #x);
#define LUAGiARB(x)		lua_pushinteger(L, x##_ARB); lua_setfield(L, -2, #x);
#define LUAGiAPPLE(x)	lua_pushinteger(L, x##_APPLE); lua_setfield(L, -2, #x);

// heres all my lua function shortcuts:

#define LUA_GLUE(x)			{#x, PF_##x},
#define LUA_GLUEARB(x)		{#x, PF_##x##ARB},
#define LUA_GLUEivARB(x)	{#x, PF_##x##ivARB},
#define LUA_GLUEsARB(x)		{#x, PF_##x##sARB},
#define LUA_GLUEvARB(x)		{#x, PF_##x##vARB},
#define LUA_GLUEd(x)		{#x, PF_##x##d},
#define LUA_GLUEf(x)		{#x, PF_##x##f},
#define LUA_GLUEEXT(x)		{#x, PF_##x##EXT},
#define LUA_GLUEi(x)		{#x, PF_##x##i},
#define LUA_GLUEs(x)		{#x, PF_##x##s},
#define LUA_GLUEsEXT(x)		{#x, PF_##x##sEXT},

//this doesn't check function existence like we want to do for opengl bindings
//	to fix that ... we need to step away from the array addition and macro-add all the functions
//	... then do the test in the macro
inline void luaAddBindings(lua_State *L, const luaL_reg *reg) {
	assert(lua_istable(L, -1));
	for (; reg->func; reg++) {
		lua_pushcfunction(L, reg->func);
		lua_setfield(L, -2, reg->name);
	}
}

#define LUA_GLUE_FUNC_SAFE_GENERAL(name, func)	\
	if (func) {	\
		lua_pushcfunction(L, PF_##func);	\
		lua_setfield(L, -2, #name);	\
	}

#define LUA_GLUE_FUNC_SAFE_SUFFIX(x,y)	LUA_GLUE_FUNC_SAFE_GENERAL(x, x##y)
	
#define LUA_GLUE_FUNC_SAFE(x)		LUA_GLUE_FUNC_SAFE_GENERAL(x, x)
#define LUA_GLUE_FUNC_SAFE_s(x)		LUA_GLUE_FUNC_SAFE_SUFFIX(x, s)
#define LUA_GLUE_FUNC_SAFE_ARB(x)	LUA_GLUE_FUNC_SAFE_SUFFIX(x,ARB)
#define LUA_GLUE_FUNC_SAFE_sARB(x)	LUA_GLUE_FUNC_SAFE_SUFFIX(x,sARB)
#define LUA_GLUE_FUNC_SAFE_EXT(x)	LUA_GLUE_FUNC_SAFE_SUFFIX(x,EXT)
#define LUA_GLUE_FUNC_SAFE_sEXT(x)	LUA_GLUE_FUNC_SAFE_SUFFIX(x,sEXT)

//#define LUA_GL_TRACE

template<int DIM, typename T>
inline _vec<DIM,T> luaReadVec(lua_State *L) {
	_vec<DIM,T> v( (T)0.0);
	//read indices 1,2,3,4
	int t = lua_gettop(L);
	if (!lua_istable(L, t)) {
		con << "luaReadVec didnt get a table" << std::endl;
		return v;	//for 1 arg, better be a table
	}
	for (int i = 0; i < DIM; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, t);
		v[i] = (T)lua_tonumber(L,-1);
		lua_pop(L,1);
	}
	assert(lua_gettop(L) == t);
	return v;
}

template<typename T>
inline void luaReadVecDyn(lua_State *L, _vec<0,T> &v) {
	//read indices 1,2,3,4
	int t = lua_gettop(L);
	if (!lua_istable(L, t)) {
		con << "luaReadVec didnt get a table" << std::endl;
		return;	//for 1 arg, better be a table
	}
	int len = lua_objlen(L, t);
	v.resize(len);
	for (int i = 0; i < len; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, t);
		v[i] = (T)lua_tonumber(L,-1);
		lua_pop(L,1);
	}
	assert(lua_gettop(L) == t);
}

typedef _vec<1,double> vec1d;

inline vec2i luaReadVec1i(lua_State *L) { return luaReadVec<2,int>(L); }	//dirty hack.  _vec<1,T> has problems with the constructor, so i just switched it up.  oh well, i hope no one ever uses this
inline vec2i luaReadVec2i(lua_State *L) { return luaReadVec<2,int>(L); }
inline vec3i luaReadVec3i(lua_State *L) { return luaReadVec<3,int>(L); }
inline vec4i luaReadVec4i(lua_State *L) { return luaReadVec<4,int>(L); }
inline vec2f luaReadVec1f(lua_State *L) { return luaReadVec<2,float>(L); }	//dirty hack.  _vec<1,T> has problems with the constructor, so i just switched it up.  oh well, i hope no one ever uses this
inline vec2f luaReadVec2f(lua_State *L) { return luaReadVec<2,float>(L); }
inline vec3f luaReadVec3f(lua_State *L) { return luaReadVec<3,float>(L); }
inline vec4f luaReadVec4f(lua_State *L) { return luaReadVec<4,float>(L); }
inline vec2d luaReadVec1d(lua_State *L) { return luaReadVec<2,double>(L); }	//dirty hack.  _vec<1,T> has problems with the constructor, so i just switched it up.  oh well, i hope no one ever uses this
inline vec2d luaReadVec2d(lua_State *L) { return luaReadVec<2,double>(L); }
inline vec3d luaReadVec3d(lua_State *L) { return luaReadVec<3,double>(L); }
inline vec4d luaReadVec4d(lua_State *L) { return luaReadVec<4,double>(L); }


#ifdef LUA_GL_TRACE
#define LUAFstart(x)	static int PF_##x(lua_State *L) { int argc = lua_gettop(L);	con.print("lua trace: " #x "\n");
#else
#define LUAFstart(x)	static int PF_##x(lua_State *L) { int argc = lua_gettop(L);
#endif
#define LUAFend() return 0; }

#define LUAPb(i)		lua_toboolean(L, i)
#define LUAPd(i)		lua_tonumber(L, i)
#define LUAPf(i)		LUAPd(i)
#define LUAPi(i)		lua_tointeger(L, i)
#define LUAPs(i)		lua_tostring(L, i)

//some implementations define GLhandleARB as "unsigned int"
//others (apple) define it as "void*" ... even though the returned values are all integers ...
#define LUAPglh(i)		(GLhandleARB)lua_tointeger(L, i)

#define LUAF(x)			LUAFstart(x) x(); LUAFend()
#define LUAFb(x)		LUAFstart(x) x(LUAPb(1)); LUAFend()
#define LUAFd(x)		LUAFstart(x) x(LUAPd(1)); LUAFend()
#define LUAFi(x)		LUAFstart(x) x(LUAPi(1)); LUAFend()
#define LUAFglh(x)		LUAFstart(x) x(LUAPglh(1)); LUAFend()
#define LUAF2i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2)); LUAFend()
#define LUAF2d(x)		LUAFstart(x) x(LUAPd(1), LUAPd(2)); LUAFend()
#define LUAF2glh(x)		LUAFstart(x) x(LUAPglh(1), LUAPglh(2)); LUAFend()
#define LUAF1i1d(x)		LUAFstart(x) x(LUAPi(1), LUAPd(2)); LUAFend()
#define LUAF3i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2), LUAPi(3)); LUAFend()
#define LUAF3d(x)		LUAFstart(x) x(LUAPd(1), LUAPd(2), LUAPd(3)); LUAFend()
#define LUAF4b(x)		LUAFstart(x) x(LUAPb(1), LUAPb(2), LUAPb(3), LUAPb(4)); LUAFend()
#define LUAF4i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2), LUAPi(3), LUAPi(4)); LUAFend()
#define LUAF4d(x)		LUAFstart(x) x(LUAPd(1), LUAPd(2), LUAPd(3), LUAPd(4)); LUAFend()
#define LUAF5i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5)); LUAFend()
#define LUAF6d(x)		LUAFstart(x) x(LUAPd(1), LUAPd(2), LUAPd(3), LUAPd(4), LUAPd(5), LUAPd(6)); LUAFend()
#define LUAF6i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5), LUAPi(6)); LUAFend()
#define LUAF7i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5), LUAPi(6), LUAPi(7)); LUAFend()
#define LUAF8i(x)		LUAFstart(x) x(LUAPi(1), LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5), LUAPi(6), LUAPi(7), LUAPi(8)); LUAFend()

//returning values...
#define LUAFri(x)		LUAFstart(x) lua_pushinteger(L, x()); return 1; LUAFend()
#define LUAFrglh(x)		LUAFstart(x) lua_pushinteger(L, (int)x()); return 1; LUAFend();
#define LUAFirb(x)		LUAFstart(x) lua_pushboolean(L, x(LUAPi(1))); return 1; LUAFend()
#define LUAFiri(x)		LUAFstart(x) lua_pushinteger(L, x(LUAPi(1))); return 1; LUAFend()
#define LUAFirglh(x)	LUAFstart(x) lua_pushinteger(L, (int)x(LUAPi(1))); return 1; LUAFend();

//special case for generating one thing and returning it.  glGenTextures, glGenFramebuffers, glGenRenderbuffers, etc ...
#define LUAFuip(x)	LUAFstart(x) GLuint id; x(1, &id); lua_pushinteger(L, id); return 1; LUAFend()

//used for vertex attributes which have different deafults for different parameters
//this allows multiple parameter/function bindings for one lua function
#define	LUAFvarcase(count, code)	if (argc == count) { code; } else
#define LUAFvarendcase(expect)	{ luaL_error(L, "expected " expect " parameters"); }

#define LUAF2ivar1i4d(x)	\
LUAFstart(x)	\
	LUAFvarcase(3, x##i(LUAPi(1), LUAPi(2), LUAPi(3) ) )	\
	LUAFvarcase(6, x##dv(LUAPi(1), LUAPi(2), vec4d(LUAPd(3), LUAPd(4), LUAPd(5), LUAPd(6)) ) )	\
	LUAFvarendcase("3 or 6")	\
LUAFend()

#define LUAF2ivar1f4f(x)	\
LUAFstart(x)	\
	LUAFvarcase(3, x##f(LUAPi(1), LUAPi(2), LUAPf(3) ) )	\
	LUAFvarcase(6, x##fv(LUAPi(1), LUAPi(2), vec4f(LUAPf(3), LUAPf(4), LUAPf(5), LUAPf(6)) ) )	\
	LUAFvarendcase("3 or 6")	\
LUAFend()

//hack to let the user specify many enable/disable's at once
#define LUAFimany(x)	\
LUAFstart(x)	\
	for (int i = 1; i <= argc; i++) {	\
		x(LUAPi(i));	\
	}	\
LUAFend()


#define LUAFloopCaseBegin()	\
	if (false) {
	
#define LUAFloopCaseString()	\
	} else if (lua_isstring(L, -2)) {	\
		const char *name = lua_tostring(L, -2);
		
#define LUAFloopCaseNumber()	\
		} else if (lua_isnumber(L, -2)) {	\
			float num = lua_tonumber(L, -2);	\

#define LUAFloopCaseEnd()	\
	} else {	\
		luaL_error(L, "found an unknown param type: %s", lua_typename(L, lua_type(L, -2)));	\
	}
	
#define LUAFloopCaseStringEnd()	\
	} else {	\
		luaL_error(L, "found an unknown param: %s", safestr(name));	\
	}

#define LUAFbeginTableLoop()	\
	int t = lua_gettop(L);	\
	if (!lua_istable(L,t)) {	\
		luaL_error(L, "expected a table");	\
	}	\
	lua_pushnil(L);	\
	while(lua_next(L, t) != 0) {
	
#define LUAFendTableLoop()	\
		lua_pop(L,1);	\
	}
//	assert(lua_gettop(L) == t);

//for now it assumes if we have only one param then it'd better be a table
//and that we want to loop through its params
#define LUAFbeginTableLoopCase()	\
	if (argc == 1) {	\
		LUAFbeginTableLoop()
		
//code goes here
#define LUAFendTableLoopCase()	\
		LUAFendTableLoop()	\
	} else


#endif	//GLUSH_PROGRAM_H
